#!/usr/bin/perl -w
#
# Copyright 2001 Jeff Garzik <jgarzik@mandrakesoft.com>
#
# This software may be used and distributed according to the terms
# of the GNU General Public License, incorporated herein by reference.
#
#
# Run "mkbuild.pl -h" for a usage summary.
#
# Makefile.meta datafile format:
# 2D table with columns separated by formfeeds,
# and rows separated by newlines.  Columns 1-N meanings
# differ based on column 0 value.
#
# Column 0: 'D' or 'F' character -- directory or file
# Column 1(D): sub-directory containing kernel module
# Column 2(D): Config.in CONFIG_xxx symbol for this sub-directory
# Column 3(D): Config.in line, verbatim, for this sub-directory
# Column 4(D): Filename of target, from [LO]_TARGET in subdir/Makefile
# Column 1(F): Base name of kernel module
# Column 2(F): Config.in CONFIG_xxx symbol for this module
# Column 3(F): Config.in line, verbatim, for this module
# Column 4(F): Export symbols flag: 'Y' or 'N'


use strict;
use Getopt::Std;

my $MetadataFile = "Makefile.meta";

my (%opts, $InstDir, $InstFile);
my ($FileDesc, $ConfigIn, $ExportSyms, $ConfigSym, @db);


&parse_cmd_line;

if ($opts{'c'}) {
	print STDERR "Creating new database (ignoring any existing one).\n";
} else {
	&read_database;
}

if ($opts{'r'}) {
	# do nothing
} elsif ($InstDir) {
	&update_from_dir;
} else {
	&update_from_file;
}

&update_database;
&regen_files;

exit(0);

##########################################################################

sub regen_makefile {
	my (@ent, @export_objs, @dirs);
	my ($modname, @modules, $rc);
	
	foreach (@db) {
		@ent = split(/\f/);
		if ($ent[0] eq "D") {
			push(@dirs, $ent[1]);
		}
		elsif ($ent[0] eq "F") {
			$modname = $ent[1] . ".o";
			if ($ent[4] eq "Y") {
				push(@export_objs, $modname);
			}
		}
	}

	open (F, ">Makefile") or die "Cannot create new Makefile: $!\n";
	print F <<"EOM";
#
# THIS IS AN AUTOMATICALLY GENERATED FILE.  DO NOT EDIT.
#

EOM

	if (!@dirs && !$modname) {
		close(F);
		return;
	}

	print F <<"EOM";

O_TARGET := 3rdparty.o

EOM
	
	print F "export-objs := ", join(" ", @export_objs), "\n\n"
		if (@export_objs);
	print F "mod-subdirs := ", join(" ", @dirs), "\n\n"
		if (@dirs);

	foreach (@db) {
		@ent = split(/\f/);
		if ($ent[0] eq "D") {
			printf F "subdir-\$(%s) += %s\n", $ent[2], $ent[1];
		}
		elsif ($ent[0] eq "F") {
			printf F "obj-\$(%s) += %s.o\n", $ent[2], $ent[1];
		}
	}

	print F <<"EOM";

include \$(TOPDIR)/Rules.make

EOM

	close(F);

	$rc = 0;
	$rc |= 0x01 if @dirs;
	$rc |= 0x02 if $modname;
	return $rc;
}


sub regen_drivers_makefile {
	my ($havemask) = @_;
	my $have_dirs = $havemask & 0x01;
	my $have_files = $havemask & 0x02;
	my (@ent);
	
	open (F, ">Makefile.drivers") or
		die "Cannot create Makefile.drivers: $!\n";
	print F <<"EOM";
#
# THIS IS AN AUTOMATICALLY GENERATED FILE.  DO NOT EDIT.
#

DRIVERS-y :=
DRIVERS += 3rdparty/3rdparty.o
EOM

	foreach (@db) {
		@ent = split(/\f/);
		if ($ent[0] eq "D") {
			(my $t = $ent[4]) =~ s|\.o$||;
			printf F "DRIVERS-\$(%s) += 3rdparty/%s/%s\n",
				 $ent[2], $t, $ent[4];
		}
	}

	print F <<"EOM";

DRIVERS += \$(DRIVERS-y)

EOM

	close(F);
}


sub regen_config_in {
	my (@ent);

	open (F, ">Config.in") or
		die "Cannot create Config.in: $!\n";
	print F <<"EOM";
#
# THIS IS AN AUTOMATICALLY GENERATED FILE.  DO NOT EDIT.
#

mainmenu_option next_comment
comment 'Unofficial 3rd party kernel additions'

EOM
	
	foreach (@db) {
		@ent = split(/\f/);
		print F $ent[3], "\n";
	}

	print F "\n\nendmenu\n";
	close(F);
}


sub regen_files {
	my $havemask = &regen_makefile;
	&regen_drivers_makefile($havemask);
	&regen_config_in;
}


sub update_database {
	my ($s);

	open (F, ">$MetadataFile") or
		die "Cannot create new file $MetadataFile: $!\n";

	foreach $s (@db) {
		print F "$s\n";
	}

	close F;
}


sub find_target {
	open(F, "$InstDir/Makefile") or die "$InstDir/Makefile: $!\n";
	while (<F>) {
		chomp;
		return $1 if (/[LO]_TARGET.*:=\s+(\S+)/);
	}
	close(F);
}


sub update_from_dir {
	die "No CONFIG_xxx symbol specified for directory\n"
		unless ($ConfigSym);
	die "No Makefile in $InstDir.\n" if (! -r "$InstDir/Makefile");
	if (-r "$InstDir/Config.in") {
		$ConfigIn = sprintf "source 3rdparty/%s/Config.in", $InstDir;
	} else {
		$ConfigIn = sprintf "tristate '%s' %s", $FileDesc, $ConfigSym;
	}
	my $target = &find_target;
	push(@db, "D\f$InstDir\f$ConfigSym\f$ConfigIn\f$target");
}


sub file_find_config_sym {
	open(F, $InstFile) or die "Cannot open file $InstFile: $!\n";
	while (<F>) {
		chomp;
		$ConfigSym = $1 if (/^c-help-symbol:\s+(CONFIG_\S+)/);
	}
	close(F);
}


sub file_find_config_in {
	open(F, $InstFile) or die "Cannot open file $InstFile: $!\n";
	while (<F>) {
		chomp;
		$ConfigIn = $1 if (/^config-in:\s+(.*)$/);
	}
	close(F);
}


sub update_from_file {
	my ($basename) = ($InstFile =~ /^(.*)\.c$/);

	if (!$ConfigSym) {
		&file_find_config_sym;
		die "No CONFIG_xxx variable assigned to $InstFile\n"
			if (!$ConfigSym);
	}

	&file_find_config_in;
	$ConfigIn = sprintf "tristate '%s' %s\n", $FileDesc, $ConfigSym
		if (!$ConfigIn);

	my $ent = sprintf "F\f$basename\f$ConfigSym\f%s\f%s\n",
			  $ConfigIn, $ExportSyms ? "Y" : "N";
	push(@db, $ent);
}


sub read_database {
	if (open F, $MetadataFile) {
		@db = <F>;
		close F;
		chomp @db;
	} else {
		print STDERR "No database found, creating new one.\n";
	}
}


sub show_usage {
	print <<"EOM";
Usage:
mkbuild.pl [options]

Options:
-h, -?		Display this usage summary.
-c		Clear database (do not read it on startup).
-d DIR		Setup module installed in DIR sub-directory.
-D \"my desc\"	Override Config.in description for FILE, with -f.
		Provide Config.in line instead of using Config.in
		file in sub-directory DIR, with -d.
-e		Flag FILE as a module which exports symbols.
		Only valid with -f.
-f FILE		Setup module from file FILE.
-r		Do not add anything, simply regenerate
		make and config.in files from metadata.
		-d, -f, and related options are ignored.
-s PREFIX	Override Config.in and Makefile CONFIG_xxx
		symbol with PREFIX.  If the string 'CONFIG_'
		is not present, it is prepended.  If any chars
		in the PREFIX are lowercase, they are made uppercase.
		If -s is missing and -d is present, the prefix will
		be derived from the sub-directory name.
EOM
}


sub parse_cmd_line {
	getopts('cd:D:ef:hrs:?', \%opts);

	if ($opts{'?'} || $opts{'h'}) {
		&show_usage;
		exit(0);
	}

	if ($opts{'s'}) {
		if ($opts{'s'} !~ /^config_/i) {
			$ConfigSym = "CONFIG_" . $opts{'s'};
		} else {
			$ConfigSym = $opts{'s'};
		}
		$ConfigSym =~ tr/a-z/A-Z/;
	}
	$FileDesc = $opts{'D'} if $opts{'D'};

	if ($opts{'r'}) {
		# do nothing
	}

	elsif ($opts{'d'}) {
		$InstDir = $opts{'d'};
		die "Directory $InstDir does not exist\n" if (! -e $InstDir);
		die "$InstDir is not a directory\n" if (! -d $InstDir);
		die "Directory $InstDir is not readable\n" if (! -r $InstDir);
		die "Directory $InstDir is not a subdir\n"
			if (($InstDir =~ m#^/#) || ($InstDir =~ m#^\.#));

		if (!$ConfigSym) {
			$ConfigSym = 'CONFIG_' . $InstDir;
			$ConfigSym =~ tr/a-z/A-Z/;
			$ConfigSym =~ s/[\-\. ]/_/g;
		}

		print "Updating from directory $InstDir.\n";
	}

	elsif ($opts{'f'} ) {
		$InstFile = $opts{'f'};
		die "$InstFile is not a .c file\n"
			if ($InstFile !~ /\.c$/);
		die "File $InstFile does not exist\n" if (! -e $InstFile);
		die "$InstFile is not a file\n" if (! -f $InstFile);
		die "File $InstFile is not readable\n" if (! -r $InstFile);
		print "Updating from file $InstFile.\n";
	
		$ExportSyms = $opts{'e'};
	}
	
	else {
		print STDERR "No valid operation mode, aborting.\n";
		exit(1);
	}
}


