/*
 *  Copyright (c) by Jaroslav Kysela <perex@suse.cz>
 *                   Cirrus Logic, Inc.
 *  Routines for control of Cirrus Logic CS461x chips
 *
 *  BUGS:
 *    --
 *
 *  TODO:
 *    --
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define SND_MAIN_OBJECT_FILE
#include "../../include/driver.h"
#include "../../include/control.h"
#include "../../include/info.h"
#include "../../include/cs461x.h"

/*
 *  constants
 */

#if 0
#define ACCEPT_VALID		/* REQUIRED ONLY FOR OSS EMULATION */
#endif

#define CS461X_BA0_SIZE		0x2000
#define CS461X_BA1_DATA0_SIZE	0x3000
#define CS461X_BA1_DATA1_SIZE	0x3800
#define CS461X_BA1_PRG_SIZE	0x7000
#define CS461X_BA1_REG_SIZE	0x0100

/*
 *  common I/O routines
 */

static void snd_cs461x_poke(cs461x_t *codec, unsigned long reg, unsigned int val)
{
	writel(val, &codec->ba1.idx[(reg >> 16) & 3][(reg >> 2) & 0x3fff]);
}

static unsigned int snd_cs461x_peek(cs461x_t *codec, unsigned long reg)
{
	return readl(&codec->ba1.idx[(reg >> 16) & 3][(reg >> 2) & 0x3fff]);
}

static void snd_cs461x_pokeBA0(cs461x_t *codec, unsigned long reg, unsigned int val)
{
	writel(val, &codec->ba0[reg >> 2]);
}

static unsigned int snd_cs461x_peekBA0(cs461x_t *codec, unsigned long reg)
{
	return readl(&codec->ba0[reg >> 2]);
}

static void snd_cs461x_codec_write(void *private_data,
				   unsigned short reg,
				   unsigned short val)
{
	/*
	 *  1. Write ACCAD = Command Address Register = 46Ch for AC97 register address
	 *  2. Write ACCDA = Command Data Register = 470h    for data to write to AC97
	 *  3. Write ACCTL = Control Register = 460h for initiating the write
	 *  4. Read ACCTL = 460h, DCV should be reset by now and 460h = 07h
	 *  5. if DCV not cleared, break and return error
	 */
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, );
	int count;

	/*
	 *  Setup the AC97 control registers on the CS461x to send the
	 *  appropriate command to the AC97 to perform the read.
	 *  ACCAD = Command Address Register = 46Ch
	 *  ACCDA = Command Data Register = 470h
	 *  ACCTL = Control Register = 460h
	 *  set DCV - will clear when process completed
	 *  reset CRW - Write command
	 *  set VFRM - valid frame enabled
	 *  set ESYN - ASYNC generation enabled
	 *  set RSTN - ARST# inactive, AC97 codec not reset
         */
	snd_cs461x_pokeBA0(codec, BA0_ACCAD, reg);
	snd_cs461x_pokeBA0(codec, BA0_ACCDA, val);
	snd_cs461x_pokeBA0(codec, BA0_ACCTL, ACCTL_DCV | ACCTL_VFRM |
				             ACCTL_ESYN | ACCTL_RSTN);
	for (count = 0; count < 4000; count++) {
		/*
		 *  First, we want to wait for a short time.
		 */
		udelay(10);
		/*
		 *  Now, check to see if the write has completed.
		 *  ACCTL = 460h, DCV should be reset by now and 460h = 07h
		 */
		if (!(snd_cs461x_peekBA0(codec, BA0_ACCTL) & ACCTL_DCV))
			return;
	}
	snd_printk("cs461x: AC'97 write problem, reg = 0x%x, val = 0x%x\n", reg, val);
}

static unsigned short snd_cs461x_codec_read(void *private_data,
					    unsigned short reg)
{
	/*
	 *  1. Write ACCAD = Command Address Register = 46Ch for AC97 register address
	 *  2. Write ACCDA = Command Data Register = 470h    for data to write to AC97 
	 *  3. Write ACCTL = Control Register = 460h for initiating the write
	 *  4. Read ACCTL = 460h, DCV should be reset by now and 460h = 17h
	 *  5. if DCV not cleared, break and return error
	 *  6. Read ACSTS = Status Register = 464h, check VSTS bit
	 */

	int count;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);

	snd_cs461x_peekBA0(codec, BA0_ACSDA);

	/*
	 *  Setup the AC97 control registers on the CS461x to send the
	 *  appropriate command to the AC97 to perform the read.
	 *  ACCAD = Command Address Register = 46Ch
	 *  ACCDA = Command Data Register = 470h
	 *  ACCTL = Control Register = 460h
	 *  set DCV - will clear when process completed
	 *  set CRW - Read command
	 *  set VFRM - valid frame enabled
	 *  set ESYN - ASYNC generation enabled
	 *  set RSTN - ARST# inactive, AC97 codec not reset
	 */

	snd_cs461x_pokeBA0(codec, BA0_ACCAD, reg);
	snd_cs461x_pokeBA0(codec, BA0_ACCDA, 0);
	snd_cs461x_pokeBA0(codec, BA0_ACCTL, ACCTL_DCV | ACCTL_CRW |
					     ACCTL_VFRM | ACCTL_ESYN |
					     ACCTL_RSTN);


	/*
	 *  Wait for the read to occur.
	 */
	for (count = 0; count < 1000; count++) {
		/*
		 *  First, we want to wait for a short time.
	 	 */
		udelay(10);
		/*
		 *  Now, check to see if the read has completed.
		 *  ACCTL = 460h, DCV should be reset by now and 460h = 17h
		 */
		if (!(snd_cs461x_peekBA0(codec, BA0_ACCTL) & ACCTL_DCV))
			goto ok1;
	}

	snd_printk("cs461x: AC'97 read problem (ACCTL_DCV), reg = 0x%x\n", reg);
	return 0xffff;

ok1:
	/*
	 *  Wait for the valid status bit to go active.
	 */
	for (count = 0; count < 100; count++) {
		/*
		 *  Read the AC97 status register.
		 *  ACSTS = Status Register = 464h
		 *  VSTS - Valid Status
		 */
		if (snd_cs461x_peekBA0(codec, BA0_ACSTS) & ACSTS_VSTS)
			goto ok2;
		udelay(10);
	}
	
	snd_printk("cs461x: AC'97 read problem (ACSTS_VSTS), reg = 0x%x\n", reg);
	return 0xffff;

ok2:
	/*
	 *  Read the data returned from the AC97 register.
	 *  ACSDA = Status Data Register = 474h
	 */
#if 0
	printk("e) reg = 0x%x, val = 0x%x, BA0_ACCAD = 0x%x\n", reg,
			snd_cs461x_peekBA0(codec, BA0_ACSDA),
			snd_cs461x_peekBA0(codec, BA0_ACCAD));
#endif
	return snd_cs461x_peekBA0(codec, BA0_ACSDA);
}

/*
 *  Chip initialization
 */

int snd_cs461x_download(cs461x_t *codec,
			unsigned int *src,
                        unsigned long offset,
                        unsigned long len)
{
	unsigned int *dst;

	if ((offset & 3) || (len & 3))
		return -EINVAL;
	dst = codec->ba1.idx[(offset >> 16) & 3];
	dst += ((offset & 0xFFFF) << 2) / sizeof(unsigned int);
	len /= sizeof(unsigned int);

	/* writel already converts 32-bit value to right endianess */
	while (len-- > 0) {
		writel(*src++, dst);
		dst++;
	}
	return 0;
}

/* 3*1024 parameter, 3.5*1024 sample, 2*3.5*1024 code */
#define BA1_DWORD_SIZE		(13 * 1024 + 512)
#define BA1_MEMORY_COUNT	3

struct BA1struct {
	struct {
		unsigned long offset;
		unsigned long size;
	} memory[BA1_MEMORY_COUNT];
	unsigned int map[BA1_DWORD_SIZE];
};

static
#include "cs461x_image.h"

int snd_cs461x_download_image(cs461x_t *codec)
{
	int idx, err;
	unsigned long offset = 0;

	for (idx = 0; idx < BA1_MEMORY_COUNT; idx++) {
		if ((err = snd_cs461x_download(codec,
					       &BA1Struct.map[offset],
					       BA1Struct.memory[idx].offset,
					       BA1Struct.memory[idx].size)) < 0)
			return err;
		offset += BA1Struct.memory[idx].size >> 2;
	}	
	return 0;
}

/*
 *  Chip reset
 */

static void snd_cs461x_reset(cs461x_t *codec)
{
	int idx;

	/*
	 *  Write the reset bit of the SP control register.
	 */
	snd_cs461x_poke(codec, BA1_SPCR, SPCR_RSTSP);

	/*
	 *  Write the control register.
	 */
	snd_cs461x_poke(codec, BA1_SPCR, SPCR_DRQEN);

	/*
	 *  Clear the trap registers.
	 */
	for (idx = 0; idx < 8; idx++) {
		snd_cs461x_poke(codec, BA1_DREG, DREG_REGID_TRAP_SELECT + idx);
		snd_cs461x_poke(codec, BA1_TWPR, 0xFFFF);
	}
	snd_cs461x_poke(codec, BA1_DREG, 0);

	/*
	 *  Set the frame timer to reflect the number of cycles per frame.
	 */
	snd_cs461x_poke(codec, BA1_FRMT, 0xadf);
}

static void snd_cs461x_clear_serial_FIFOs(cs461x_t *codec)
{
	int idx, loop, powerdown = 0;
	unsigned int tmp;

	/*
	 *  See if the devices are powered down.  If so, we must power them up first
	 *  or they will not respond.
	 */
	tmp = snd_cs461x_peekBA0(codec, BA0_CLKCR1);
	if (!(tmp & CLKCR1_SWCE)) {
		snd_cs461x_pokeBA0(codec, BA0_CLKCR1, tmp | CLKCR1_SWCE);
		powerdown = 1;
	}

	/*
	 *  We want to clear out the serial port FIFOs so we don't end up playing
	 *  whatever random garbage happens to be in them.  We fill the sample FIFOS
	 *  with zero (silence).
         */
	snd_cs461x_pokeBA0(codec, BA0_SERBWP, 0);

	/*
	 *  Fill all 256 sample FIFO locations.
	 */
	for (idx = 0; idx < 256; idx++) {
		/*
		 *  Make sure the previous FIFO write operation has completed.
		 */
		for (loop = 0; loop < 5; loop++) {
			udelay(50);
			if (!(snd_cs461x_peekBA0(codec, BA0_SERBST) & SERBST_WBSY))
				break;
		}
		if (snd_cs461x_peekBA0(codec, BA0_SERBST) & SERBST_WBSY) {
			if (powerdown)
				snd_cs461x_pokeBA0(codec, BA0_CLKCR1, tmp);
		}
		/*
		 *  Write the serial port FIFO index.
		 */
		snd_cs461x_pokeBA0(codec, BA0_SERBAD, idx);
		/*
		 *  Tell the serial port to load the new value into the FIFO location.
		 */
		snd_cs461x_pokeBA0(codec, BA0_SERBCM, SERBCM_WRC);
	}
	/*
	 *  Now, if we powered up the devices, then power them back down again.
	 *  This is kinda ugly, but should never happen.
	 */
	if (powerdown)
		snd_cs461x_pokeBA0(codec, BA0_CLKCR1, tmp);
}

static void snd_cs461x_powerup_dac(cs461x_t *codec)
{
	int count;
	unsigned int tmp;

	/*
	 *  Power on the DACs on the AC97 codec.  We turn off the DAC
	 *  powerdown bit and write the new value of the power control
	 *  register.
	 */
	tmp = snd_cs461x_codec_read(codec, AC97_POWERDOWN);
	if (tmp & 2)	/* already */
		return;
	snd_cs461x_codec_write(codec, AC97_POWERDOWN, tmp & 0xfdff);

	/*
	 *  Now, we wait until we sample a DAC ready state.
	 */
	for (count = 0; count < 32; count++) {
		/*
		 *  First, lets wait a short while to let things settle out a
		 *  bit, and to prevent retrying the read too quickly.
		 */
		udelay(50);

		/*
		 *  Read the current state of the power control register.
		 */
		if (snd_cs461x_codec_read(codec, AC97_POWERDOWN) & 2)
			break;
	}
	
	/*
	 *  Check the status..
	 */
	if (!(snd_cs461x_codec_read(codec, AC97_POWERDOWN) & 2))
		snd_printk("cs461x: powerup DAC failed\n");
}

static void snd_cs461x_powerup_adc(cs461x_t *codec)
{
	int count;
	unsigned int tmp;

	/*
	 *  Power on the ADCs on the AC97 codec.  We turn off the DAC
	 *  powerdown bit and write the new value of the power control
	 *  register.
	 */
	tmp = snd_cs461x_codec_read(codec, AC97_POWERDOWN);
	if (tmp & 1)	/* already */
		return;
	snd_cs461x_codec_write(codec, AC97_POWERDOWN, tmp & 0xfeff);

	/*
	 *  Now, we wait until we sample a ADC ready state.
	 */
	for (count = 0; count < 32; count++) {
		/*
		 *  First, lets wait a short while to let things settle out a
		 *  bit, and to prevent retrying the read too quickly.
		 */
		udelay(50);

		/*
		 *  Read the current state of the power control register.
		 */
		if (snd_cs461x_codec_read(codec, AC97_POWERDOWN) & 1)
			break;
	}

	/*
	 *  Check the status..
	 */
	if (!(snd_cs461x_codec_read(codec, AC97_POWERDOWN) & 1))
		snd_printk("cs461x: powerup ADC failed\n");
}

static void snd_cs461x_proc_start(cs461x_t *codec)
{
	int cnt;

	/*
	 *  Set the frame timer to reflect the number of cycles per frame.
	 */
	snd_cs461x_poke(codec, BA1_FRMT, 0xadf);
	/*
	 *  Turn on the run, run at frame, and DMA enable bits in the local copy of
	 *  the SP control register.
	 */
	snd_cs461x_poke(codec, BA1_SPCR, SPCR_RUN | SPCR_RUNFR | SPCR_DRQEN);
	/*
	 *  Wait until the run at frame bit resets itself in the SP control
	 *  register.
	 */
	for (cnt = 0; cnt < 25; cnt++) {
		udelay(50);
		if (!(snd_cs461x_peek(codec, BA1_SPCR) & SPCR_RUNFR))
			break;
	}

	if (snd_cs461x_peek(codec, BA1_SPCR) & SPCR_RUNFR)
		snd_printk("cs461x: SPCR_RUNFR never reset\n");
}

static void snd_cs461x_proc_stop(cs461x_t *codec)
{
	/*
	 *  Turn off the run, run at frame, and DMA enable bits in the local copy of
	 *  the SP control register.
	 */
	snd_cs461x_poke(codec, BA1_SPCR, 0);
}

/*
 *  Sample rate routines
 */

#define GOF_PER_SEC 200

static void snd_cs461x_set_play_sample_rate(cs461x_t *codec, unsigned int rate)
{
	unsigned long flags;
	unsigned int tmp1, tmp2;
	unsigned int phiIncr;
	unsigned int correctionPerGOF, correctionPerSec;

	/*
	 *  Compute the values used to drive the actual sample rate conversion.
	 *  The following formulas are being computed, using inline assembly
	 *  since we need to use 64 bit arithmetic to compute the values:
	 *
	 *  phiIncr = floor((Fs,in * 2^26) / Fs,out)
	 *  correctionPerGOF = floor((Fs,in * 2^26 - Fs,out * phiIncr) /
         *                                   GOF_PER_SEC)
         *  ulCorrectionPerSec = Fs,in * 2^26 - Fs,out * phiIncr -M
         *                       GOF_PER_SEC * correctionPerGOF
	 *
	 *  i.e.
	 *
	 *  phiIncr:other = dividend:remainder((Fs,in * 2^26) / Fs,out)
	 *  correctionPerGOF:correctionPerSec =
	 *      dividend:remainder(ulOther / GOF_PER_SEC)
	 */
	tmp1 = rate << 16;
	phiIncr = tmp1 / 48000;
	tmp1 -= phiIncr * 48000;
	tmp1 <<= 10;
	phiIncr <<= 10;
	tmp2 = tmp1 / 48000;
	phiIncr += tmp2;
	tmp1 -= tmp2 * 48000;
	correctionPerGOF = tmp1 / GOF_PER_SEC;
	tmp1 -= correctionPerGOF * GOF_PER_SEC;
	correctionPerSec = tmp1;

	/*
	 *  Fill in the SampleRateConverter control block.
	 */
	spin_lock_irqsave(&codec->reg_lock, flags);
	snd_cs461x_poke(codec, BA1_PSRC,
	  ((correctionPerSec << 16) & 0xFFFF0000) | (correctionPerGOF & 0xFFFF));
	snd_cs461x_poke(codec, BA1_PPI, phiIncr);
	spin_unlock_irqrestore(&codec->reg_lock, flags);
}

static void snd_cs461x_set_capture_sample_rate(cs461x_t *codec, unsigned int rate)
{
	unsigned long flags;
	unsigned int phiIncr, coeffIncr, tmp1, tmp2;
	unsigned int correctionPerGOF, correctionPerSec, initialDelay;
	unsigned int frameGroupLength, cnt;

	/*
	 *  We can only decimate by up to a factor of 1/9th the hardware rate.
	 *  Correct the value if an attempt is made to stray outside that limit.
	 */
	if ((rate * 9) < 48000)
		rate = 48000 / 9;

	/*
	 *  We can not capture at at rate greater than the Input Rate (48000).
	 *  Return an error if an attempt is made to stray outside that limit.
	 */
	if (rate > 48000)
		rate = 48000;

	/*
	 *  Compute the values used to drive the actual sample rate conversion.
	 *  The following formulas are being computed, using inline assembly
	 *  since we need to use 64 bit arithmetic to compute the values:
	 *
	 *     coeffIncr = -floor((Fs,out * 2^23) / Fs,in)
	 *     phiIncr = floor((Fs,in * 2^26) / Fs,out)
	 *     correctionPerGOF = floor((Fs,in * 2^26 - Fs,out * phiIncr) /
	 *                                GOF_PER_SEC)
	 *     correctionPerSec = Fs,in * 2^26 - Fs,out * phiIncr -
	 *                          GOF_PER_SEC * correctionPerGOF
	 *     initialDelay = ceil((24 * Fs,in) / Fs,out)
	 *
	 * i.e.
	 *
	 *     coeffIncr = neg(dividend((Fs,out * 2^23) / Fs,in))
	 *     phiIncr:ulOther = dividend:remainder((Fs,in * 2^26) / Fs,out)
	 *     correctionPerGOF:correctionPerSec =
	 * 	    dividend:remainder(ulOther / GOF_PER_SEC)
	 *     initialDelay = dividend(((24 * Fs,in) + Fs,out - 1) / Fs,out)
	 */

	tmp1 = rate << 16;
	coeffIncr = tmp1 / 48000;
	tmp1 -= coeffIncr * 48000;
	tmp1 <<= 7;
	coeffIncr <<= 7;
	coeffIncr += tmp1 / 48000;
	coeffIncr ^= 0xFFFFFFFF;
	coeffIncr++;
	tmp1 = 48000 << 16;
	phiIncr = tmp1 / rate;
	tmp1 -= phiIncr * rate;
	tmp1 <<= 10;
	phiIncr <<= 10;
	tmp2 = tmp1 / rate;
	phiIncr += tmp2;
	tmp1 -= tmp2 * rate;
	correctionPerGOF = tmp1 / GOF_PER_SEC;
	tmp1 -= correctionPerGOF * GOF_PER_SEC;
	correctionPerSec = tmp1;
	initialDelay = ((48000 * 24) + rate - 1) / rate;

	/*
	 *  Fill in the VariDecimate control block.
	 */
	spin_lock_irqsave(&codec->reg_lock, flags);
	snd_cs461x_poke(codec, BA1_CSRC,
		((correctionPerSec << 16) & 0xFFFF0000) | (correctionPerGOF & 0xFFFF));
	snd_cs461x_poke(codec, BA1_CCI, coeffIncr);
	snd_cs461x_poke(codec, BA1_CD,
		(((BA1_VARIDEC_BUF_1 + (initialDelay << 2)) << 16) & 0xFFFF0000) | 0x80);
	snd_cs461x_poke(codec, BA1_CPI, phiIncr);
	spin_unlock_irqrestore(&codec->reg_lock, flags);

	/*
	 *  Figure out the frame group length for the write back task.  Basically,
	 *  this is just the factors of 24000 (2^6*3*5^3) that are not present in
	 *  the output sample rate.
	 */
	frameGroupLength = 1;
	for (cnt = 2; cnt <= 64; cnt *= 2) {
		if (((rate / cnt) * cnt) != rate)
			frameGroupLength *= 2;
	}
	if (((rate / 3) * 3) != rate) {
		frameGroupLength *= 3;
	}
	for (cnt = 5; cnt <= 125; cnt *= 5) {
		if (((rate / cnt) * cnt) != rate) 
			frameGroupLength *= 5;
        }

	/*
	 * Fill in the WriteBack control block.
	 */
	spin_lock_irqsave(&codec->reg_lock, flags);
	snd_cs461x_poke(codec, BA1_CFG1, frameGroupLength);
	snd_cs461x_poke(codec, BA1_CFG2, (0x00800000 | frameGroupLength));
	snd_cs461x_poke(codec, BA1_CCST, 0x0000FFFF);
	snd_cs461x_poke(codec, BA1_CSPB, ((65536 * rate) / 24000));
	snd_cs461x_poke(codec, (BA1_CSPB + 4), 0x0000FFFF);
	spin_unlock_irqrestore(&codec->reg_lock, flags);
}

/*
 *  PCM part
 */

static int snd_cs461x_playback_ioctl(void *private_data,
				     snd_pcm_subchn_t * subchn,
				     unsigned int cmd,
				     unsigned long *arg)
{
	int result;

	result = snd_pcm_lib_ioctl(private_data, subchn, cmd, arg);
	if (result < 0)
		return result;		
	return result;
}

static int snd_cs461x_capture_ioctl(void *private_data,
				    snd_pcm_subchn_t * subchn,
				    unsigned int cmd,
				    unsigned long *arg)
{
	int result;

	result = snd_pcm_lib_ioctl(private_data, subchn, cmd, arg);
	if (result < 0)
		return result;		
	return result;
}

static void snd_cs461x_copy_playback_buffer(snd_pcm_subchn_t * subchn,
					    cs461x_t * codec)
{
	snd_pcm_runtime_t *runtime = subchn->runtime;
	if (snd_pcm_format_width(runtime->format.format) == 16) {
		memcpy(codec->pbuf + (2048 * codec->ppingbuf++),
		       codec->pbuffer + (2048 * codec->pringbuf++),
		       2048);
		codec->ppingbuf &= 1;
		if (codec->psize <= (codec->pringbuf * 2048))
			codec->pringbuf = 0;
	} else {
		unsigned char *dest;
		unsigned char *src;
		int count = 1024;
		
		dest = codec->pbuf + (2048 * codec->ppingbuf++);
		src = codec->pbuffer + (1024 * codec->pringbuf++);
		while (count-- > 0) {
			*dest++ = 0x80;
			*dest++ = *src++;
		}
		codec->ppingbuf &= 1;
		if (codec->psize <= (codec->pringbuf * 1024))
			codec->pringbuf = 0;
	}
}

static int snd_cs461x_playback_trigger(void *private_data,
				       snd_pcm_subchn_t * subchn,
				       int cmd)
{
	unsigned long flags;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	unsigned int tmp;
	int result = 0;

	spin_lock_irqsave(&codec->reg_lock, flags);
	if (cmd == SND_PCM_TRIGGER_GO) {
		snd_cs461x_copy_playback_buffer(subchn, codec);
		tmp = snd_cs461x_peek(codec, BA1_PCTL);
		tmp &= 0x0000ffff;
		snd_cs461x_poke(codec, BA1_PCTL, codec->pctl | tmp);
	} else if (cmd == SND_PCM_TRIGGER_STOP) {
		tmp = snd_cs461x_peek(codec, BA1_PCTL);
		tmp &= 0x0000ffff;
		snd_cs461x_poke(codec, BA1_PCTL, tmp);
	} else {
		result = -EINVAL;
	}
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return result;
}

static void snd_cs461x_copy_capture_buffer(snd_pcm_subchn_t * subchn,
					   cs461x_t * codec)
{
	signed int b16;
	snd_pcm_runtime_t *runtime = subchn->runtime;

	if (codec->cff == 1) {
		memcpy(codec->cbuffer + (2048 * codec->cringbuf++),
		       codec->cbuf + (2048 * codec->cpingbuf++),
		       2048);
		codec->cpingbuf &= 1;
		if (codec->csize <= (codec->cringbuf * 2048))
			codec->cringbuf = 0;
	} else if (codec->cff == 2) {
		unsigned char *dest;
		unsigned char *src;
		int count = 512;

		dest = codec->cbuffer + (1024 * codec->cringbuf++);
		src = codec->cbuf + (2048 * codec->cpingbuf++);
		if (snd_pcm_format_width(runtime->format.format) == 16) {
			while (count-- > 0) {
				b16 = *((signed short *)src)++;
				b16 += *((signed short *)src)++;
				*((unsigned short *)dest)++ = b16 >> 1;
			}
		} else {
			while (count-- > 0) {
				src++;
				*dest++ = *src++ ^ 0x80;
				src++;
				*dest++ = *src++ ^ 0x80;
			}
		}
		codec->cpingbuf &= 1;
		if (codec->csize <= (codec->cringbuf * 1024))
			codec->cringbuf = 0;
	} else if (codec->cff == 4) {
		unsigned char *dest;
		unsigned char *src;
		int count = 512;
		
		dest = codec->cbuffer + (512 * codec->cringbuf++);
		src = codec->cbuf + (2048 * codec->cpingbuf++);
		while (count-- > 0) {
			b16 = *((signed short *)src)++;
			b16 += *((signed short *)src)++;
			*dest++ = (b16 >> 9) ^ 0x80;
		}
		codec->cpingbuf &= 1;
		if (codec->csize <= (codec->cringbuf * 512))
			codec->cringbuf = 0;
	}
}

static int snd_cs461x_capture_trigger(void *private_data,	
				      snd_pcm_subchn_t * subchn,
				      int cmd)
{
	unsigned long flags;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	unsigned int tmp;
	int result = 0;

	spin_lock_irqsave(&codec->reg_lock, flags);
	if (cmd == SND_PCM_TRIGGER_GO) {
		tmp = snd_cs461x_peek(codec, BA1_CCTL);
		tmp &= 0xffff0000;
		snd_cs461x_poke(codec, BA1_CCTL, codec->cctl | tmp);
	} else if (cmd == SND_PCM_TRIGGER_STOP) {
		tmp = snd_cs461x_peek(codec, BA1_CCTL);
		tmp &= 0xffff0000;
		snd_cs461x_poke(codec, BA1_CCTL, tmp);
	} else {
		result = -EINVAL;
	}
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return result;
}

static int snd_cs461x_playback_prepare(void *private_data,
				       snd_pcm_subchn_t * subchn)
{
	unsigned int tmp, tmp1;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;

	if (snd_pcm_lib_transfer_fragment(subchn) != 2048)
		snd_printk("cs461x: wrong playback count - %i\n", snd_pcm_lib_transfer_fragment(subchn));
	/* set the attenuation to 0dB */ 
	snd_cs461x_poke(codec, BA1_PVOL, 0x80008000);
	snd_cs461x_poke(codec, BA1_PBA, virt_to_bus(codec->pbuf));
	tmp1 = 16;
	if (runtime->format.voices < 2)
		tmp1 >>= 1;
	tmp = snd_cs461x_peek(codec, BA1_PDTC);
	tmp &= ~0x000003ff;
	tmp |= tmp1 - 1;
	snd_cs461x_poke(codec, BA1_PDTC, tmp);
	tmp = snd_cs461x_peek(codec, BA1_PFIE);
	tmp &= ~0x0000f03f;
	if (snd_pcm_format_unsigned(runtime->format.format))
		tmp |= 0x00008000;
	if (snd_pcm_format_big_endian(runtime->format.format))
		tmp |= 0x00004000;
	if (runtime->format.voices < 2)
		tmp |= 0x00002000;
	snd_cs461x_poke(codec, BA1_PFIE, tmp);
	codec->ppingbuf = 0;
	codec->pbuffer = runtime->dma_area->buf;
	codec->psize = snd_pcm_lib_transfer_size(subchn);
	codec->pringbuf = 0;
	snd_cs461x_set_play_sample_rate(codec, runtime->format.rate);
	return 0;
}

static int snd_cs461x_capture_prepare(void *private_data,
				      snd_pcm_subchn_t * subchn)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;

	if (snd_pcm_lib_transfer_fragment(subchn) != 2048)
		snd_printk("cs461x: wrong record count - %i\n", snd_pcm_lib_transfer_fragment(subchn));
	/* set the attenuation to 0dB */
	snd_cs461x_poke(codec, BA1_CVOL, 0x80008000);
	snd_cs461x_poke(codec, BA1_CBA, virt_to_bus(codec->cbuf));
	codec->cpingbuf = 0;
	codec->cbuffer = runtime->dma_area->buf;
	codec->csize = snd_pcm_lib_transfer_size(subchn);
	codec->cringbuf = 0;
	snd_cs461x_set_capture_sample_rate(codec, runtime->format.rate);
	codec->cff = 1;
	codec->cffc = 0;
	if (runtime->format.voices < 2)
		codec->cff <<= 1;
	if (snd_pcm_format_width(runtime->format.format) != 16)
		codec->cff <<= 1;
	return 0;
}

static unsigned int snd_cs461x_playback_pointer(void *private_data,
						snd_pcm_subchn_t * subchn)
{
	unsigned int result;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;

	result = snd_cs461x_peek(codec, BA1_PBA) - virt_to_bus(codec->pbuf);
	if (result > 2048)
		result -= 2048;
	if (snd_pcm_format_width(runtime->format.format) != 16)
		result >>= 1;
	return result;
}

static unsigned int snd_cs461x_capture_pointer(void *private_data,
					       snd_pcm_subchn_t * subchn)
{
	unsigned int result;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	snd_pcm_runtime_t *runtime = subchn->runtime;

	result = snd_cs461x_peek(codec, BA1_CBA) - virt_to_bus(codec->cbuf);
	if (result > 2048)
		result -= 2048;
	if (snd_pcm_format_width(runtime->format.format) != 16)
		result >>= 1;
	return result;
}

void snd_cs461x_interrupt(cs461x_t *codec)
{
	unsigned int status;

	/*
	 *  Read the Interrupt Status Register to clear the interrupt
	 */
	status = snd_cs461x_peekBA0(codec, BA0_HISR);
	if ((status & 0x7fffffff) == 0) {
		snd_cs461x_pokeBA0(codec, BA0_HICR, HICR_CHGM | HICR_IEV);
		return;
	}

	if ((status & HISR_VC0) && codec->pcm) {
		if (codec->playback_subchn) {
			snd_cs461x_copy_playback_buffer(codec->playback_subchn, codec);
			if (snd_pcm_format_width(codec->playback_subchn->runtime->format.format) == 16 ||
			    codec->ppingbuf == 1)
				snd_pcm_transfer_done(codec->playback_subchn);
		}
	}
	if ((status & HISR_VC1) && codec->pcm) {
		if (codec->capture_subchn) {
			snd_cs461x_copy_capture_buffer(codec->capture_subchn, codec);
			codec->cffc++;
			if (codec->cffc >= codec->cff) {
				codec->cffc = 0;
				snd_pcm_transfer_done(codec->capture_subchn);
			}
		}
	}
	if ((status & HISR_MIDI) && codec->rmidi) {
		unsigned char c;
		
		spin_lock(&codec->reg_lock);
		while ((snd_cs461x_peekBA0(codec, BA0_MIDSR) & MIDSR_RBE) == 0) {
			c = snd_cs461x_peekBA0(codec, BA0_MIDRP);
			if ((codec->midcr & MIDCR_RIE) == 0)
				continue;
			spin_unlock(&codec->reg_lock);
			snd_rawmidi_receive(codec->rmidi, &c, 1);
			spin_lock(&codec->reg_lock);
		}
		while ((snd_cs461x_peekBA0(codec, BA0_MIDSR) & MIDSR_TBF) == 0) {
			if ((codec->midcr & MIDCR_TIE) == 0)
				break;
			if (snd_rawmidi_transmit(codec->rmidi, &c, 1) != 1) {
				codec->midcr &= ~MIDCR_TIE;
				snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
				break;
			}
			snd_cs461x_pokeBA0(codec, BA0_MIDWP, c);
		}
		spin_unlock(&codec->reg_lock);
	}
	/*
	 *  EOI to the PCI part....reenables interrupts
	 */
	snd_cs461x_pokeBA0(codec, BA0_HICR, HICR_CHGM | HICR_IEV);
}

static snd_pcm_hardware_t snd_cs461x_playback =
{
	SND_PCM_CHNINFO_MMAP |
#ifdef ACCEPT_VALID
	SND_PCM_CHNINFO_MMAP_VALID |	/* NOT TRUE!!! OSS REQUIRES IT */
#endif
	SND_PCM_CHNINFO_BLOCK | SND_PCM_CHNINFO_INTERLEAVE |
	SND_PCM_CHNINFO_BLOCK_TRANSFER,		/* flags */
	SND_PCM_FMT_U8 | SND_PCM_FMT_S16_LE,	/* hardware formats */
	SND_PCM_RATE_CONTINUOUS | SND_PCM_RATE_8000_48000,
	5500,			/* min. rate */
	48000,			/* max. rate */
	1,			/* min. voices */
	2,			/* max. voices */
	2048,			/* min. fragment size */
	2048,			/* max. fragment size */
	3,			/* fragment align */
	0,			/* FIFO size (unknown) */
	4,			/* transfer block size */
	snd_cs461x_playback_ioctl,
	snd_cs461x_playback_prepare,
	snd_cs461x_playback_trigger,
	snd_cs461x_playback_pointer
};

static snd_pcm_hardware_t snd_cs461x_capture =
{
	SND_PCM_CHNINFO_MMAP |
#ifdef ACCEPT_VALID
	SND_PCM_CHNINFO_MMAP_VALID |	/* NOT TRUE!!! OSS REQUIRES IT */
#endif
	SND_PCM_CHNINFO_BLOCK | SND_PCM_CHNINFO_INTERLEAVE |
	SND_PCM_CHNINFO_BLOCK_TRANSFER,		/* flags */
	SND_PCM_FMT_U8 | SND_PCM_FMT_S16_LE,	/* hardware formats */
	SND_PCM_RATE_CONTINUOUS | SND_PCM_RATE_8000_48000,
	5500,			/* min. rate */
	48000,			/* max. rate */
	1,			/* min. voices */
	2,			/* max. voices */
	2048,			/* min. fragment size */
	2048,			/* max. fragment size */
	3,			/* fragment align */
	0,			/* FIFO size (unknown) */
	4,			/* transfer block size */
	snd_cs461x_capture_ioctl,
	snd_cs461x_capture_prepare,
	snd_cs461x_capture_trigger,
	snd_cs461x_capture_pointer
};

static int snd_cs461x_playback_open(void *private_data,
				    snd_pcm_subchn_t * subchn)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	int err, pg;

	if ((err = snd_pcm_dma_alloc(subchn, codec->dma1ptr, "CS461x - DAC")) < 0)
		return err;
	codec->pbuf = snd_malloc_pages(PAGE_SIZE, &pg, GFP_KERNEL);
	if (codec->pbuf == NULL) {
		snd_pcm_dma_free(subchn);
		return -ENOMEM;
	}
	codec->playback_subchn = subchn;
	subchn->runtime->hw = &snd_cs461x_playback;
	snd_pcm_set_mixer(subchn, codec->mixer->device, codec->ac97->me_playback);
	codec->active_ctrl(codec, 1);
	codec->amplifier_ctrl(codec, 1);

	return 0;
}

static int snd_cs461x_capture_open(void *private_data,
				   snd_pcm_subchn_t * subchn)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	int err, pg;

	if ((err = snd_pcm_dma_alloc(subchn, codec->dma2ptr, "CS461x - ADC")) < 0)
		return err;
	codec->cbuf = snd_malloc_pages(PAGE_SIZE, &pg, GFP_KERNEL);
	if (codec->cbuf == NULL) {
		snd_pcm_dma_free(subchn);
		return -ENOMEM;
	}
	codec->capture_subchn = subchn;
	subchn->runtime->hw = &snd_cs461x_capture;
	snd_pcm_set_mixer(subchn, codec->mixer->device, codec->ac97->me_capture);
	codec->active_ctrl(codec, 1);
	codec->amplifier_ctrl(codec, 1);

	return 0;
}

static int snd_cs461x_playback_close(void *private_data,
				     snd_pcm_subchn_t * subchn)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);

	codec->playback_subchn = NULL;
	if (codec->pbuf)
		snd_free_pages(codec->pbuf, PAGE_SIZE);
	snd_pcm_dma_free(subchn);
	codec->active_ctrl(codec, -1);
	codec->amplifier_ctrl(codec, -1);

	return 0;
}

static int snd_cs461x_capture_close(void *private_data,
				    snd_pcm_subchn_t * subchn)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);

	codec->capture_subchn = NULL;
	if (codec->cbuf)
		snd_free_pages(codec->cbuf, PAGE_SIZE);
	snd_pcm_dma_free(subchn);
	codec->active_ctrl(codec, -1);
	codec->amplifier_ctrl(codec, -1);

	return 0;
}

static void snd_cs461x_pcm_free(void *private_data)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, );
	codec->pcm = NULL;
}

int snd_cs461x_pcm(cs461x_t * codec, int device, snd_pcm_t ** rpcm)
{
	snd_pcm_t *pcm;
	int err;

	*rpcm = NULL;
	if ((err = snd_pcm_new(codec->card, "CS461x", device, 1, 1, &pcm)) < 0)
		return err;
	pcm->private_data = codec;
	pcm->private_free = snd_cs461x_pcm_free;

	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].private_data = codec;
	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].open = snd_cs461x_playback_open;
	pcm->chn[SND_PCM_CHANNEL_PLAYBACK].close = snd_cs461x_playback_close;
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].private_data = codec;
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].open = snd_cs461x_capture_open;
	pcm->chn[SND_PCM_CHANNEL_CAPTURE].close = snd_cs461x_capture_close;

	/* global setup */
	pcm->info_flags = SND_PCM_INFO_PLAYBACK | SND_PCM_INFO_CAPTURE |
			  SND_PCM_INFO_DUPLEX;
	strcpy(pcm->name, "CS461x");
	*rpcm = codec->pcm = pcm;
	return 0;
}

/*
 *  Mixer routines
 */

static void snd_cs461x_mixer_write(void *private_data, unsigned short reg, unsigned short val)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, );
	int val2 = 0;
	codec->active_ctrl(codec, 1);
	if (reg == AC97_CD)
		val2 = snd_cs461x_codec_read(codec, AC97_CD);

	snd_cs461x_codec_write(private_data, reg, val);
	
	/*
	 *	Adjust power if the mixer is selected/deselected according
	 *	to the CD.
	 *
	 *	IF the CD is a valid input source (mixer or direct) AND
	 *		the CD is not muted THEN power is needed
	 *
	 *	We do two things. When record select changes the input to
	 *	add/remove the CD we adjust the power count if the CD is
	 *	unmuted.
	 *
	 *	When the CD mute changes we adjust the power level if the
	 *	CD was a valid input.
	 *
	 *      We also check for CD volume != 0, as the CD mute isn't
	 *      normally tweaked from userspace.
	 */
	 
	/* CD mute change ? */
	
	if (reg == AC97_CD) {
		/* Mute bit change ? */
		if ((val2^val)&0x8000 || ((val2 == 0x1f1f || val == 0x1f1f) && val2 != val)) {
			/* Mute on */
			if(val&0x8000 || val == 0x1f1f)
				codec->amplifier_ctrl(codec, -1);
			else /* Mute off power on */
				codec->amplifier_ctrl(codec, 1);
		}
	}

	codec->active_ctrl(codec, -1);
}

static unsigned short snd_cs461x_mixer_read(void *private_data, unsigned short reg)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	unsigned short val;
	codec->active_ctrl(codec, 1);
	val = snd_cs461x_codec_read(private_data, reg);
	codec->active_ctrl(codec, -1);
	return val;
}

static void snd_cs461x_mixer_free_ac97(void *private_data)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, );
	codec->mixer = NULL;
}

int snd_cs461x_mixer(cs461x_t * codec, int device, snd_pcm_t * pcm, snd_kmixer_t ** rmixer)
{
	ac97_t ac97;
	snd_kmixer_t *mixer;
	int err;

	*rmixer = NULL;
	memset(&ac97, 0, sizeof(ac97));
	ac97.write = snd_cs461x_mixer_write;
	ac97.read = snd_cs461x_mixer_read;
	ac97.private_data = codec;
	ac97.private_free = snd_cs461x_mixer_free_ac97;
	if ((err = snd_ac97_mixer(codec->card, device, &ac97, 1, &pcm->device, &mixer)) < 0)
		return err;
	codec->ac97 = snd_magic_cast(ac97_t, mixer->private_data, -ENXIO);
	*rmixer = codec->mixer = mixer;
	return 0;
}

/*
 *  RawMIDI interface
 */

static void snd_cs461x_midi_reset(cs461x_t *codec)
{
	snd_cs461x_pokeBA0(codec, BA0_MIDCR, MIDCR_MRST);
	udelay(100);
	snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
}

static int snd_cs461x_midi_input_open(snd_rawmidi_t * rmidi)
{
	unsigned long flags;
	cs461x_t *codec = snd_magic_cast(cs461x_t, rmidi->private_data, -ENXIO);

	codec->active_ctrl(codec, 1);
	spin_lock_irqsave(&codec->reg_lock, flags);
	codec->uartm |= CS461X_MODE_INPUT;
	codec->midcr |= MIDCR_RXE;
	if (!(codec->uartm & CS461X_MODE_OUTPUT)) {
		snd_cs461x_midi_reset(codec);
	} else {
		snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
	}
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return 0;
}

static int snd_cs461x_midi_input_close(snd_rawmidi_t * rmidi)
{
	unsigned long flags;
	cs461x_t *codec = snd_magic_cast(cs461x_t, rmidi->private_data, -ENXIO);

	spin_lock_irqsave(&codec->reg_lock, flags);
	codec->midcr &= ~(MIDCR_RXE | MIDCR_RIE);
	if (!(codec->uartm & CS461X_MODE_OUTPUT)) {
		snd_cs461x_midi_reset(codec);
	} else {
		snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
	}
	codec->uartm &= ~CS461X_MODE_INPUT;
	spin_unlock_irqrestore(&codec->reg_lock, flags);

	codec->active_ctrl(codec, -1);
	return 0;
}

static int snd_cs461x_midi_output_open(snd_rawmidi_t * rmidi)
{
	unsigned long flags;
	cs461x_t *codec = snd_magic_cast(cs461x_t, rmidi->private_data, -ENXIO);

	codec->active_ctrl(codec, 1);

	spin_lock_irqsave(&codec->reg_lock, flags);
	codec->uartm |= CS461X_MODE_OUTPUT;
	codec->midcr |= MIDCR_TXE;
	if (!(codec->uartm & CS461X_MODE_INPUT)) {
		snd_cs461x_midi_reset(codec);
	} else {
		snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
	}
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return 0;
}

static int snd_cs461x_midi_output_close(snd_rawmidi_t * rmidi)
{
	unsigned long flags;
	cs461x_t *codec = snd_magic_cast(cs461x_t, rmidi->private_data, -ENXIO);

	spin_lock_irqsave(&codec->reg_lock, flags);
	codec->midcr &= ~(MIDCR_TXE | MIDCR_TIE);
	if (!(codec->uartm & CS461X_MODE_INPUT)) {
		snd_cs461x_midi_reset(codec);
	} else {
		snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
	}
	codec->uartm &= ~CS461X_MODE_OUTPUT;
	spin_unlock_irqrestore(&codec->reg_lock, flags);

	codec->active_ctrl(codec, -1);
	return 0;
}

static void snd_cs461x_midi_input_trigger(snd_rawmidi_t * rmidi, int up)
{
	unsigned long flags;
	cs461x_t *codec = snd_magic_cast(cs461x_t, rmidi->private_data, );

	spin_lock_irqsave(&codec->reg_lock, flags);
	if (up) {
		if ((codec->midcr & MIDCR_RIE) == 0) {
			codec->midcr |= MIDCR_RIE;
			snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
		}
	} else {
		if (codec->midcr & MIDCR_RIE) {
			codec->midcr &= ~MIDCR_RIE;
			snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
		}
	}
	spin_unlock_irqrestore(&codec->reg_lock, flags);
}

static void snd_cs461x_midi_output_trigger(snd_rawmidi_t * rmidi, int up)
{
	unsigned long flags;
	cs461x_t *codec = snd_magic_cast(cs461x_t, rmidi->private_data, );
	unsigned char byte;

	spin_lock_irqsave(&codec->reg_lock, flags);
	if (up) {
		if ((codec->midcr & MIDCR_TIE) == 0) {
			codec->midcr |= MIDCR_TIE;
			/* fill UART FIFO buffer at first, and turn Tx interrupts only if necessary */
			while ((codec->midcr & MIDCR_TIE) &&
			       (snd_cs461x_peekBA0(codec, BA0_MIDSR) & MIDSR_TBF) == 0) {
				if (snd_rawmidi_transmit(rmidi, &byte, 1) != 1) {
					codec->midcr &= ~MIDCR_TIE;
				} else {
					snd_cs461x_pokeBA0(codec, BA0_MIDWP, byte);
				}
			}
			snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
		}
	} else {
		if (codec->midcr & MIDCR_TIE) {
			codec->midcr &= ~MIDCR_TIE;
			snd_cs461x_pokeBA0(codec, BA0_MIDCR, codec->midcr);
		}
	}
	spin_unlock_irqrestore(&codec->reg_lock, flags);
}

static struct snd_stru_rawmidi_channel_hw snd_cs461x_midi_output =
{
	open:           snd_cs461x_midi_output_open,
	close:          snd_cs461x_midi_output_close,
	trigger:        snd_cs461x_midi_output_trigger,
};

static struct snd_stru_rawmidi_channel_hw snd_cs461x_midi_input =
{
	open:           snd_cs461x_midi_input_open,
	close:          snd_cs461x_midi_input_close,
	trigger:        snd_cs461x_midi_input_trigger,
};

int snd_cs461x_midi(cs461x_t * codec, int device, snd_rawmidi_t **rrawmidi)
{
	snd_rawmidi_t *rmidi;
	int err;

	*rrawmidi = NULL;
	if ((err = snd_rawmidi_new(codec->card, "CS461X", device, &rmidi)) < 0)
		return err;
	strcpy(rmidi->name, "CS461X");
	memcpy(&rmidi->chn[SND_RAWMIDI_CHANNEL_OUTPUT].hw, &snd_cs461x_midi_output, sizeof(snd_cs461x_midi_output));
	memcpy(&rmidi->chn[SND_RAWMIDI_CHANNEL_INPUT].hw, &snd_cs461x_midi_input, sizeof(snd_cs461x_midi_input));
	rmidi->info_flags |= SND_RAWMIDI_INFO_OUTPUT | SND_RAWMIDI_INFO_INPUT | SND_RAWMIDI_INFO_DUPLEX;
	rmidi->private_data = codec;
	*rrawmidi = codec->rmidi = rmidi;
	return 0;
}

/*
 *  proc interface
 */

static long snd_cs461x_BA0_read(void *private_data, void *file_private_data,
				struct file *file, char *buf, long count)
{
	long size;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	
	size = count;
	if (file->f_pos + size > CS461X_BA0_SIZE)
		size = (long)CS461X_BA0_SIZE - file->f_pos;
	if (size > 0) {
		if (copy_to_user(buf, codec->ba0 + file->f_pos, size))
			return -EFAULT;
		file->f_pos += size;
		return size;
	}
	return 0;
}

static long snd_cs461x_BA1_data0_read(void *private_data, void *file_private_data,
				      struct file *file, char *buf, long count)
{
	long size;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	
	size = count;
	if (file->f_pos + size > CS461X_BA1_DATA0_SIZE)
		size = (long)CS461X_BA1_DATA0_SIZE - file->f_pos;
	if (size > 0) {
		if (copy_to_user(buf, codec->ba1.name.data0 + file->f_pos, size))
			return -EFAULT;
		file->f_pos += size;
		return size;
	}
	return 0;
}

static long snd_cs461x_BA1_data1_read(void *private_data, void *file_private_data,
				      struct file *file, char *buf, long count)
{
	long size;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	
	size = count;
	if (file->f_pos + size > CS461X_BA1_DATA1_SIZE)
		size = (long)CS461X_BA1_DATA1_SIZE - file->f_pos;
	if (size > 0) {
		if (copy_to_user(buf, codec->ba1.name.data1 + file->f_pos, size))
			return -EFAULT;
		file->f_pos += size;
		return size;
	}
	return 0;
}

static long snd_cs461x_BA1_prg_read(void *private_data, void *file_private_data,
				    struct file *file, char *buf, long count)
{
	long size;
	cs461x_t *codec = snd_magic_cast(cs461x_t, private_data, -ENXIO);
	
	size = count;
	if (file->f_pos + size > CS461X_BA1_PRG_SIZE)
		size = (long)CS461X_BA1_PRG_SIZE - file->f_pos;
	if (size > 0) {
		if (copy_to_user(buf, codec->ba1.name.pmem + file->f_pos, size))
			return -EFAULT;
		file->f_pos += size;
		return size;
	}
	return 0;
}

static int snd_cs461x_proc_init(snd_card_t * card, cs461x_t * codec)
{
	snd_info_entry_t *entry;
	
	entry = snd_info_create_entry(card, "cs461x_BA0");
	if (entry) {
		entry->type = SND_INFO_ENTRY_DATA;
		entry->private_data = codec;
		entry->t.data.read = snd_cs461x_BA0_read;
		entry->size = CS461X_BA0_SIZE;
		if (snd_info_register(entry) < 0) {
			snd_info_unregister(entry);
			entry = NULL;
		}
	}
	codec->entry_proc_BA0 = entry;
	entry = snd_info_create_entry(card, "cs461x_BA1_data0");
	if (entry) {
		entry->type = SND_INFO_ENTRY_DATA;
		entry->private_data = codec;
		entry->t.data.read = snd_cs461x_BA1_data0_read;
		entry->size = CS461X_BA1_DATA0_SIZE;
		if (snd_info_register(entry) < 0) {
			snd_info_unregister(entry);
			entry = NULL;
		}
	}
	codec->entry_proc_BA1_data0 = entry;
	entry = snd_info_create_entry(card, "cs461x_BA1_data1");
	if (entry) {
		entry->type = SND_INFO_ENTRY_DATA;
		entry->private_data = codec;
		entry->t.data.read = snd_cs461x_BA1_data1_read;
		entry->size = CS461X_BA1_DATA1_SIZE;
		if (snd_info_register(entry) < 0) {
			snd_info_unregister(entry);
			entry = NULL;
		}
	}
	codec->entry_proc_BA1_data1 = entry;
	entry = snd_info_create_entry(card, "cs461x_BA1_prg");
	if (entry) {
		entry->type = SND_INFO_ENTRY_DATA;
		entry->private_data = codec;
		entry->t.data.read = snd_cs461x_BA1_prg_read;
		entry->size = CS461X_BA1_PRG_SIZE;
		if (snd_info_register(entry) < 0) {
			snd_info_unregister(entry);
			entry = NULL;
		}
	}
	codec->entry_proc_BA1_prg = entry;
	return 0;
}

static int snd_cs461x_proc_done(cs461x_t * codec)
{
	if (codec->entry_proc_BA0)
		snd_info_unregister((snd_info_entry_t *) codec->entry_proc_BA0);
	if (codec->entry_proc_BA1_data0)
		snd_info_unregister((snd_info_entry_t *) codec->entry_proc_BA1_data0);
	if (codec->entry_proc_BA1_data1)
		snd_info_unregister((snd_info_entry_t *) codec->entry_proc_BA1_data1);
	if (codec->entry_proc_BA1_prg)
		snd_info_unregister((snd_info_entry_t *) codec->entry_proc_BA1_prg);
	return 0;
}

/*
 *  initialize chip
 */

static int snd_cs461x_chip_init(cs461x_t *codec, int busywait)
{
	unsigned int tmp;
	int timeout;

	/* 
	 *  First, blast the clock control register to zero so that the PLL starts
         *  out in a known state, and blast the master serial port control register
         *  to zero so that the serial ports also start out in a known state.
         */
        snd_cs461x_pokeBA0(codec, BA0_CLKCR1, 0);
        snd_cs461x_pokeBA0(codec, BA0_SERMC1, 0);

	/*
	 *  If we are in AC97 mode, then we must set the part to a host controlled
         *  AC-link.  Otherwise, we won't be able to bring up the link.
         */        
        snd_cs461x_pokeBA0(codec, BA0_SERACC, SERACC_HSP | SERACC_CODEC_TYPE_1_03);	/* 1.03 codec */
        /* snd_cs461x_pokeBA0(codec, BA0_SERACC, SERACC_HSP | SERACC_CODEC_TYPE_2_0); */ /* 2.00 codec */

        /*
         *  Drive the ARST# pin low for a minimum of 1uS (as defined in the AC97
         *  spec) and then drive it high.  This is done for non AC97 modes since
         *  there might be logic external to the CS461x that uses the ARST# line
         *  for a reset.
         */
        snd_cs461x_pokeBA0(codec, BA0_ACCTL, 0);
        udelay(500);
        snd_cs461x_pokeBA0(codec, BA0_ACCTL, ACCTL_RSTN);

	/*
	 *  The first thing we do here is to enable sync generation.  As soon
	 *  as we start receiving bit clock, we'll start producing the SYNC
	 *  signal.
	 */
	snd_cs461x_pokeBA0(codec, BA0_ACCTL, ACCTL_ESYN | ACCTL_RSTN);

	/*
	 *  Now wait for a short while to allow the AC97 part to start
	 *  generating bit clock (so we don't try to start the PLL without an
	 *  input clock).
	 */
	mdelay(10);

	/*
	 *  Set the serial port timing configuration, so that
	 *  the clock control circuit gets its clock from the correct place.
	 */
	snd_cs461x_pokeBA0(codec, BA0_SERMC1, SERMC1_PTC_AC97);

	/*
	 *  Write the selected clock control setup to the hardware.  Do not turn on
	 *  SWCE yet (if requested), so that the devices clocked by the output of
	 *  PLL are not clocked until the PLL is stable.
	 */
	snd_cs461x_pokeBA0(codec, BA0_PLLCC, PLLCC_LPF_1050_2780_KHZ | PLLCC_CDR_73_104_MHZ);
	snd_cs461x_pokeBA0(codec, BA0_PLLM, 0x3a);
	snd_cs461x_pokeBA0(codec, BA0_CLKCR2, CLKCR2_PDIVS_8);

	/*
	 *  Power up the PLL.
	 */
	snd_cs461x_pokeBA0(codec, BA0_CLKCR1, CLKCR1_PLLP);

	/*
         *  Wait until the PLL has stabilized.
	 */
	mdelay(100);

	/*
	 *  Turn on clocking of the core so that we can setup the serial ports.
	 */
	tmp = snd_cs461x_peekBA0(codec, BA0_CLKCR1) | CLKCR1_SWCE;
	snd_cs461x_pokeBA0(codec, BA0_CLKCR1, tmp);

	/*
	 *  Fill the serial port FIFOs with silence.
	 */
	snd_cs461x_clear_serial_FIFOs(codec);

	/*
	 *  Set the serial port FIFO pointer to the first sample in the FIFO.
	 */
	/* snd_cs461x_pokeBA0(codec, BA0_SERBSP, 0); */

	/*
	 *  Write the serial port configuration to the part.  The master
	 *  enable bit is not set until all other values have been written.
	 */
	snd_cs461x_pokeBA0(codec, BA0_SERC1, SERC1_SO1F_AC97 | SERC1_SO1EN);
	snd_cs461x_pokeBA0(codec, BA0_SERC2, SERC2_SI1F_AC97 | SERC1_SO1EN);
	snd_cs461x_pokeBA0(codec, BA0_SERMC1, SERMC1_PTC_AC97 | SERMC1_MSPE);

	mdelay(5);		/* Shouldnt be needed ?? */
	
	/*
	 * Wait for the codec ready signal from the AC97 codec.
	 */
	timeout = 150;
	while (timeout-- > 0) {
		/*
		 *  Read the AC97 status register to see if we've seen a CODEC READY
		 *  signal from the AC97 codec.
		 */
		if (snd_cs461x_peekBA0(codec, BA0_ACSTS) & ACSTS_CRDY)
			goto ok1;
		if (busywait)
			mdelay(10);
		else {
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout((HZ+99)/100);
		}
	}

	snd_printk("cs461x: create - never read codec ready from AC'97\n");
	snd_printk("cs461x: it is not probably bug, try to use CS4236 driver\n");
	return -EIO;
ok1:
	/*
	 *  Assert the vaid frame signal so that we can start sending commands
	 *  to the AC97 codec.
	 */
	snd_cs461x_pokeBA0(codec, BA0_ACCTL, ACCTL_VFRM | ACCTL_ESYN | ACCTL_RSTN);

	/*
	 *  Wait until we've sampled input slots 3 and 4 as valid, meaning that
	 *  the codec is pumping ADC data across the AC-link.
	 */
	timeout = 150;
	while (timeout-- > 0) {
		/*
		 *  Read the input slot valid register and see if input slots 3 and
		 *  4 are valid yet.
		 */
		if ((snd_cs461x_peekBA0(codec, BA0_ACISV) & (ACISV_ISV3 | ACISV_ISV4)) == (ACISV_ISV3 | ACISV_ISV4))
			goto ok2;
		if (busywait)
			mdelay(10);
		else {
			set_current_state(TASK_UNINTERRUPTIBLE);
			schedule_timeout((HZ+99)/100);
		}
	}

	snd_printk("cs461x: create - never read ISV3 & ISV4 from AC'97\n");
	return -EIO;
ok2:
	/*
	 *  Now, assert valid frame and the slot 3 and 4 valid bits.  This will
	 *  commense the transfer of digital audio data to the AC97 codec.
	 */
	snd_cs461x_pokeBA0(codec, BA0_ACOSV, ACOSV_SLV3 | ACOSV_SLV4);

	/*
	 *  Power down the DAC and ADC.  We will power them up (if) when we need
	 *  them.
	 */
	/* snd_cs461x_pokeBA0(codec, BA0_AC97_POWERDOWN, 0x300); */

	/*
	 *  Turn off the Processor by turning off the software clock enable flag in 
	 *  the clock control register.
	 */
	/* tmp = snd_cs461x_peekBA0(codec, BA0_CLKCR1) & ~CLKCR1_SWCE; */
	/* snd_cs461x_pokeBA0(codec, BA0_CLKCR1, tmp); */

	/*
         *  Reset the processor.
         */
	snd_cs461x_reset(codec);

	/*
         *  Download the image to the processor.
	 */
	if (snd_cs461x_download_image(codec) < 0) {
		snd_printk("cs461x: image download error\n");
		return -EIO;
	}

	/*
         *  Stop playback DMA.
	 */
	tmp = snd_cs461x_peek(codec, BA1_PCTL);
	codec->pctl = tmp & 0xffff0000;
	snd_cs461x_poke(codec, BA1_PCTL, tmp & 0x0000ffff);

	/*
         *  Stop capture DMA.
	 */
	tmp = snd_cs461x_peek(codec, BA1_CCTL);
	codec->cctl = tmp & 0x0000ffff;
	snd_cs461x_poke(codec, BA1_CCTL, tmp & 0xffff0000);

	mdelay(5);		/* Do we need this ?? */
	
	snd_cs461x_powerup_adc(codec);
	snd_cs461x_powerup_dac(codec);

	snd_cs461x_set_play_sample_rate(codec, 8000);
	snd_cs461x_set_capture_sample_rate(codec, 8000);

	snd_cs461x_proc_start(codec);

	/*
	 *  Enable interrupts on the part.
	 */
	snd_cs461x_pokeBA0(codec, BA0_HICR, HICR_IEV | HICR_CHGM);

	tmp = snd_cs461x_peek(codec, BA1_PFIE);
	tmp &= ~0x0000f03f;
	snd_cs461x_poke(codec, BA1_PFIE, tmp);	/* playback interrupt enable */

	tmp = snd_cs461x_peek(codec, BA1_CIE);
	tmp &= ~0x0000003f;
	tmp |=  0x00000001;
	snd_cs461x_poke(codec, BA1_CIE, tmp);	/* capture interrupt enable */	
	return 0;
}


/*
 * stop the h/w
 */
static void snd_cs461x_hw_stop(cs461x_t *codec)
{
	unsigned int tmp;

	tmp = snd_cs461x_peek(codec, BA1_PFIE);
	tmp &= ~0x0000f03f;
	tmp |=  0x00000010;
	snd_cs461x_poke(codec, BA1_PFIE, tmp);	/* playback interrupt disable */

	tmp = snd_cs461x_peek(codec, BA1_CIE);
	tmp &= ~0x0000003f;
	tmp |=  0x00000011;
	snd_cs461x_poke(codec, BA1_CIE, tmp);	/* capture interrupt disable */

	/*
         *  Stop playback DMA.
	 */
	tmp = snd_cs461x_peek(codec, BA1_PCTL);
	snd_cs461x_poke(codec, BA1_PCTL, tmp & 0x0000ffff);

	/*
         *  Stop capture DMA.
	 */
	tmp = snd_cs461x_peek(codec, BA1_CCTL);
	snd_cs461x_poke(codec, BA1_CCTL, tmp & 0xffff0000);

	/*
         *  Reset the processor.
         */
	snd_cs461x_reset(codec);

	snd_cs461x_proc_stop(codec);

	/*
	 *  Power down the DAC and ADC.  We will power them up (if) when we need
	 *  them.
	 */
	snd_cs461x_codec_write(codec, AC97_POWERDOWN, 0x300);

	/*
	 *  Power down the PLL.
	 */
	snd_cs461x_pokeBA0(codec, BA0_CLKCR1, 0);

	/*
	 *  Turn off the Processor by turning off the software clock enable flag in 
	 *  the clock control register.
	 */
	tmp = snd_cs461x_peekBA0(codec, BA0_CLKCR1) & ~CLKCR1_SWCE;
	snd_cs461x_pokeBA0(codec, BA0_CLKCR1, tmp);
}


/*
 *	AMP control - null AMP
 */
 
static void amp_none(cs461x_t *codec, int change)
{	
}

/*
 *	Crystal EAPD mode
 */
 
static void amp_voyetra(cs461x_t *codec, int change)
{
	/* Manage the EAPD bit on the Crystal 4297 
	   and the Analog AD1885 */
	   
	int old = codec->amplifier;
	
	codec->amplifier += change;
	if (codec->amplifier && !old) {
		/* Turn the EAPD amp on */
		snd_cs461x_codec_write(codec,  AC97_POWERDOWN, 
				       snd_cs461x_codec_read(codec, AC97_POWERDOWN) |
				       0x8000);
	} else if (old && !codec->amplifier) {
		/* Turn the EAPD amp off */
		snd_cs461x_codec_write(codec,  AC97_POWERDOWN, 
				       snd_cs461x_codec_read(codec, AC97_POWERDOWN) &
				       ~0x8000);
	}
}


/*
 *	Game Theatre XP card - EGPIO[2] is used to enable the external amp.
 */
 
static void amp_hercules(cs461x_t *chip, int change)
{
	int old = chip->amplifier;

	chip->amplifier += change;
	if (chip->amplifier && !old) {
		snd_cs461x_pokeBA0(chip, BA0_EGPIODR, 
				   EGPIODR_GPOE2);     /* enable EGPIO2 output */
		snd_cs461x_pokeBA0(chip, BA0_EGPIOPTR, 
				   EGPIOPTR_GPPT2);   /* open-drain on output */
	} else if (old && !chip->amplifier) {
		snd_cs461x_pokeBA0(chip, BA0_EGPIODR, 0); /* disable */
		snd_cs461x_pokeBA0(chip, BA0_EGPIOPTR, 0); /* disable */
	}
}


#if 0
/*
 *	Untested
 */
 
static void amp_voyetra_4294(cs461x_t *codec, int change)
{
	codec->amplifier+=change;

	if(codec->amplifier) {
		/* Switch the GPIO pins 7 and 8 to open drain */
		snd_cs461x_codec_write(codec, 0x4C,
				       snd_cs461x_codec_read(codec, 0x4C) & 0xFE7F);
		snd_cs461x_codec_write(codec, 0x4E,
				       snd_cs461x_codec_read(codec, 0x4E) | 0x0180);
		/* Now wake the AMP (this might be backwards) */
		snd_cs461x_codec_write(codec, 0x54,
				       snd_cs461x_codec_read(codec, 0x54) & ~0x0180);
	} else {
		snd_cs461x_codec_write(codec, 0x54,
				       snd_cs461x_codec_read(codec, 0x54) | 0x0180);
	}
}
#endif


/*
 * piix4 pci ids
 */
#ifndef PCI_VENDOR_ID_INTEL
#define PCI_VENDOR_ID_INTEL 0x8086
#endif /* PCI_VENDOR_ID_INTEL */

#ifndef PCI_DEVICE_ID_INTEL_82371AB_3
#define PCI_DEVICE_ID_INTEL_82371AB_3 0x7113
#endif /* PCI_DEVICE_ID_INTEL_82371AB_3 */

/*
 *	Handle the CLKRUN on a thinkpad. We must disable CLKRUN support
 *	whenever we need to beat on the chip.
 *
 *	The original idea and code for this hack comes from David Kaiser at
 *	Linuxcare. Perhaps one day Crystal will document their chips well
 *	enough to make them useful.
 */
 
static void clkrun_hack(cs461x_t *codec, int change)
{
	u16 control;
	int old;
	
	if (codec->acpi_dev == NULL)
		return;

	old = codec->amplifier;
	codec->amplifier += change;
	
	/* Read ACPI port */	
	control = inw(codec->acpi_port + 0x10);

	/* Flip CLKRUN off while running */
	if (! codec->amplifier && old)
		outw(control | 0x2000, codec->acpi_port + 0x10);
	else if (codec->amplifier && ! old)
		outw(control & ~0x2000, codec->acpi_port + 0x10);
}

	
/*
 * detect intel piix4
 */
static void clkrun_init(cs461x_t *codec)
{
	u8 pp;

	codec->acpi_dev = pci_find_device(PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82371AB_3, NULL);
	if (codec->acpi_dev == NULL)
		return;		/* Not a thinkpad thats for sure */

	/* Find the control port */		
	pci_read_config_byte(codec->acpi_dev, 0x41, &pp);
	codec->acpi_port = pp << 8;
}


/*
 * Card subid table
 */
 
struct cs_card_type
{
	u16 vendor;
	u16 id;
	char *name;
	void (*init)(cs461x_t *);
	void (*amp)(cs461x_t *, int);
	void (*active)(cs461x_t *, int);
};

static struct cs_card_type __initdata cards[]={
	{0x1489, 0x7001, "Genius Soundmaker 128 value", NULL, amp_none, NULL},
	{0x5053, 0x3357, "Voyetra", NULL, amp_voyetra, NULL},
	{0x1071, 0x6003, "Mitac MI6020/21", NULL, amp_voyetra, NULL},
	{0x14AF, 0x0050, "Hercules Game Theatre XP", NULL, amp_hercules, NULL},
	{0x1681, 0x0050, "Hercules Game Theatre XP", NULL, amp_hercules, NULL},
	{0x1681, 0x0051, "Hercules Game Theatre XP", NULL, amp_hercules, NULL},
	{0x1681, 0x0052, "Hercules Game Theatre XP", NULL, amp_hercules, NULL},
	{0x1681, 0x0053, "Hercules Game Theatre XP", NULL, amp_hercules, NULL},
	{0x1681, 0x0054, "Hercules Game Theatre XP", NULL, amp_hercules, NULL},
	/* Not sure if the 570 needs the clkrun hack */
	{PCI_VENDOR_ID_IBM, 0x0132, "Thinkpad 570", clkrun_init, NULL, clkrun_hack},
	{PCI_VENDOR_ID_IBM, 0x0153, "Thinkpad 600X/A20/T20", clkrun_init, NULL, clkrun_hack},
	{PCI_VENDOR_ID_IBM, 0x1010, "Thinkpad 600E (unsupported)", NULL, NULL, NULL},
	{0, 0, "Card without SSID set", NULL, NULL, NULL },
	{0, 0, NULL, NULL, NULL, NULL}
};


/*
 * APM support
 */
#ifdef CONFIG_PM
static int snd_cs461x_pm_callback(struct pm_dev *dev, pm_request_t rqst, void *data)
{
	cs461x_t *codec = snd_magic_cast(cs461x_t, dev->data, 0);
	int i, amp_saved;

	switch (rqst) {
	case PM_SUSPEND:
		if (codec->playback_subchn)
			snd_cs461x_playback_trigger(codec, codec->playback_subchn, SND_PCM_TRIGGER_STOP);
		
		if (codec->capture_subchn)
			snd_cs461x_capture_trigger(codec, codec->capture_subchn, SND_PCM_TRIGGER_STOP);

		snd_cs461x_hw_stop(codec);
		break;

	case PM_RESUME:
		amp_saved = codec->amplifier;
		codec->amplifier = 0;
		codec->active_ctrl(codec, 1); /* force to on */

		snd_cs461x_chip_init(codec, 1);

		snd_cs461x_codec_write(codec, AC97_MASTER, 0x8000);
		for (i = 0; i < 100; i++) {
			if (snd_cs461x_codec_read(codec, AC97_MASTER) == 0x8000)
				break;
			mdelay(10); /* Zzzz.. */
		}

		/* restore ac97 registers */
		snd_ac97_resume(codec->ac97);
#if 0
		for (i = 2; i <= 0x20; i += 2) {
			if (i == AC97_MASTER_TONE)
				continue;
			snd_cs461x_codec_write(codec, i, codec->ac97->regs[i]);
		}
#endif
		if (amp_saved)
			codec->amplifier_ctrl(codec, 1); /* try to turn on */
		if (codec->playback_subchn) {
			snd_cs461x_playback_prepare(codec, codec->playback_subchn);
			snd_cs461x_playback_trigger(codec, codec->playback_subchn, SND_PCM_TRIGGER_GO);
		}
		if (codec->capture_subchn) {
			snd_cs461x_capture_prepare(codec, codec->capture_subchn);
			snd_cs461x_capture_trigger(codec, codec->capture_subchn, SND_PCM_TRIGGER_GO);
		}
		if (! amp_saved) {
			codec->amplifier = 1;
			codec->active_ctrl(codec, -1);
		}
		codec->amplifier = amp_saved;
		break;
	}
	return 0;
}
#endif /* CONFIG_PM */


/*
 * create a cs461x card instance
 */
int snd_cs461x_create(snd_card_t * card,
		      struct pci_dev * pci,
		      snd_dma_t * dma1ptr,
		      snd_dma_t * dma2ptr,
		      snd_irq_t * irqptr,
		      int external_amp, int thinkpad,
		      cs461x_t ** rcodec)
{
	cs461x_t *codec;
	unsigned short cmdw;
	unsigned char cmdb;
	int err;
	struct cs_card_type *cp;
	u16 ss_card, ss_vendor;
	static snd_device_ops_t ops = {
		(snd_dev_free_t *)snd_cs461x_free,
		NULL,
		NULL
	};
	
	*rcodec = NULL;
	codec = snd_magic_kcalloc(cs461x_t, 0, GFP_KERNEL);
	if (codec == NULL)
		return -ENOMEM;
	spin_lock_init(&codec->reg_lock);
	codec->card = card;
	codec->pci = pci;
	codec->dma1ptr = dma1ptr;
	codec->dma2ptr = dma2ptr;
	codec->irqptr = irqptr;
	codec->ba0_addr = pci_resource_start(pci, 0);
	codec->ba1_addr = pci_resource_start(pci, 1);
	if (codec->ba0_addr == 0 || codec->ba0_addr == ~0 ||
	    codec->ba1_addr == 0 || codec->ba1_addr == ~0) {
	    	snd_printk("cs461x: wrong address(es) - ba0 = 0x%lx, ba1 = 0x%lx\n", codec->ba0_addr, codec->ba1_addr);
	    	snd_cs461x_free(codec);
	    	return -ENOMEM;
	}

	/* set up amp and clkrun hack */
	pci_read_config_word(pci, PCI_SUBSYSTEM_VENDOR_ID, &ss_vendor);
	pci_read_config_word(pci, PCI_SUBSYSTEM_ID, &ss_card);

	for (cp = &cards[0]; cp->name; cp++) {
		if (cp->vendor == ss_vendor && cp->id == ss_card) {
			snd_printk("cs461x: hack for %s enabled\n", cp->name);
			if (cp->init)
				cp->init(codec);
			codec->amplifier_ctrl = cp->amp;
			codec->active_ctrl = cp->active;
			break;
		}
	}

	if (external_amp) {
		snd_printk("cs461x: Crystal EAPD support forced on.\n");
		codec->amplifier_ctrl = amp_voyetra;
	}

	if (thinkpad) {
		snd_printk("cs461x: Activating CLKRUN hack for Thinkpad.\n");
		codec->active_ctrl = clkrun_hack;
		clkrun_init(codec);
	}
	
	if (codec->amplifier_ctrl == NULL)
		codec->amplifier_ctrl = amp_none;
	if (codec->active_ctrl == NULL)
		codec->active_ctrl = amp_none;

	codec->active_ctrl(codec, 1);

	codec->ba0 = ioremap(codec->ba0_addr, CS461X_BA0_SIZE);
	codec->ba1.name.data0 = ioremap(codec->ba1_addr + BA1_SP_DMEM0, CS461X_BA1_DATA0_SIZE);
	codec->ba1.name.data1 = ioremap(codec->ba1_addr + BA1_SP_DMEM1, CS461X_BA1_DATA1_SIZE);
	codec->ba1.name.pmem = ioremap(codec->ba1_addr + BA1_SP_PMEM, CS461X_BA1_PRG_SIZE);
	codec->ba1.name.reg = ioremap(codec->ba1_addr + BA1_SP_REG, CS461X_BA1_REG_SIZE);
	if (codec->ba0 == NULL || codec->ba1.name.data0 == NULL ||
	    codec->ba1.name.data1 == NULL || codec->ba1.name.pmem == NULL ||
	    codec->ba1.name.reg == NULL) {
		snd_cs461x_free(codec);
		return -ENOMEM;
	}

	pci_set_master(pci);
	pci_read_config_word(pci, PCI_COMMAND, &cmdw);
	if ((cmdw & PCI_COMMAND_MEMORY) != PCI_COMMAND_MEMORY) {
		cmdw |= PCI_COMMAND_MEMORY;
		pci_write_config_word(pci, PCI_COMMAND, cmdw);
	}
	pci_read_config_byte(pci, PCI_LATENCY_TIMER, &cmdb);
	if (cmdb < 32)
		cmdb = 32;
	pci_write_config_byte(pci, PCI_LATENCY_TIMER, cmdb);

	err = snd_cs461x_chip_init(codec, 0);
	if (err < 0) {
		snd_cs461x_free(codec);
		return err;
	}

	snd_cs461x_proc_init(card, codec);

#ifdef CONFIG_PM
	codec->pm_dev = pm_register(PM_PCI_DEV, 0, snd_cs461x_pm_callback);
	if (codec->pm_dev)
		codec->pm_dev->data = codec;
#endif

	if ((err = snd_device_new(card, SND_DEV_LOWLEVEL, codec, 0, &ops, NULL)) < 0) {
		snd_cs461x_free(codec);
		return err;
	}

	codec->active_ctrl(codec, -1);

	*rcodec = codec;
	return 0;
}

int snd_cs461x_free(cs461x_t * codec)
{
	snd_debug_check(codec == NULL, -EINVAL);

	if (codec->active_ctrl)
		codec->active_ctrl(codec, 1);

#ifdef CONFIG_PM
	if (codec->pm_dev)
		pm_unregister(codec->pm_dev);
#endif

	snd_cs461x_proc_done(codec);

	if (codec->ba0)
		snd_cs461x_hw_stop(codec);

	if (codec->ba0)
		iounmap(codec->ba0);
	if (codec->ba1.name.data0)
		iounmap(codec->ba1.name.data0);
	if (codec->ba1.name.data1)
		iounmap(codec->ba1.name.data1);
	if (codec->ba1.name.pmem)
		iounmap(codec->ba1.name.pmem);
	if (codec->ba1.name.reg)
		iounmap(codec->ba1.name.reg);

	if (codec->active_ctrl)
		codec->active_ctrl(codec, -codec->amplifier); /* force to off */
	if (codec->amplifier_ctrl)
		codec->amplifier_ctrl(codec, -codec->amplifier); /* force to off */

	snd_magic_kfree(codec);
	return 0;
}

EXPORT_SYMBOL(snd_cs461x_create);
EXPORT_SYMBOL(snd_cs461x_interrupt);
EXPORT_SYMBOL(snd_cs461x_pcm);
EXPORT_SYMBOL(snd_cs461x_mixer);
EXPORT_SYMBOL(snd_cs461x_midi);

/*
 *  INIT part
 */

static int __init alsa_cs461x_init(void)
{
	return 0;
}

static void __exit alsa_cs461x_exit(void)
{
}

module_init(alsa_cs461x_init)
module_exit(alsa_cs461x_exit)

MODULE_LICENSE("GPL");
