/*
 *  Copyright (c) 1999 by Uros Bizjak <uros@kss-loka.si>
 *                        Takashi Iwai <tiwai@suse.de>
 *
 *  SB16ASP/AWE32 CSP control (http://www.kss-loka.si/~uros/CSP.html)
 *
 *  CSP microcode loader:
 *   http://www.kss-loka.si/~uros/csptools-0.6.0pre1.tar.gz
 *
 *   This program is free software; you can redistribute it and/or modify 
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 */

#define SND_MAIN_OBJECT_FILE
#include "../../include/driver.h"
#include "../../include/info.h"
#include "../../include/sb16_csp.h"

#ifdef MODULE_PARM
MODULE_AUTHOR("Uros Bizjak <uros@kss-loka.si>");
MODULE_DESCRIPTION("ALSA driver for SB16 Creative Signal Processor\n\
classes: {sound}");
MODULE_LICENSE("GPL");
#endif /* MODULE_PARM */

#ifdef SND_LITTLE_ENDIAN
#define CSP_HDR_VALUE(a,b,c,d)	((a) | ((b)<<8) | ((c)<<16) | ((d)<<24))
#define LE_SHORT(v)		(v)
#define LE_INT(v)		(v)
#else
#define CSP_HDR_VALUE(a,b,c,d)	((d) | ((c)<<8) | ((b)<<16) | ((a)<<24))
#define LE_SHORT(v)		le16_to_cpu(v)
#define LE_INT(v)		le32_to_cpu(v)
#endif

#define RIFF_HEADER	CSP_HDR_VALUE('R', 'I', 'F', 'F')
#define CSP__HEADER	CSP_HDR_VALUE('C', 'S', 'P', ' ')
#define LIST_HEADER	CSP_HDR_VALUE('L', 'I', 'S', 'T')
#define FUNC_HEADER	CSP_HDR_VALUE('f', 'u', 'n', 'c')
#define CODE_HEADER	CSP_HDR_VALUE('c', 'o', 'd', 'e')
#define INIT_HEADER	CSP_HDR_VALUE('i', 'n', 'i', 't')
#define MAIN_HEADER	CSP_HDR_VALUE('m', 'a', 'i', 'n')

/*
 * RIFF data format
 */
typedef struct riff_header {
	__u32 name;
	__u32 len;
} riff_header_t;

typedef struct desc_header {
	riff_header_t info;
	__u16 func_nr;
	__u16 VOC_type;
	__u16 flags_play_rec;
	__u16 flags_16bit_8bit;
	__u16 flags_stereo_mono;
	__u16 flags_rates;
} desc_header_t;

/*
 * prototypes
 */
static void snd_sb_csp_free(void *private_data);
static int snd_sb_csp_open(snd_hwdep_t * hw, struct file *file);
static int snd_sb_csp_ioctl(snd_hwdep_t * hw, struct file *file, unsigned int cmd, unsigned long arg);
static int snd_sb_csp_release(snd_hwdep_t * hw, struct file *file);

static int csp_detect(sbdsp_t * codec, int *version);
static int set_codec_parameter(sbdsp_t * codec, unsigned char par, unsigned char val);
static int set_register(sbdsp_t * codec, unsigned char reg, unsigned char val);
static int read_register(sbdsp_t * codec, unsigned char reg);
static int set_mode_register(sbdsp_t * codec, unsigned char mode);
static int get_version(sbdsp_t * codec);

static int snd_sb_csp_riff_load(snd_sb_csp_t * p, snd_sb_csp_microcode_t * code);
static int snd_sb_csp_unload(snd_sb_csp_t * p);
static int snd_sb_csp_load(snd_sb_csp_t * p, const unsigned char *buf, int size, int load_flags);
static int snd_sb_csp_autoload(snd_sb_csp_t * p, int pcm_sfmt, int play_rec_mode);
static int snd_sb_csp_check_version(snd_sb_csp_t * p);

static int snd_sb_csp_use(snd_sb_csp_t * p);
static int snd_sb_csp_unuse(snd_sb_csp_t * p);
static int snd_sb_csp_start(snd_sb_csp_t * p, int sample_width, int channels);
static int snd_sb_csp_stop(snd_sb_csp_t * p);
static int snd_sb_csp_pause(snd_sb_csp_t * p);
static int snd_sb_csp_restart(snd_sb_csp_t * p);

static int snd_sb16mixer_qsound_control(snd_kmixer_element_t *element, int w_flag,
					struct snd_mixer_element_3d_effect1 *effect1);
static int snd_sb16_mixer_qsound_build(snd_sb_csp_t * p);
static int snd_sb16_mixer_qsound_destroy(snd_sb_csp_t * p);
static int snd_sb_csp_qsound_transfer(snd_sb_csp_t * p);

static int init_proc_entry(snd_sb_csp_t * p, int device);
static void delete_proc_entry(snd_sb_csp_t * p);
static void info_read(snd_info_buffer_t * buffer, void *private_data);

/*
 * Detect CSP chip and create a new instance
 */
int snd_sb_csp_new(sbdsp_t * codec, int device, snd_hwdep_t ** rhwdep)
{
	snd_sb_csp_t *p;
	int version, err;
	snd_hwdep_t *hw;

	*rhwdep = NULL;

	if (csp_detect(codec, &version))
		return -ENODEV;

	if ((err = snd_hwdep_new(codec->card, "SB16-CSP", device, &hw)) < 0)
		return err;

	if ((p = snd_magic_kcalloc(snd_sb_csp_t, 0, GFP_KERNEL)) == NULL) {
		snd_device_free(codec->card, hw);
		return -ENOMEM;
	}
	p->codec = codec;
	p->version = version;

	init_MUTEX(&p->access_mutex);
	sprintf(hw->name, "CSP v%d.%d", (version >> 4), (version & 0x0f));
	hw->type = SND_HWDEP_TYPE_SB16CSP;
	hw->private_data = p;
	hw->private_free = snd_sb_csp_free;

	/* operators - only write/ioctl */
	hw->ops.open = snd_sb_csp_open;
	hw->ops.ioctl = snd_sb_csp_ioctl;
	hw->ops.release = snd_sb_csp_release;

	/* create a proc entry */
	init_proc_entry(p, device);
	*rhwdep = hw;
	return 0;
}

/*
 * Call-back structure
 */
static snd_sb_csp_callback_t csp_callbacks =
{
	snd_sb_csp_use,
	snd_sb_csp_unuse,
	snd_sb_csp_autoload,
	snd_sb_csp_start,
	snd_sb_csp_stop,
	snd_sb_csp_qsound_transfer
};

/*
 * register callbacks
 */
void snd_sb_csp_register_callbacks(snd_sb_csp_callback_t ** callbacks_ptr)
{
	*callbacks_ptr = &csp_callbacks;
}

/*
 * free_private for hwdep instance
 */
static void snd_sb_csp_free(void *private_data)
{
	snd_sb_csp_t *p = snd_magic_cast(snd_sb_csp_t, private_data, );
	if (p) {
		if (p->running & SND_SB_CSP_ST_RUNNING)
			snd_sb_csp_stop(p);
		delete_proc_entry(p);
		snd_magic_kfree(p);
	}
}

/* ------------------------------ */

/*
 * open the device exclusively
 */
static int snd_sb_csp_open(snd_hwdep_t * hw, struct file *file)
{
	snd_sb_csp_t *p = snd_magic_cast(snd_sb_csp_t, hw->private_data, -ENXIO);
	return (snd_sb_csp_use(p));
}

/*
 * ioctl for hwdep device:
 */
static int snd_sb_csp_ioctl(snd_hwdep_t * hw, struct file *file, unsigned int cmd, unsigned long arg)
{
	snd_sb_csp_t *p = snd_magic_cast(snd_sb_csp_t, hw->private_data, -ENXIO);
	snd_sb_csp_info_t info;
	snd_sb_csp_start_t start_info;
	int err;

	snd_debug_check(p == NULL, -EINVAL);

	if (snd_sb_csp_check_version(p))
		return -ENODEV;

	switch (cmd) {
		/* get information */
	case SND_SB_CSP_IOCTL_INFO:
		memcpy(info.codec_name, p->codec_name, sizeof(p->codec_name));
		info.func_nr = p->func_nr;
		info.acc_format = p->acc_format;
		info.acc_channels = p->acc_channels;
		info.acc_width = p->acc_width;
		info.acc_rates = p->acc_rates;
		info.csp_mode = p->mode;
		info.run_channels = p->run_channels;
		info.run_width = p->run_width;
		info.version = p->version;
		info.state = p->running;
		err = copy_to_user((void *) arg, &info, sizeof(info));
		break;

		/* load CSP microcode */
	case SND_SB_CSP_IOCTL_LOAD_CODE:
		err = (p->running & SND_SB_CSP_ST_RUNNING ?
		       -EBUSY : snd_sb_csp_riff_load(p, (snd_sb_csp_microcode_t *) arg));
		break;
	case SND_SB_CSP_IOCTL_UNLOAD_CODE:
		err = (p->running & SND_SB_CSP_ST_RUNNING ?
		       -EBUSY : snd_sb_csp_unload(p));
		break;

		/* change CSP running state */
	case SND_SB_CSP_IOCTL_START:
		if (copy_from_user(&start_info, (void *) arg, sizeof(start_info)))
			err = -EFAULT;
		else
			err = snd_sb_csp_start(p, start_info.sample_width, start_info.channels);
		break;
	case SND_SB_CSP_IOCTL_STOP:
		err = snd_sb_csp_stop(p);
		break;
	case SND_SB_CSP_IOCTL_PAUSE:
		err = snd_sb_csp_pause(p);
		break;
	case SND_SB_CSP_IOCTL_RESTART:
		err = snd_sb_csp_restart(p);
		break;
	default:
		err = -EINVAL;
		break;
	}

	return err;
}

/*
 * close the device
 */
static int snd_sb_csp_release(snd_hwdep_t * hw, struct file *file)
{
	snd_sb_csp_t *p = snd_magic_cast(snd_sb_csp_t, hw->private_data, -ENXIO);
	return (snd_sb_csp_unuse(p));
}

/* ------------------------------ */

/*
 * acquire device
 */
static int snd_sb_csp_use(snd_sb_csp_t * p)
{
	down(&p->access_mutex);
	if (p->used) {
		up(&p->access_mutex);
		return -EAGAIN;
	}
	p->used++;
	up(&p->access_mutex);

	return 0;

}

/*
 * release device
 */
static int snd_sb_csp_unuse(snd_sb_csp_t * p)
{
	down(&p->access_mutex);
	p->used--;
	up(&p->access_mutex);

	return 0;
}

/*
 * load microcode via ioctl: 
 * code is user-space pointer
 */
static int snd_sb_csp_riff_load(snd_sb_csp_t * p, snd_sb_csp_microcode_t * mcode)
{
	snd_sb_csp_mc_header_t info;

	unsigned char *data_ptr, *data_end;
	unsigned short func_nr = 0;

	riff_header_t file_h, item_h, code_h;
	__u32 item_type;
	desc_header_t funcdesc_h;

	unsigned long flags;
	int err;

	if (copy_from_user(&info, mcode, sizeof(info)))
		return -EFAULT;
	data_ptr = mcode->data;

	if (copy_from_user(&file_h, data_ptr, sizeof(file_h)))
		return -EFAULT;
	if ((file_h.name != RIFF_HEADER) ||
	    (LE_INT(file_h.len) >= SND_SB_CSP_MAX_MICROCODE_FILE_SIZE - sizeof(file_h))) {
		snd_printd(__FUNCTION__ ": Invalid RIFF header\n");
		return -EINVAL;
	}
	data_ptr += sizeof(file_h);
	data_end = data_ptr + LE_INT(file_h.len);

	if (copy_from_user(&item_type, data_ptr, sizeof(item_type)))
		return -EFAULT;
	if (item_type != CSP__HEADER) {
		snd_printd(__FUNCTION__ ": Invalid RIFF file type\n");
		return -EINVAL;
	}
	data_ptr += sizeof (item_type);

	for (; data_ptr < data_end; data_ptr += LE_INT(item_h.len)) {
		if (copy_from_user(&item_h, data_ptr, sizeof(item_h)))
			return -EFAULT;
		data_ptr += sizeof(item_h);
		if (item_h.name != LIST_HEADER)
			continue;

		if (copy_from_user(&item_type, data_ptr, sizeof(item_type)))
			 return -EFAULT;
		switch (item_type) {
		case FUNC_HEADER:
			if (copy_from_user(&funcdesc_h, data_ptr + sizeof(item_type), sizeof(funcdesc_h)))
				return -EFAULT;
			func_nr = LE_SHORT(funcdesc_h.func_nr);
			break;
		case CODE_HEADER:
			if (func_nr != info.func_req)
				break;	/* not required function, try next */
			data_ptr += sizeof(item_type);

			/* destroy QSound mixer element */
			if (p->mode == SND_SB_CSP_MODE_QSOUND) {
				snd_sb16_mixer_qsound_destroy(p);
			}
			/* Clear all flags */
			p->running = 0;
			p->mode = 0;

			/* load microcode blocks */
			for (;;) {
				if (data_ptr >= data_end)
					return -EINVAL;
				if (copy_from_user(&code_h, data_ptr, sizeof(code_h)))
					return -EFAULT;

				/* init microcode blocks */
				if (code_h.name != INIT_HEADER)
					break;
				data_ptr += sizeof(code_h);
				err = snd_sb_csp_load(p, data_ptr, LE_INT(code_h.len),
						      SND_SB_CSP_LOAD_INITBLOCK | SND_SB_CSP_LOAD_FROMUSER);
				if (err)
					return err;
				data_ptr += LE_INT(code_h.len);
			}
			/* main microcode block */
			if (copy_from_user(&code_h, data_ptr, sizeof(code_h)))
				return -EFAULT;

			if (code_h.name != MAIN_HEADER) {
				snd_printd(__FUNCTION__ ": Missing 'main' microcode\n");
				return -EINVAL;
			}
			data_ptr += sizeof(code_h);
			err = snd_sb_csp_load(p, data_ptr, LE_INT(code_h.len),
					      SND_SB_CSP_LOAD_FROMUSER);
			if (err)
				return err;

			/* fill in codec header */
			strncpy(p->codec_name, info.codec_name, sizeof(p->codec_name) - 1);
			p->codec_name[sizeof(p->codec_name) - 1] = 0;
			p->func_nr = func_nr;
			p->mode = LE_SHORT(funcdesc_h.flags_play_rec);
			switch (LE_SHORT(funcdesc_h.VOC_type)) {
			case 0x0001:	/* QSound decoder */
				if (LE_SHORT(funcdesc_h.flags_play_rec) == SND_SB_CSP_MODE_DSP_WRITE) {
					if (snd_sb16_mixer_qsound_build(p) == 0)
						/* set QSound flag and clear all other mode flags */
						p->mode = SND_SB_CSP_MODE_QSOUND;
				}
				p->acc_format = 0;
				break;
			case 0x0006:	/* A Law codec */
				p->acc_format = SND_PCM_FMT_A_LAW;
				break;
			case 0x0007:	/* Mu Law codec */
				p->acc_format = SND_PCM_FMT_MU_LAW;
				break;
			case 0x0011:	/* what Creative thinks is IMA ADPCM codec */
			case 0x0200:	/* Creative ADPCM codec */
				p->acc_format = SND_PCM_FMT_IMA_ADPCM;
				break;
			case    201:	/* Text 2 Speech decoder */
				/* TODO: Text2Speech handling routines */
				p->acc_format = 0;
				break;
			case 0x0202:	/* Fast Speech 8 codec */
			case 0x0203:	/* Fast Speech 10 codec */
				p->acc_format = SND_PCM_FMT_SPECIAL;
				break;
			default:	/* other codecs are unsupported */
				p->acc_format = p->acc_width = p->acc_rates = 0;
				p->mode = 0;
				snd_printd(__FUNCTION__ ": Unsupported CSP codec type: 0x%04x\n",
					   LE_SHORT(funcdesc_h.VOC_type));
				return -EINVAL;
			}
			p->acc_channels = LE_SHORT(funcdesc_h.flags_stereo_mono);
			p->acc_width = LE_SHORT(funcdesc_h.flags_16bit_8bit);
			p->acc_rates = LE_SHORT(funcdesc_h.flags_rates);

			/* Decouple CSP from IRQ and DMAREQ lines */
			spin_lock_irqsave(&p->codec->reg_lock, flags);
			set_mode_register(p->codec, 0xfc);
			set_mode_register(p->codec, 0x00);
			spin_unlock_irqrestore(&p->codec->reg_lock, flags);

			/* finished loading successfully */
			p->running = SND_SB_CSP_ST_LOADED;	/* set LOADED flag */
			return 0;
		}
	}
	snd_printd(__FUNCTION__ ": Function #%d not found\n", info.func_req);
	return -EINVAL;
}

/*
 * unload CSP microcode
 */
static int snd_sb_csp_unload(snd_sb_csp_t * p)
{
	if (p->running & SND_SB_CSP_ST_RUNNING)
		return -EBUSY;
	if (!(p->running & SND_SB_CSP_ST_LOADED))
		return -ENXIO;

	/* clear supported formats */
	p->acc_format = 0;
	p->acc_channels = p->acc_width = p->acc_rates = 0;
	/* destroy QSound mixer element */
	if (p->mode == SND_SB_CSP_MODE_QSOUND) {
		snd_sb16_mixer_qsound_destroy(p);
	}
	/* clear all flags */
	p->running = 0;
	p->mode = 0;
	return 0;
}

/*
 * send command sequence to DSP
 */
static inline int command_seq(sbdsp_t * codec, const unsigned char *seq, int size)
{
	int i;
	for (i = 0; i < size; i++) {
		if (!snd_sb16dsp_command(codec, seq[i]))
			return -EIO;
	}
	return 0;
}

/*
 * set CSP codec parameter
 */
static int set_codec_parameter(sbdsp_t * codec, unsigned char par, unsigned char val)
{
	unsigned char dsp_cmd[3];

	dsp_cmd[0] = 0x05;	/* CSP set codec parameter */
	dsp_cmd[1] = val;	/* Parameter value */
	dsp_cmd[2] = par;	/* Parameter */
	command_seq(codec, dsp_cmd, 3);
	snd_sb16dsp_command(codec, 0x03);	/* DSP read? */
	if (snd_sb16dsp_get_byte(codec) != par)
		return -EIO;
	return 0;
}

/*
 * set CSP register
 */
static int set_register(sbdsp_t * codec, unsigned char reg, unsigned char val)
{
	unsigned char dsp_cmd[3];

	dsp_cmd[0] = 0x0e;	/* CSP set register */
	dsp_cmd[1] = reg;	/* CSP Register */
	dsp_cmd[2] = val;	/* value */
	return command_seq(codec, dsp_cmd, 3);
}

/*
 * read CSP register
 * return < 0 -> error
 */
static int read_register(sbdsp_t * codec, unsigned char reg)
{
	unsigned char dsp_cmd[2];

	dsp_cmd[0] = 0x0f;	/* CSP read register */
	dsp_cmd[1] = reg;	/* CSP Register */
	command_seq(codec, dsp_cmd, 2);
	return snd_sb16dsp_get_byte(codec);	/* Read DSP value */
}

/*
 * set CSP mode register
 */
static int set_mode_register(sbdsp_t * codec, unsigned char mode)
{
	unsigned char dsp_cmd[2];

	dsp_cmd[0] = 0x04;	/* CSP set mode register */
	dsp_cmd[1] = mode;	/* mode */
	return command_seq(codec, dsp_cmd, 2);
}

/*
 * Detect CSP
 * return 0 if CSP exists.
 */
static int csp_detect(sbdsp_t * codec, int *version)
{
	unsigned char csp_test1, csp_test2;
	unsigned long flags;
	int result = -ENODEV;

	spin_lock_irqsave(&codec->reg_lock, flags);

	set_codec_parameter(codec, 0x00, 0x00);
	set_mode_register(codec, 0xfc);		/* 0xfc = ?? */

	csp_test1 = read_register(codec, 0x83);
	set_register(codec, 0x83, ~csp_test1);
	csp_test2 = read_register(codec, 0x83);
	if (csp_test2 != (csp_test1 ^ 0xff))
		goto __fail;

	set_register(codec, 0x83, csp_test1);
	csp_test2 = read_register(codec, 0x83);
	if (csp_test2 != csp_test1)
		goto __fail;

	set_mode_register(codec, 0x00);		/* 0x00 = ? */

	*version = get_version(codec);
	snd_sb16dsp_reset(codec);	/* reset DSP after getversion! */
	if (*version >= 0x10 && *version <= 0x1f)
		result = 0;		/* valid version id */

      __fail:
	spin_unlock_irqrestore(&codec->reg_lock, flags);
	return result;
}

/*
 * get CSP version number
 */
static int get_version(sbdsp_t * codec)
{
	unsigned char dsp_cmd[2];

	dsp_cmd[0] = 0x08;	/* SB_DSP_!something! */
	dsp_cmd[1] = 0x03;	/* get chip version id? */
	command_seq(codec, dsp_cmd, 2);

	return (snd_sb16dsp_get_byte(codec));
}

/*
 * check if the CSP version is valid
 */
static int snd_sb_csp_check_version(snd_sb_csp_t * p)
{
	if (p->version < 0x10 || p->version > 0x1f) {
		snd_printd(__FUNCTION__ ": Invalid CSP version: 0x%x\n", p->version);
		return 1;
	}
	return 0;
}

/*
 * download microcode to CSP (microcode should have one "main" block).
 */
static int snd_sb_csp_load(snd_sb_csp_t * p, const unsigned char *buf, int size, int load_flags)
{
	int status, i;
	int err;
	int result = -EIO;
	unsigned long flags;

	spin_lock_irqsave(&p->codec->reg_lock, flags);
	snd_sb16dsp_command(p->codec, 0x01);	/* CSP download command */
	if (snd_sb16dsp_get_byte(p->codec)) {
		snd_printd(__FUNCTION__ ": Download command failed\n");
		goto __fail;
	}
	/* Send CSP low byte (size - 1) */
	snd_sb16dsp_command(p->codec, (unsigned char)(size - 1));
	/* Send high byte */
	snd_sb16dsp_command(p->codec, (unsigned char)((size - 1) >> 8));
	/* send microcode sequence */
	if (load_flags & SND_SB_CSP_LOAD_FROMUSER) {
		/* copy microcode from user space */
		unsigned char *kbuf, *_kbuf;
		_kbuf = kbuf = snd_kmalloc (size, GFP_KERNEL);
		if (copy_from_user(kbuf, buf, size)) {
			result = -EFAULT;
			snd_kfree (_kbuf);
			goto __fail;
		}
		while (size--) {
			if (!snd_sb16dsp_command(p->codec, *kbuf++)) {
				snd_kfree (_kbuf);
				goto __fail;
			}
		}
		snd_kfree (_kbuf);
	} else {
		/* load from kernel space */
		while (size--) {
			if (!snd_sb16dsp_command(p->codec, *buf++))
				goto __fail;
		}
	}
	if (snd_sb16dsp_get_byte(p->codec))
		goto __fail;

	if (load_flags & SND_SB_CSP_LOAD_INITBLOCK) {
		/* some codecs (FastSpeech) take some time to initialize */
		status = 0;  /* to make gcc happy */
		for (i = 10; i; i--) {
			snd_sb16dsp_command(p->codec, 0x03);
			status = snd_sb16dsp_get_byte(p->codec);
			if (status == 0x55)
				break;
			udelay (10);
		}
		if (status != 0x55) {
			snd_printd(__FUNCTION__ ": Microcode initialization failed\n");
			goto __fail;
		}
	} else {
		/*
		 * Read mixer register SB_DSP4_DMASETUP after loading 'main' code.
		 * Start CSP chip if no 16bit DMA channel is set - some kind
		 * of autorun or perhaps a bugfix?
		 */
		spin_lock(&p->codec->mixer.lock);
		status = snd_sb16mixer_read(&p->codec->mixer, SB_DSP4_DMASETUP);
		spin_unlock(&p->codec->mixer.lock);
		if (!(status & (SB_DMASETUP_DMA7 | SB_DMASETUP_DMA6 | SB_DMASETUP_DMA5))) {
			err = (set_codec_parameter(p->codec, 0xaa, 0x00) ||
			       set_codec_parameter(p->codec, 0xff, 0x00));
			snd_sb16dsp_reset(p->codec);		/* really! */
			if (err)
				goto __fail;
			set_mode_register(p->codec, 0xc0);	/* c0 = STOP */
			set_mode_register(p->codec, 0x70);	/* 70 = RUN */
		}
	}
	result = 0;

      __fail:
	spin_unlock_irqrestore(&p->codec->reg_lock, flags);
	return result;
}

#include "sb16_csp_codecs.h"

/*
 * autoload hardware codec if necessary
 * return 0 if CSP is loaded and ready to run (p->running != 0)
 */
static int snd_sb_csp_autoload(snd_sb_csp_t * p, int pcm_sfmt, int play_rec_mode)
{
	unsigned long flags;
	int err = 0;

	/* if CSP is running or manually loaded then exit */
	if (p->running & (SND_SB_CSP_ST_RUNNING | SND_SB_CSP_ST_LOADED)) 
		return -EBUSY;

	/* autoload microcode only if requested hardware codec is not already loaded */
	if (((1 << pcm_sfmt) & p->acc_format) && (play_rec_mode & p->mode)) {
		p->running = SND_SB_CSP_ST_AUTO;
	} else {
		switch (pcm_sfmt) {
		case SND_PCM_SFMT_MU_LAW:
			err = snd_sb_csp_load(p, &mulaw_main[0], sizeof(mulaw_main), 0);
			p->acc_format = SND_PCM_FMT_MU_LAW;
			p->mode = SND_SB_CSP_MODE_DSP_READ | SND_SB_CSP_MODE_DSP_WRITE;
			break;
		case SND_PCM_SFMT_A_LAW:
			err = snd_sb_csp_load(p, &alaw_main[0], sizeof(alaw_main), 0);
			p->acc_format = SND_PCM_FMT_A_LAW;
			p->mode = SND_SB_CSP_MODE_DSP_READ | SND_SB_CSP_MODE_DSP_WRITE;
			break;
		case SND_PCM_SFMT_IMA_ADPCM:
			err = snd_sb_csp_load(p, &ima_adpcm_init[0], sizeof(ima_adpcm_init),
					      SND_SB_CSP_LOAD_INITBLOCK);
			if (err)
				break;
			if (play_rec_mode == SND_SB_CSP_MODE_DSP_WRITE) {
				err = snd_sb_csp_load(p, &ima_adpcm_playback[0],
						      sizeof(ima_adpcm_playback), 0);
				p->mode = SND_SB_CSP_MODE_DSP_WRITE;
			} else {
				err = snd_sb_csp_load(p, &ima_adpcm_capture[0],
						      sizeof(ima_adpcm_capture), 0);
				p->mode = SND_SB_CSP_MODE_DSP_READ;
			}
			p->acc_format = SND_PCM_FMT_IMA_ADPCM;
			break;				  
		default:
			/* Decouple CSP from IRQ and DMAREQ lines */
			if (p->running & SND_SB_CSP_ST_AUTO) {
				spin_lock_irqsave(&p->codec->reg_lock, flags);
				set_mode_register(p->codec, 0xfc);
				set_mode_register(p->codec, 0x00);
				spin_unlock_irqrestore(&p->codec->reg_lock, flags);
				p->running = 0;			/* clear autoloaded flag */
			}
			return -EINVAL;
		}
		if (err) {
			p->acc_format = 0;
			p->acc_channels = p->acc_width = p->acc_rates = 0;

			p->running = 0;				/* clear autoloaded flag */
			p->mode = 0;
			return (err);
		} else {
			p->running = SND_SB_CSP_ST_AUTO;	/* set autoloaded flag */
			p->acc_width = SND_SB_CSP_SAMPLE_16BIT;	/* only 16 bit data */
			p->acc_channels = SND_SB_CSP_MONO | SND_SB_CSP_STEREO;
			p->acc_rates = SND_SB_CSP_RATE_ALL;	/* HW codecs accept all rates */
		}   

	}
	return (p->running & SND_SB_CSP_ST_AUTO) ? 0 : -ENXIO;
}

/*
 * start CSP
 */
static int snd_sb_csp_start(snd_sb_csp_t * p, int sample_width, int channels)
{
	unsigned char s_type;	/* sample type */
	unsigned char mixL, mixR;
	int result = -EIO;
	unsigned long flags;

	if (!(p->running & (SND_SB_CSP_ST_LOADED | SND_SB_CSP_ST_AUTO))) {
		snd_printd(__FUNCTION__ ": Microcode not loaded\n");
		return -ENXIO;
	}
	if (p->running & SND_SB_CSP_ST_RUNNING) {
		snd_printd(__FUNCTION__ ": CSP already running\n");
		return -EBUSY;
	}
	if (!(sample_width & p->acc_width)) {
		snd_printd(__FUNCTION__ ": Unsupported PCM sample width\n");
		return -EINVAL;
	}
	if (!(channels & p->acc_channels)) {
		snd_printd(__FUNCTION__ ": Invalid number of channels\n");
		return -EINVAL;
	}

	/* Mute PCM volume */
	spin_lock_irqsave(&p->codec->mixer.lock, flags);
	mixL = snd_sb16mixer_read(&p->codec->mixer, SB_DSP4_PCM_DEV);
	mixR = snd_sb16mixer_read(&p->codec->mixer, SB_DSP4_PCM_DEV + 1);
	snd_sb16mixer_write(&p->codec->mixer, SB_DSP4_PCM_DEV, mixL & 0x7);
	snd_sb16mixer_write(&p->codec->mixer, SB_DSP4_PCM_DEV + 1, mixR & 0x7);

	spin_lock(&p->codec->reg_lock);
	set_mode_register(p->codec, 0xc0);	/* c0 = STOP */
	set_mode_register(p->codec, 0x70);	/* 70 = RUN */

	s_type = 0x00;
	if (channels == SND_SB_CSP_MONO)
		s_type = 0x11;	/* 000n 000n    (n = 1 if mono) */
	if (sample_width == SND_SB_CSP_SAMPLE_8BIT)
		s_type |= 0x22;	/* 00dX 00dX    (d = 1 if 8 bit samples) */

	if (set_codec_parameter(p->codec, 0x81, s_type)) {
		snd_printd(__FUNCTION__ ": Set sample type command failed\n");
		goto __fail;
	}
	if (set_codec_parameter(p->codec, 0x80, 0x00)) {
		snd_printd(__FUNCTION__ ": Codec start command failed\n");
		goto __fail;
	}
	p->run_width = sample_width;
	p->run_channels = channels;

	p->running |= SND_SB_CSP_ST_RUNNING;

	if (p->mode & SND_SB_CSP_MODE_QSOUND) {
		set_codec_parameter(p->codec, 0xe0, 0x01);
		/* enable QSound decoder */
		set_codec_parameter(p->codec, 0x00, 0xff);
		set_codec_parameter(p->codec, 0x01, 0xff);
		p->running |= SND_SB_CSP_ST_QSOUND;
		/* set QSound startup value */
		snd_sb_csp_qsound_transfer(p);
	}
	result = 0;

      __fail:
	spin_unlock(&p->codec->reg_lock);

	/* restore PCM volume */
	snd_sb16mixer_write(&p->codec->mixer, SB_DSP4_PCM_DEV, mixL);
	snd_sb16mixer_write(&p->codec->mixer, SB_DSP4_PCM_DEV + 1, mixR);
	spin_unlock_irqrestore(&p->codec->mixer.lock, flags);

	return result;
}

/*
 * stop CSP
 */
static int snd_sb_csp_stop(snd_sb_csp_t * p)
{
	int result;
	unsigned char mixL, mixR;
	unsigned long flags;

	if (!(p->running & SND_SB_CSP_ST_RUNNING))
		return 0;

	/* Mute PCM volume */
	spin_lock_irqsave(&p->codec->mixer.lock, flags);
	mixL = snd_sb16mixer_read(&p->codec->mixer, SB_DSP4_PCM_DEV);
	mixR = snd_sb16mixer_read(&p->codec->mixer, SB_DSP4_PCM_DEV + 1);
	snd_sb16mixer_write(&p->codec->mixer, SB_DSP4_PCM_DEV, mixL & 0x7);
	snd_sb16mixer_write(&p->codec->mixer, SB_DSP4_PCM_DEV + 1, mixR & 0x7);

	spin_lock(&p->codec->reg_lock);
	if (p->running & SND_SB_CSP_ST_QSOUND) {
		set_codec_parameter(p->codec, 0xe0, 0x01);
		/* disable QSound decoder */
		set_codec_parameter(p->codec, 0x00, 0x00);
		set_codec_parameter(p->codec, 0x01, 0x00);

		p->running &= ~SND_SB_CSP_ST_QSOUND;
	}
	result = set_mode_register(p->codec, 0xc0);	/* c0 = STOP */
	spin_unlock(&p->codec->reg_lock);

	/* restore PCM volume */
	snd_sb16mixer_write(&p->codec->mixer, SB_DSP4_PCM_DEV, mixL);
	snd_sb16mixer_write(&p->codec->mixer, SB_DSP4_PCM_DEV + 1, mixR);
	spin_unlock_irqrestore(&p->codec->mixer.lock, flags);

	if (!(result))
		p->running &= ~(SND_SB_CSP_ST_PAUSED | SND_SB_CSP_ST_RUNNING);
	return result;
}

/*
 * pause CSP codec and hold DMA transfer
 */
static int snd_sb_csp_pause(snd_sb_csp_t * p)
{
	int result;
	unsigned long flags;

	if (!(p->running & SND_SB_CSP_ST_RUNNING))
		return -EBUSY;

	spin_lock_irqsave(&p->codec->reg_lock, flags);
	result = set_codec_parameter(p->codec, 0x80, 0xff);
	spin_unlock_irqrestore(&p->codec->reg_lock, flags);
	if (!(result))
		p->running |= SND_SB_CSP_ST_PAUSED;

	return result;
}

/*
 * restart CSP codec and resume DMA transfer
 */
static int snd_sb_csp_restart(snd_sb_csp_t * p)
{
	int result;
	unsigned long flags;

	if (!(p->running & SND_SB_CSP_ST_PAUSED))
		return -EBUSY;

	spin_lock_irqsave(&p->codec->reg_lock, flags);
	result = set_codec_parameter(p->codec, 0x80, 0x00);
	spin_unlock_irqrestore(&p->codec->reg_lock, flags);
	if (!(result))
		p->running &= ~SND_SB_CSP_ST_PAUSED;

	return result;
}

/* ------------------------------ */

/*
 * QSound mixer element for PCM
 */
static int snd_sb16mixer_qsound_control(snd_kmixer_element_t *element, int w_flag,
					struct snd_mixer_element_3d_effect1 *effect1)
{
	snd_sb_csp_t *p = snd_magic_cast(snd_sb_csp_t, element->private_data, -ENXIO);
	unsigned int flags;
	int change = 0;

	/* lock access to data structure (interrupt driven qsound_transfer) */
	spin_lock_irqsave(&p->q_lock, flags);
	if (!w_flag) {
		effect1->sw = p->q_enabled;
		effect1->space = p->qpos_left;
	} else {
		if (p->q_enabled != effect1->sw) {
			p->q_enabled = effect1->sw;
			change = 1;
		}
		if (p->qpos_left != effect1->space) {
			p->qpos_left = effect1->space;
			p->qpos_right = effect1->space;
			p->qpos_changed = 1;
			change = 1;
		}
	}
	spin_unlock_irqrestore(&p->q_lock, flags);
	return change;
}

static int snd_sb16_mixer_qsound_build(snd_sb_csp_t * p)
{
	snd_kmixer_t *mixer = p->mixer;
	static struct snd_mixer_element_3d_effect1_info eff1_info;

	memset(&eff1_info, 0, sizeof(eff1_info));
	eff1_info.effect = SND_MIXER_EFF1_SPACE | SND_MIXER_EFF1_SW;
	eff1_info.max_space = SND_SB_CSP_QSOUND_MAX_RIGHT;

	snd_debug_check(mixer == NULL, -EINVAL);

	spin_lock_init(&p->q_lock);

	snd_mixer_lock(mixer, 0);
	/* build 3d effect1 group */
	if ((p->qsound_group = snd_mixer_lib_group(mixer, SND_MIXER_GRP_EFFECT_3D, 0)) == NULL)
		goto __error;
	/* QSound element */
	if ((p->qsound_element = snd_mixer_lib_3d_effect1(mixer, "QSound Position", 0, &eff1_info, snd_sb16mixer_qsound_control, p)) == NULL)
		goto __error;
	if (snd_mixer_element_route_add(mixer, p->qsound_element, p->mixer_dest))
		goto __error;
	if (snd_mixer_group_element_add(mixer, p->qsound_group, p->qsound_element) < 0)
		goto __error;

	snd_mixer_lock(mixer, 1);
	return 0;

     __error:
	snd_mixer_lock(mixer, 1);
	snd_sb16_mixer_qsound_destroy(p);
	return -ENOMEM;
}

static int snd_sb16_mixer_qsound_destroy(snd_sb_csp_t * p)
{
	snd_kmixer_t *mixer = p->mixer;
	unsigned long flags;

	snd_debug_check((mixer == NULL) || (p->qsound_group == NULL), -EINVAL);
	snd_mixer_lock(mixer, 0);
	/* remove all the routes, elements */
	/* remove qsound element  */
	if (p->qsound_element) {
		snd_mixer_group_element_remove(mixer, p->qsound_group, p->qsound_element);
		snd_mixer_element_remove(mixer, p->qsound_element);
		p->qsound_element = NULL;
	}
	/* remove 3d effect group */
	if (p->qsound_group) {
		snd_mixer_group_remove(mixer, p->qsound_group);
		p->qsound_group = NULL;
	}
	snd_mixer_lock(mixer, 1);

	/* cancel pending transfer of QSound parameters */
	spin_lock_irqsave (&p->q_lock, flags);
	p->qpos_changed = 0;
	spin_unlock_irqrestore (&p->q_lock, flags);

	return 0;
}

/*
 * Transfer qsound parameters to CSP,
 * function should be called from interrupt routine
 */
static int snd_sb_csp_qsound_transfer(snd_sb_csp_t * p)
{
	int err = -ENXIO;

	spin_lock(&p->q_lock);
	if (p->running & SND_SB_CSP_ST_QSOUND) {
		set_codec_parameter(p->codec, 0xe0, 0x01);
		/* left channel */
		set_codec_parameter(p->codec, 0x00, p->qpos_left);
		set_codec_parameter(p->codec, 0x02, 0x00);
		/* right channel */
		set_codec_parameter(p->codec, 0x00, p->qpos_right);
		set_codec_parameter(p->codec, 0x03, 0x00);
		err = 0;
	}
	p->qpos_changed = 0;
	spin_unlock(&p->q_lock);
	return err;
}

/* ------------------------------ */

/*
 * proc interface
 */
static int init_proc_entry(snd_sb_csp_t * p, int device)
{
	char name[16];
	snd_info_entry_t *entry;
	sprintf(name, "cspD%d", device);
	entry = p->proc = snd_info_create_entry(p->codec->card, name);
	if (entry) {
		entry->type = SND_INFO_ENTRY_TEXT;
		entry->t.text.read_size = 256;
		entry->t.text.read = info_read;
		entry->private_data = p;
		if (snd_info_register(entry) < 0) {
			snd_info_free_entry(entry);
			p->proc = NULL;
		}
	}
	return 0;
}

static void delete_proc_entry(snd_sb_csp_t * p)
{
	if (p->proc) {
		snd_info_unregister(p->proc);
		p->proc = NULL;
	}
}

static void info_read(snd_info_buffer_t * buffer, void *private_data)
{
	snd_sb_csp_t *p = snd_magic_cast(snd_sb_csp_t, private_data, );

	snd_iprintf(buffer, "Creative Signal Processor [v%d.%d]\n", (p->version >> 4), (p->version & 0x0f));
	snd_iprintf(buffer, "State: %cx%c%c%c\n", ((p->running & SND_SB_CSP_ST_QSOUND) ? 'Q' : '-'),
		    ((p->running & SND_SB_CSP_ST_PAUSED) ? 'P' : '-'),
		    ((p->running & SND_SB_CSP_ST_RUNNING) ? 'R' : '-'),
		    ((p->running & SND_SB_CSP_ST_LOADED) ? 'L' : '-'));
	if (p->running & SND_SB_CSP_ST_LOADED) {
		snd_iprintf(buffer, "Codec: %s [func #%d]\n", p->codec_name, p->func_nr);
		snd_iprintf(buffer, "Sample rates: ");
		if (p->acc_rates == SND_SB_CSP_RATE_ALL) {
			snd_iprintf(buffer, "All\n");
		} else {
			snd_iprintf(buffer, "%s%s%s%s\n",
				    ((p->acc_rates & SND_SB_CSP_RATE_8000) ? "8000Hz " : ""),
				    ((p->acc_rates & SND_SB_CSP_RATE_11025) ? "11025Hz " : ""),
				    ((p->acc_rates & SND_SB_CSP_RATE_22050) ? "22050Hz " : ""),
				    ((p->acc_rates & SND_SB_CSP_RATE_44100) ? "44100Hz" : ""));
		}
		if (p->mode == SND_SB_CSP_MODE_QSOUND) {
			snd_iprintf(buffer, "QSound decoder %sabled\n",
				    p->q_enabled ? "en" : "dis");
		} else {
			snd_iprintf(buffer, "PCM format ID: 0x%x (%s/%s) [%s/%s] [%s/%s]\n",
				    p->acc_format,
				    ((p->acc_width & SND_SB_CSP_SAMPLE_16BIT) ? "16bit" : "-"),
				    ((p->acc_width & SND_SB_CSP_SAMPLE_8BIT) ? "8bit" : "-"),
				    ((p->acc_channels & SND_SB_CSP_MONO) ? "mono" : "-"),
				    ((p->acc_channels & SND_SB_CSP_STEREO) ? "stereo" : "-"),
				    ((p->mode & SND_SB_CSP_MODE_DSP_WRITE) ? "playback" : "-"),
				    ((p->mode & SND_SB_CSP_MODE_DSP_READ) ? "capture" : "-"));
		}
	}
	if (p->running & SND_SB_CSP_ST_AUTO) {
		snd_iprintf(buffer, "Autoloaded Mu-Law, A-Law or Ima-ADPCM hardware codec\n");
	}
	if (p->running & SND_SB_CSP_ST_RUNNING) {
		snd_iprintf(buffer, "Processing %dbit %s PCM samples\n",
			    ((p->run_width & SND_SB_CSP_SAMPLE_16BIT) ? 16 : 8),
			    ((p->run_channels & SND_SB_CSP_MONO) ? "mono" : "stereo"));
	}
	if (p->running & SND_SB_CSP_ST_QSOUND) {
		snd_iprintf(buffer, "Qsound position: left = 0x%x, right = 0x%x\n",
			    p->qpos_left, p->qpos_right);
	}
}

/* */

EXPORT_SYMBOL(snd_sb_csp_new);
EXPORT_SYMBOL(snd_sb_csp_register_callbacks);

/*
 * INIT part
 */

static int __init alsa_sb16_csp_init(void)
{
	return 0;
}

static void __exit alsa_sb16_csp_exit(void)
{
}

module_init(alsa_sb16_csp_init)
module_exit(alsa_sb16_csp_exit)
