/*
 * Cyclades PC300 synchronous serial card driver for Linux
 *
 * Copyright (C) 2000-2001 Krzysztof Halasa <khc@pm.waw.pl>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * For information see http://hq.pm.waw.pl/hdlc/
 *
 * Current status:
 *    - this is work in progress
 *
 * Sources of information:
 *    Hitachi HD64572 SCA-II User's Manual
 *    Cyclades PC300 Linux driver
 */

#include <linux/config.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/slab.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/fcntl.h>
#include <linux/in.h>
#include <linux/string.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/netdevice.h>
#include <linux/hdlc.h>
#include <linux/pci.h>
#include <asm/io.h>
#include <asm/delay.h>

#include "hd64572.h"


static const char* version = "Cyclades PC300 driver version 1.05";
static const char* devname = "PC300";

#define	PC300_PLX_WIN	0x80    /* PLX control window size (128b) */
#define	PC300_RAMSIZE	0x80000 /* RAM window size (512Kb) */
#define	PC300_SCASIZE	0x400   /* SCA window size (1Kb) */
#define ALL_PAGES_ALWAYS_MAPPED

#define PC300_OSC_CLOCK	24576000
#define PC300_PCI_CLOCK	33000000


#ifdef CONFIG_PC300_PCI_CLOCK
#define CLOCK_BASE PC300_PCI_CLOCK
#else
#define CLOCK_BASE PC300_OSC_CLOCK
#endif


/* Masks to access the init_ctrl PLX register */
#define	PC300_CLKSEL_MASK		(0x00000004UL)
#define	PC300_CHMEDIA_MASK(port)	(0x00000020UL << (port * 3))
#define	PC300_CTYPE_MASK		(0x00000800UL)


/* Control Constant Definitions */
#define	PC300_RSV	0x01
#define	PC300_X21	0x02
#define	PC300_TE	0x03


/*
 *      PLX PCI9050-1 local configuration and shared runtime registers.
 *	This structure can be used to access 9050 registers (memory mapped).
 */
typedef struct {
	u32 loc_addr_range[4];	/* 00-0Ch : Local Address Ranges */
	u32 loc_rom_range;	/* 10h : Local ROM Range */
	u32 loc_addr_base[4];	/* 14-20h : Local Address Base Addrs */
	u32 loc_rom_base;	/* 24h : Local ROM Base */
	u32 loc_bus_descr[4];	/* 28-34h : Local Bus Descriptors */
	u32 rom_bus_descr;	/* 38h : ROM Bus Descriptor */
	u32 cs_base[4];		/* 3C-48h : Chip Select Base Addrs */
	u32 intr_ctrl_stat;	/* 4Ch : Interrupt Control/Status */
	u32 init_ctrl;		/* 50h : EEPROM ctrl, Init Ctrl, etc */
}plx9050;



typedef struct port_s {
	hdlc_device hdlc;	/* HDLC device struct - must be first */
	struct card_s *card;
	spinlock_t lock;	/* TX lock */
	sync_serial_settings settings;
	unsigned short encoding;
	unsigned short parity;
	unsigned int iface;
	u8 rxs, txs, tmc;	/* SCA registers */
	u8 phy_node;		/* physical port # - 0 or 1 */
	u8 rxin;		/* rx ring buffer 'in' pointer */
	u8 txin;		/* tx ring buffer 'in' and 'last' pointers */
	u8 txlast;
	u8 rxpart;		/* partial frame received, next frame invalid*/
}port_t;



typedef struct card_s {
	int type;		/* RSV, X21, etc. */
	u32 ramphys;		/* dual-port memory base (physical) */
	u8* rambase;		/* dual-port memory base (virtual) */
	u32 ramsize;		/* dual-port memory size */
	u32 scaphys;		/* dual-port memory base (physical) */
	u8* scabase;		/* dual-port memory base (virtual) */
	u32 plxphys;		/* PLX registers memory base (physical) */
	plx9050* plxbase;	/* PLX registers memory base (virtual) */
	u16 buff_offset;	/* offset of first buffer of first channel */
	u8 irq;			/* interrupt request level */
	u8 ring_buffers;	/* number of buffers in a ring */
	int clock;		/* Board clock */

	port_t ports[2];
}card_t;


#define sca_in(reg, card)	     readb(card->scabase + (reg))
#define sca_out(value, reg, card)    writeb(value, card->scabase + (reg))
#define sca_inw(reg, card)	     readw(card->scabase + (reg))
#define sca_outw(value, reg, card)   writew(value, card->scabase + (reg))
#define sca_inl(reg, card)	     readl(card->scabase + (reg))
#define sca_outl(value, reg, card)   writel(value, card->scabase + (reg))

#define port_to_card(port)	     (port->card)
#define log_node(port)		     (port->phy_node)
#define phy_node(port)		     (port->phy_node)
#define winsize(card)		     (PC300_RAMSIZE)
#define win0base(card)		     (card->rambase)
#define winbase(card)      	     (card->rambase)
#define get_port(card, port)	     (&card->ports[port])

#include "hd6457x.c"



static int pc300_set_iface(port_t *port)
{
	card_t *card = port->card;
	u32* init_ctrl = &card->plxbase->init_ctrl;
	u8 msci = get_msci(port);
	u8 rxs = port->rxs & CLK_BRG_MASK;
	u8 txs = port->txs & CLK_BRG_MASK;

	sca_out(EXS_TES1, (phy_node(port) ? MSCI1_OFFSET : MSCI0_OFFSET) + EXS,
		port_to_card(port));
	switch(port->settings.clock_type) {
	case CLOCK_EXT:
		rxs |= CLK_LINE; /* RXC input */
		txs |= CLK_PIN_OUT | CLK_LINE; /* TXC input */
		break;

	case CLOCK_INT:
		rxs |= CLK_BRG; /* BRG output */
		txs |= CLK_PIN_OUT | CLK_TX_RXCLK; /* RX clock */
		break;

	case CLOCK_TXINT:
		rxs |= CLK_LINE; /* RXC input */
		txs |= CLK_PIN_OUT | CLK_BRG; /* BRG output */
		break;

	case CLOCK_TXFROMRX:
		rxs |= CLK_LINE; /* RXC input */
		txs |= CLK_PIN_OUT | CLK_TX_RXCLK; /* RX clock */
		break;

	default:
		return -EINVAL;
	}

	port->rxs = rxs;
	port->txs = txs;
	sca_out(rxs, msci + RXS, card);
	sca_out(txs, msci + TXS, card);
	sca_set_port(port);

	switch(port->iface) {
	case IF_IFACE_X21:
		if (port->card->type != PC300_X21)
			return -EINVAL;
		break;

	case IF_IFACE_V35:
		if (port->card->type != PC300_RSV)
			return -EINVAL;
		writel(readl(init_ctrl) | PC300_CHMEDIA_MASK(port->phy_node),
		       init_ctrl);
		break;

	case IF_IFACE_V24:
		if (port->card->type != PC300_RSV)
			return -EINVAL;
		writel(readl(init_ctrl) & ~PC300_CHMEDIA_MASK(port->phy_node),
		       init_ctrl);
		break;

	default:
		return -EINVAL;
	}
	return 0;
}



static int pc300_open(struct net_device *dev)
{
	hdlc_device *hdlc = dev_to_hdlc(dev);
	port_t *port = hdlc_to_port(hdlc);

	int result = hdlc_open(hdlc);
	if (result)
		return result;

	MOD_INC_USE_COUNT;
	sca_open(hdlc);
	return pc300_set_iface(port);
}



static int pc300_close(struct net_device *dev)
{
	hdlc_device *hdlc = dev_to_hdlc(dev);
	sca_close(hdlc);
	hdlc_close(hdlc);
	MOD_DEC_USE_COUNT;
	return 0;
}



static int pc300_ioctl(struct net_device *dev, struct ifreq *ifr, int cmd)
{
	hdlc_device *hdlc = dev_to_hdlc(dev);
	port_t *port = hdlc_to_port(hdlc);
	const size_t size = sizeof(sync_serial_settings);

#ifdef CONFIG_HDLC_DEBUG_RINGS
	if (cmd == SIOCDEVPRIVATE) {
		sca_dump_rings(hdlc);
		return 0;
	}
#endif
	if (cmd != SIOCDEVICE)
		return hdlc_ioctl(dev, ifr, cmd);

	switch(ifr->ifr_settings.type) {
	case IF_GET_IFACE:
		ifr->ifr_settings.type = port->iface;
		if (ifr->ifr_settings.data_length == 0)
			return 0; /* return interface type only */
		if (ifr->ifr_settings.data_length < size)
			return -ENOMEM;	/* buffer too small */
		if (copy_to_user(ifr->ifr_settings.data,
				 &port->settings, size))
			return -EFAULT;
		ifr->ifr_settings.data_length = size;
		return 0;

	case IF_IFACE_V35:
	case IF_IFACE_V24:
	case IF_IFACE_X21:
		if(!capable(CAP_NET_ADMIN))
			return -EPERM;

		if (ifr->ifr_settings.data_length != size)
			return -ENOMEM;	/* incorrect data length */

		if (copy_from_user(&port->settings,
				   ifr->ifr_settings.data, size))
			return -EFAULT;
		/* FIXME - put sanity checks here */
		port->iface = ifr->ifr_settings.type;
		return pc300_set_iface(port);

	default:
		return hdlc_ioctl(dev, ifr, cmd);
	}
}



static u32 detect_ram(card_t *card)
{
	u32 i;
	u8* rambase = card->rambase;

	/* Let's find out how much RAM is present on this board */
	for (i = 0; i < PC300_RAMSIZE ; i++) {
		writeb((u8)i, rambase + i);
		if (readb(rambase + i) != (u8)i)
			break;
	}

	return i;
}



static void plx_init(card_t *card)
{
	plx9050 *plx_ctl = (plx9050 *)card->plxbase;

	/* Reset PLX */
	writel(readl(&plx_ctl->init_ctrl) | 0x40000000, &plx_ctl->init_ctrl);
	udelay(100L);
	writel(readl(&plx_ctl->init_ctrl) & ~0x40000000, &plx_ctl->init_ctrl);

	/* Reload Config. Registers from EEPROM */
	writel(readl(&plx_ctl->init_ctrl) | 0x20000000, &plx_ctl->init_ctrl);
	udelay(100L);
	writel(readl(&plx_ctl->init_ctrl) & ~0x20000000, &plx_ctl->init_ctrl);
}



static void pc300_pci_remove_one(struct pci_dev *pdev)
{
	int i;
        card_t *card = pci_get_drvdata(pdev);

        if (!card)
                BUG();

	for(i = 0; i < 2; i++)
		if (card->ports[i].card)
			unregister_hdlc_device(&card->ports[i].hdlc);
  
	if (card->irq)
		free_irq(card->irq, card);
	iounmap(card->rambase);
	iounmap(card->scabase);
	iounmap(card->plxbase);
        pci_release_regions(pdev);
	kfree(card);
        pci_set_drvdata(pdev, NULL);
}



static int __devinit pc300_pci_init_one(struct pci_dev *pdev,
					const struct pci_device_id *ent)
{
	card_t *card;
	u8 rev_id;
	int i;

#ifndef MODULE
        static int printed_version;
        if (!printed_version++)
                printk(KERN_INFO "%s\n", version);
#endif

	i = pci_enable_device(pdev);
	if (i)
		return i;

        i = pci_request_regions(pdev, "PC300");
	if (i)
                return i;

	card = kmalloc(sizeof(card_t), GFP_KERNEL);
	if (card == NULL) {
		printk(KERN_ERR "pc300: unable to allocate memory\n");
		return -ENOBUFS;
	}
	memset(card, 0, sizeof(card_t));

	pci_read_config_byte(pdev, PCI_REVISION_ID, &rev_id);

	card->plxphys = pci_resource_start(pdev,0) & PCI_BASE_ADDRESS_MEM_MASK;
	card->plxbase = ioremap(card->plxphys, PC300_PLX_WIN);

	card->scaphys = pci_resource_start(pdev,2) & PCI_BASE_ADDRESS_MEM_MASK;
	card->scabase = ioremap(card->scaphys, PC300_SCASIZE);

	card->ramphys = pci_resource_start(pdev,3) & PCI_BASE_ADDRESS_MEM_MASK;
	card->rambase = ioremap(card->ramphys, PC300_RAMSIZE);

#ifdef CONFIG_PC300_DEBUG_PCI
	printk(KERN_DEBUG "pc300: bus=0x0%x, pci_id=0x%x, rev_id=%d IRQ%d "
	       "RAM=0x%x PLX=0x%x SCA=0x%x relocated "
	       "RAM=0x%p PLX=0x%p SCA=0x%p\n",
	       pdev->bus->number, pdev->devfn, rev_id, pdev->irq,
	       card->ramphys, card->plxphys, card->scaphys,
	       card->rambase, card->plxbase, card->scabase);
#endif
	card->ramsize = detect_ram(card);
  
	/* Enable interrupts on the PCI bridge */
	plx_init(card);
	writew(readw(&card->plxbase->intr_ctrl_stat) | 0x0040,
	       &card->plxbase->intr_ctrl_stat);
  
#ifdef CONFIG_PC300_PCI_CLOCK
	/* Set board clock to PCI clock */
	writel(readl(&card->plxbase->init_ctrl) | 0x00000004UL,
	       &card->plxbase->init_ctrl);
	card->clock = PC300_PCI_CLOCK;
#else
	/* Set board clock to internal oscillator clock */
	writel(readl(&card->plxbase->init_ctrl) &~0x00000004UL,
	       &card->plxbase->init_ctrl);
	card->clock = PC300_OSC_CLOCK;
#endif
  
	/* Set Global SCA-II registers */
	if(readl(&card->plxbase->init_ctrl) & PC300_CTYPE_MASK)
		card->type = PC300_X21;
	else
		card->type = PC300_RSV;

	pci_set_drvdata(pdev, card);

	/* Allocate IRQ */
	if(request_irq(pdev->irq, sca_intr, SA_SHIRQ, devname, card)) {
		printk(KERN_WARNING "pc300: could not allocate IRQ%d.\n",
		       pdev->irq);
		pc300_pci_remove_one(pdev);
		return -EBUSY;
	}
	card->irq = pdev->irq;

	/* 4 rings required for 2 ports, 2 rings for one port */
	card->ring_buffers = card->ramsize /
		(4 * (sizeof(pkt_desc) + HDLC_MAX_MRU));

	card->buff_offset = 4 * (sizeof(pkt_desc)) * card->ring_buffers;

	printk(KERN_INFO "pc300: PC300/%s %u KB RAM at 0x%x, IRQ%u, "
	       "using %u packets rings\n",
	       card->type == PC300_X21 ? "X21" : "RSV", card->ramsize / 1024,
	       card->ramphys, card->irq, card->ring_buffers);

	sca_init(card, 0);

	sca_out(sca_in(PCR, card) | PCR_COTE, PCR, card);
	sca_out(0x10, BTCR, card);
  
	for(i = 0; i < 2; i++) {
		port_t *port = &card->ports[i];
		struct net_device *dev = hdlc_to_dev(&port->hdlc);
		port->phy_node = i;

		spin_lock_init(&port->lock);
		dev->irq = card->irq;
		dev->mem_start = card->ramphys;
		dev->mem_end = card->ramphys + card->ramsize - 1;
		dev->tx_queue_len = 50;
		dev->do_ioctl = pc300_ioctl;
		dev->open = pc300_open;
		dev->stop = pc300_close;
		port->hdlc.attach = sca_attach;
		port->hdlc.xmit = sca_xmit;
		port->settings.clock_type = CLOCK_EXT;
		if (card->type == PC300_X21)
			port->iface = IF_IFACE_X21;
		else
			port->iface = IF_IFACE_V35;

		if(register_hdlc_device(&port->hdlc)) {
			printk("pc300: unable to register hdlc device\n");
			pc300_pci_remove_one(pdev);
			return -ENOBUFS;
		}
		port->card = card;
		sca_init_sync_port(port);	/* Set up SCA memory */

		printk(KERN_INFO "%s: PC300 node %d\n",
		       hdlc_to_name(&port->hdlc), port->phy_node);
	}
	return 0;
}



static struct pci_device_id pc300_pci_tbl[] __devinitdata = {
        { PCI_VENDOR_ID_CYCLADES, PCI_DEVICE_ID_PC300_RX_1, PCI_ANY_ID,
	  PCI_ANY_ID, 0, 0, 0 },
        { PCI_VENDOR_ID_CYCLADES, PCI_DEVICE_ID_PC300_RX_2, PCI_ANY_ID,
	  PCI_ANY_ID, 0, 0, 0 },
        { PCI_VENDOR_ID_CYCLADES, PCI_DEVICE_ID_PC300_TE_1, PCI_ANY_ID,
	  PCI_ANY_ID, 0, 0, 0 },
        { PCI_VENDOR_ID_CYCLADES, PCI_DEVICE_ID_PC300_TE_2, PCI_ANY_ID,
	  PCI_ANY_ID, 0, 0, 0 },
        { 0, }
};


static struct pci_driver pc300_pci_driver = {
        name:           "PC300",
        id_table:       pc300_pci_tbl,
        probe:          pc300_pci_init_one,
        remove:         pc300_pci_remove_one,
};


static int __init pc300_init_module(void)
{
#ifdef MODULE
        printk(KERN_INFO "%s (SCA-%s)\n", version, sca_version);
#endif
	return pci_module_init(&pc300_pci_driver);
}



static void __init pc300_cleanup_module(void)
{
	pci_unregister_driver(&pc300_pci_driver);
}

MODULE_AUTHOR("Krzysztof Halasa <khc@pm.waw.pl>");
MODULE_DESCRIPTION("Cyclades PC300 serial port driver");
MODULE_LICENSE("GPL");
MODULE_DEVICE_TABLE(pci, pc300_pci_tbl);
EXPORT_NO_SYMBOLS;
module_init(pc300_init_module);
module_exit(pc300_cleanup_module);
