/*-*- linux-c -*-
 *  linux/drivers/video/i810fb.c -- Intel 810 frame buffer device
 *
 *      Copyright (C) 2001 Antonino Daplas
 *      All Rights Reserved      
 *
 *      Contributors:
 *         Michael Vogt <mvogt@acm.org> - added support for Intel 815 chipsets and enabling the power-on
 *                                        state of external VGA connectors for secondary displays
 *
 *	The code framework is a modification of vfb.c by Geert Uytterhoeven.
 *      DotClock and PLL calculations are partly based on i810_driver.c in xfree86 v4.0.3 by
 *              Precision Insight.
 *      Watermark calculation and tables are based on i810_wmark.c in xfre86 v4.0.3 by
 *              Precision Insight.  Slight modifications only to allow for integer operations
 *              instead of floating point.
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

#include <linux/module.h>
#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/string.h>
#include <linux/mm.h>
#include <linux/tty.h>
#include <linux/version.h>

#if LINUX_VERSION_CODE <= KERNEL_VERSION(2,4,14)
#include <linux/malloc.h>
#else
#include <linux/slab.h>
#endif

#include <linux/vmalloc.h>
#include <linux/fb.h>
#include <linux/init.h>
#include <linux/pci.h>
#include <linux/pci_ids.h>
#include <linux/resource.h>
#include <linux/selection.h>
#include <linux/console.h>
#include <linux/vt_kern.h>
#include <linux/unistd.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include <asm/mtrr.h>
#include <asm/page.h>

#include <video/fbcon.h>
#include <video/fbcon-cfb8.h>
#include <video/fbcon-cfb16.h>
#include <video/fbcon-cfb24.h>

#include "i810fb.h"

#define VERSION_MAJOR            0
#define VERSION_MINOR            0
#define VERSION_TEENIE           10

#define arraysize(x)	(sizeof(x)/sizeof(*(x)))
#define I810_PAGESIZE   4096

#ifndef PCI_DEVICE_ID_INTEL_82810_MC4
  #define PCI_DEVICE_ID_INTEL_82810_MC4           0x7124
#endif
#ifndef PCI_DEVICE_ID_INTEL_82810_IG4
  #define PCI_DEVICE_ID_INTEL_82810_IG4           0x7125
#endif

/* mvo: intel i815 */
#ifndef PCI_DEVICE_ID_INTEL_82815_100
  #define PCI_DEVICE_ID_INTEL_82815_100           0x1102
#endif
#ifndef PCI_DEVICE_ID_INTEL_82815_NOAGP
  #define PCI_DEVICE_ID_INTEL_82815_NOAGP         0x1112
#endif
#ifndef PCI_DEVICE_ID_INTEL_82815_FULL_CTRL
  #define PCI_DEVICE_ID_INTEL_82815_FULL_CTRL     0x1130
#endif 
#ifndef PCI_DEVICE_ID_INTEL_82815_FULL_BRG
  #define PCI_DEVICE_ID_INTEL_82815_FULL_BRG      0x1131
#endif 
#ifndef PCI_DEVICE_ID_INTEL_82815_FULL
  #define PCI_DEVICE_ID_INTEL_82815_FULL          0x1132
#endif 

#define PCI_I810_MISCC              0x72
#define MMIO_SIZE                   512*1024
#define GTT_SIZE                    64*1024
#define RINGBUFFER_SIZE             128*1024
#define CURSOR_SIZE                 4096 
#define OFF                         0
#define ON                          1

/* Masks (AND ops) and OR's */
#define FB_START_MASK               0x3f << (32 - 6)
#define MMIO_ADDR_MASK              0x1FFF << (32 - 13)
#define FREQ_MASK                   0x1EF
#define SCR_OFF                      0x20
#define DRAM_ON                     0x08            
#define DRAM_OFF                    0xE7
#define PG_ENABLE_MASK              0x01

/* defines for restoring registers partially */
#define ADDR_MAP_MASK               0x07 << 5
#define DISP_CTRL                   ~0
#define PIXCONF_0                   0x64 << 8
#define PIXCONF_2                   0xF3 << 24
#define PIXCONF_1                   0xF0 << 16
#define MN_MASK                     0x3FF03FF
#define P_OR                        0x7 << 4                    
#define DAC_BIT                     1 << 16
#define INTERLACE_BIT               1 << 7
#define IER_MASK                    3 << 13
#define IMR_MASK                    3 << 13

/* Power Management */
#define DPMS_MASK                   0xF0000
#define POWERON                     0x00000
#define STANDBY                     0x20000
#define SUSPEND                     0x80000
#define POWERDOWN                   0xA0000

#define EMR_MASK                    ~0x3F
#define FW_BLC_MASK                 ~(0x3F | (7 << 8) | (0x3F << 12) |  (7 << 20))
#define RBUFFER_START_MASK          0xFFFFF000
#define RBUFFER_SIZE_MASK           0x001FF000
#define RBUFFER_HEAD_MASK           0x001FFFFC
#define RBUFFER_TAIL_MASK           0x001FFFF8

/* Mode */
#define VFREQ_MAX                   85
#define REF_FREQ                    24000000
#define TARGET_N_MAX                30

#ifdef CONFIG_FB_I810_NONSTD
#define VFREQ_MIN                    50
#else
#define VFREQ_MIN                    60
#endif 

/* Cursor */
#define CURSOR_ENABLE_MASK          0x1000             
#define CURSOR_MODE_64_TRANS        4
#define CURSOR_MODE_64_XOR	    5
#define CURSOR_MODE_64_3C	    6	
#define COORD_INACTIVE              0
#define COORD_ACTIVE                1 << 4
#define EXTENDED_PALETTE	    1
  
/* AGP */
#define AGP_NORMAL                  0
#define AGP_CACHE	            1
#define AGP_PHYSICAL                2


struct pgtbl_record {
	u8 flag;
};

struct ringbuffer {
	u32 tail;
	u32 head;
	u32 start;
	u32 size;
};

struct state_registers {
	struct ringbuffer lring_state;	
	u32 dclk_1d, dclk_2d, dclk_0ds;
	u32 pixconf, fw_blc, pgtbl_ctl;
	u16 bltcntl, hwstam, ier, iir, imr;
	u8 cr00, cr01, cr02, cr03, cr04;
	u8 cr05, cr06, cr07, cr08, cr09;
	u8 cr10, cr11, cr12, cr13, cr14;
	u8 cr15, cr16, cr17, cr80, gr10;
	u8 cr30, cr31, cr32, cr33, cr35;
	u8 cr39, cr41, cr70, sr01, msr;
};

struct cursor_data {
	u32 cursor_enable;
	u32 cursor_show;
	u32 blink_count;
	u32 blink_rate;
	struct timer_list *timer;
};	

struct gtt_data {
	struct pgtbl_record *i810_master;
	u32 *i810_pgtbl_phys;
	u32 *i810_pgtbl_virtual;
	u32 pgtbl_scratch;
};
/* PCI */
static struct pci_device_id i810_pci_tbl[] __devinitdata = {
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82810_IG1, 
	PCI_ANY_ID, PCI_ANY_ID, 0, 0, PCI_DEVICE_ID_INTEL_82810_MC1 }, 
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82810_IG3,
	  PCI_ANY_ID, PCI_ANY_ID, 0, 0, PCI_DEVICE_ID_INTEL_82810_MC3 },
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82810_IG4,
	  PCI_ANY_ID, PCI_ANY_ID, 0, 0, PCI_DEVICE_ID_INTEL_82810_MC4},
	/* mvo: added i815 PCI-ID */  
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82815_100,
	  PCI_ANY_ID, PCI_ANY_ID, 0, 0, PCI_DEVICE_ID_INTEL_82815_100 },
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82815_NOAGP,
	  PCI_ANY_ID, PCI_ANY_ID, 0, 0, PCI_DEVICE_ID_INTEL_82815_NOAGP },
	{ PCI_VENDOR_ID_INTEL, PCI_DEVICE_ID_INTEL_82815_FULL,
	  PCI_ANY_ID, PCI_ANY_ID, 0, 0, PCI_DEVICE_ID_INTEL_82815_FULL }
};	  
	             
static int __devinit i810fb_init_pci (struct pci_dev *dev, 
				      const struct pci_device_id *entry);
static void __devexit i810fb_remove_pci(struct pci_dev *dev);
	

static struct pci_driver i810fb_driver = {
	name:		"i810fb",
	id_table:	i810_pci_tbl,
	probe:		i810fb_init_pci,
	remove:		i810fb_remove_pci,
};	

struct i810_fbinfo {
	struct fb_info fb_info;
	struct pci_dev *dev;
	struct state_registers hw_state;
	struct mode_registers mode_params;
	struct cursor_data cursor;
	struct gtt_data i810_gtt;
	struct display disp;
	unsigned long fb_size;
	unsigned long fb_start_virtual;
	unsigned long fb_start_phys;
	unsigned long fb_base_phys;
	unsigned long fb_offset;
	unsigned long mmio_start_phys;
	unsigned long mmio_start_virtual;
	unsigned long lring_start_phys;
	unsigned long lring_start_virtual;
	unsigned long lring_offset;
	unsigned long cursor_start_phys;
	unsigned long cursor_start_virtual;
	unsigned long cursor_offset;
	unsigned long page_order;
	unsigned long fb_max;
	unsigned long aper_size;
	u8 mem_freq;
};

static int i810fb_enable = 0;
static struct i810_fbinfo *i810_info;
static struct ringbuffer xring_state;
static int currcon = 0;
static int curr_pg_flag = 1;
static u32 xcon = 6;
static u32 xgart = 0;
static int hwcur = 0;
static int rr = 0;
static int bpp = 8;
static int mtrr = 0;
static int accel = 0;
static int hsync1 = 0;
static int hsync2 = 0;
static int vsync1 = 0;
static int vsync2 = 0;
static int fifo_count = 0;
static u16 xres = 640;
static u16 yres = 480;
static u16 vyres = 480;

/* "use once" vars */
static char i810fb_name[16] __devinitdata = "i810fb";
static struct fb_var_screeninfo i810fb_default __devinitdata = {
	/* 640x480, 8 bpp */
	640, 480, 640, 480, 0, 0, 8, 0,
	{0, 8, 0}, {0, 8, 0}, {0, 8, 0}, {0, 0, 0},
	0, 0, -1, -1, 0, 20000, 64, 64, 32, 32, 64, 2,
	0, FB_VMODE_NONINTERLACED
};

static char fontname[40]  __devinitdata = { 0 };
static u32 vram __devinitdata = 4;
static int ext_vga __devinitdata = 0;

    /*
     *  Interface used by the world
     */
int __init i810fb_setup(char *options);
static int i810fb_open(struct fb_info *info, int user);
static int i810fb_release(struct fb_info *info, int user);
static int i810fb_get_fix(struct fb_fix_screeninfo *fix, int con,
		       struct fb_info *info);
static int i810fb_get_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info);
static int i810fb_set_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info);
static int i810fb_pan_display(struct fb_var_screeninfo *var, int con,
			   struct fb_info *info);
static int i810fb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info);
static int i810fb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info);
static int i810fb_ioctl(struct inode *inode, struct file *file, u_int cmd,
		     u_long arg, int con, struct fb_info *info);
static int i810fb_mmap(struct fb_info *info, struct file *file, struct vm_area_struct *vma);

    /*
     *  Interface to the low level console driver
     */
int __init i810fb_init(void);
static int i810fbcon_switch(int con, struct fb_info *info);
static int i810fbcon_updatevar(int con, struct fb_info *info);
static void i810fbcon_blank(int blank, struct fb_info *info);


    /*
     *  Internal routines
     */
static u_long get_line_length(int xres_virtual, int bpp);
static void i810fb_encode_fix(struct fb_fix_screeninfo *fix,
			   struct fb_var_screeninfo *var);
static void set_color_bitfields(struct fb_var_screeninfo *var);
static int i810fb_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info);
static int i810fb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                         u_int transp, struct fb_info *info);
static void do_install_cmap(int con, struct fb_info *info);
static void i810fb_init_monspecs(struct fb_info *fb_info);
static void __devexit i810fb_restore_regs(void);
static void i810fb_load_regs(struct display *disp);
static void i810fb_save_vgax(void);
static void i810fb_protect_regs(int mode);
static void i810fb_set_disp(struct display *disp, int con, struct fb_info *info);
static int i810fb_calc_mode_limits(struct fb_info *info, struct display *disp);
static void i810fb_calc_vga(struct fb_info *info, struct display *disp);
static void i810fb_release_resource(void);
static inline void i810fb_lring_enable(u32 mode);
static inline int not_safe(void);
static inline int find_agpgart(void);
static int i810fb_calc_monitor_limits(struct fb_info *info, int xres, int yres);
static void i810fb_enable_cursor(int mode);
static void i810fb_set_cursor(struct display *disp);
static void i810_calc_dclk(struct mode_registers *params, u32 freq);
static void i810fb_screen_off(u8 mode);
static void i810fb_init_cursor(void);
static unsigned long i810_allocate_memory(u32 *pgtbl, u32 pgsize, u32 type, u32 offset);
static int i810_free_memory(u32 *pgtbl, u32 type, u32 offset);
static void i810fb_init_ringbuffer(void);
static int i810_find_chunk(u32 size);

/* Helper routines */ 
static inline u8 i810_readb(u32 where);
static inline u16 i810_readw(u32 where);
static inline u32 i810_readl(u32 where);
static inline void i810_writeb(u32 where, u8 val);
static inline void i810_writew(u32 where, u16 val);
static inline void i810_writel(u32 where, u32 val);
static inline void flush_cache(void);


/* BLT Engine Routines */

/**
 * wait_for_space - check ring buffer free space
 * @tail: offset to current tail of buffer
 * @val: amount of ringbuffer space needed
 *
 * DESCRIPTION:
 * The function waits until a free space from the ringbuffer
 * is available 
 */	

static inline void wait_for_space(u32 tail, u32 val)
{
	u32 head, i = 0, j;
	j = (val + 2) << 1;
	while (!i) {
		head = i810_readl(LRING + 4) & RBUFFER_HEAD_MASK;	
		if ((tail == head) || 
		    (tail > head && (tail - head) > j) ||
		    (tail < head && (head + RINGBUFFER_SIZE - tail) > j))           
			++i;	
	}
}

/**
 * wait_for_blit_idle - wait until 2D engine is idle
 */
static inline void wait_for_blit_idle(void)
{
	while(i810_readw(BLTCNTL) & 1);
}

/** 
 * wait_for_engine_idle - waits for all engine to finish
 *
 * DESCRIPTION:
 * This waits for lring(0), iring(1), batch(3), render(4), mapping(5) and
 * blitter(6) to finish
 */
static inline void wait_for_engine_idle(void)
{
	while((i810_readl(INSTDONE) & 0x7B) != 0x7B);
}

/**
 * lring_valid - check if ringbuffer register is valid
 */
static inline int lring_valid(void)
{
	u32 tmp;
	
	tmp = i810_readl(LRING + 8) & RBUFFER_START_MASK;
	if (tmp != (i810_info->lring_start_phys & 0x00FFFF00))
		return 0;
	return 1;
}
 
static inline u32 begin_lring(u32 space, u32 lring_tail)
{
	u32 val = 0, i, residual, tail;
	
	tail = lring_tail & RBUFFER_TAIL_MASK;
	residual = RINGBUFFER_SIZE - tail;
	if (residual >= space)
		wait_for_space(tail, space >> 2);
	else {
		val = residual + space;
		wait_for_space(tail, val >> 2);
		for (i = residual >> 2; i--; )
			PUT_RING(NOP);
		tail = 0;
	}
	return tail;
}

static inline void end_lring(u32 tail, u32 lring_tail)
{
	if (tail >= RINGBUFFER_SIZE)
		tail = 0;
	lring_tail &= ~RBUFFER_TAIL_MASK;
	wait_for_blit_idle();  
	i810_writel(RINGBUFFER, lring_tail | tail);
}


/**
 * i810fb_issue_client_instruction - write instruction packets to ringbuffer
 * @arg - pointer to structure for instruction packets
 *
 * DESCRIPTION:
 * The i810 engine use dword-multiple packets to instruct the engine.
 * The function must wait for sufficient buffer space and blit engine
 * idle before activating the new instruction sequence.  This particular
 * function is specifically made for clients residing in user-space.
 */

static int i810fb_issue_client_instruction(u32 arg)
{
	u32 temp, tail, space;

	get_user(space, (u32 *) arg);
	temp = i810_readl(LRING);
	tail = begin_lring(space, temp);

	if (copy_from_user((void *) (i810_info->lring_start_virtual + tail), 
			  (void *)(arg + 4), space))
		return -EFAULT;
	tail += space;
	
	end_lring(tail, temp);
	return 0;
}

/**
 * flush_gfx - flushes graphics pipeline
 *
 * DESCRIPTION:
 * Flushes the graphics pipeline.  This is done via
 * sending instruction packets to the ringbuffer
 */
static inline void flush_gfx(void)
{
	u32 temp, tail, space = 8;
	
	temp = i810_readl(LRING);
	tail = begin_lring(space, temp);
	
	PUT_RING(PARSER | FLUSH);
	PUT_RING(NOP);
	end_lring(tail, temp);
}

/**
 * mono_pat_blit - monochromatic pattern BLIT
 * @dpitch: no of pixel per line of destination 
 * @dheight:  height in pixels of the pattern
 * @dwidth: width in pixels of the pattern
 * @where: where to place the pattern
 * @fg: foreground color
 * @bg: background color
 * @rop: graphics raster operation
 * @patt_1: first part of an 8x8 pixel pattern
 * @patt_2: second part of an 8x8 pixel pattern
 *
 * DESCRIPTION:
 * Immediate monochromatic pattern BLIT function.  The pattern
 * is directly written to the ringbuffer.
 */
static void mono_pat_blit(int dpitch, int dheight, int dwidth, int dest, 
			  int fg, int bg, int rop, int patt_1, int patt_2)
{
	u32 temp, tail, space = 32;

	temp = i810_readl(LRING);
	tail = begin_lring(space, temp);

	PUT_RING(BLIT | MONO_PAT_BLIT | 6);
	PUT_RING(rop << 16 | dpitch);
	PUT_RING(dheight << 16 | dwidth);
	PUT_RING(dest);
	PUT_RING(bg);
	PUT_RING(fg);
	PUT_RING(patt_1);
	PUT_RING(patt_2);

	end_lring(tail, temp);
}
/**
 * source_copy_blit - BLIT transfer operation
 * @dwidth: width of rectangular graphics data
 * @dheight: height of rectangular graphics data
 * @dpitch: pixels per line of destination buffer
 * @xdir: direction of copy (left to right or right to left)
 * @spitch: pixels per line of source buffer
 * @from: source address
 * @where: destination address
 * @rop: raster operation
 *
 * DESCRIPTION:
 * This is a BLIT operation typically used when doing
 * a 'Copy and Paste'
 */
static void source_copy_blit(int dwidth, int dheight, int dpitch, int xdir, 
			     int spitch, int src, int dest, int rop)
{
	u32 temp, tail, space = 24;

	temp = i810_readl(LRING);
	tail = begin_lring(space, temp);

	PUT_RING(BLIT | SOURCE_COPY_BLIT | 4);
	PUT_RING(xdir | rop << 16 | dpitch);
	PUT_RING(dheight << 16 | dwidth);
	PUT_RING(dest);
	PUT_RING(spitch);
	PUT_RING(src);

	end_lring(tail, temp);

}	

/**
 * color_blit - solid color BLIT operation
 * @width: width of destination
 * @height: height of destination
 * @pitch: pixels per line of the buffer
 * @where: destination
 * @rop: raster operation
 * @what: color to transfer
 *
 * DESCRIPTION:
 * A BLIT operation which can be used for  color fill/rectangular fill
 */
static void color_blit(int width, int height, int pitch, int dest, int rop, int what)
{
	u32 temp, tail, space = 24;

	temp = i810_readl(LRING);
	tail = begin_lring(space, temp);

	PUT_RING(BLIT | COLOR_BLT | 3);
	PUT_RING(rop << 16 | pitch | SOLIDPATTERN);
	PUT_RING(height << 16 | width);
	PUT_RING(dest);
	PUT_RING(what);
	PUT_RING(NOP);

	end_lring(tail, temp);
}
/* 
 * The next functions are the accelerated equivalents of the
 * generic framebuffer operation.  Each function will only 
 * proceed if the graphics table is valid.  They all finally
 * end up calling one of the BLIT functions.
 */

void i810_accel_setup(struct display *p)
{
	p->next_line = p->line_length ? p->line_length : 
		p->var.xres_virtual * (p->var.bits_per_pixel) >> 3;
	p->next_plane = 0;
}

void i810_accel_bmove(struct display *p, int sy, int sx, int dy, int dx, 
		       int height, int width)
{
	int pitch, xdir, src, dest, mult;
	
	if (not_safe()) return;
	if (!lring_valid) 
		i810fb_init_ringbuffer();
	dy *= fontheight(p);
	sy *= fontheight(p);
	height *= fontheight(p);
	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	
	sx *= fontwidth(p) * mult;
	dx *= fontwidth(p) * mult;
	width *= fontwidth(p) * mult;			

	if (dx <= sx) 
		xdir = INCREMENT;
	else {
		xdir = DECREMENT;
		sx += width - 1;
		dx += width - 1;
	}
	if (dy <= sy) 
		pitch = p->next_line;
	else {
		pitch = (-(p->next_line)) & 0xFFFF; /* NOTE: pitch needs to be quadword aligned */
		sy += height - 1;
		dy += height - 1;
	}
	src = (sy * p->next_line) + sx; 
	dest = (dy * p->next_line) + dx;
	source_copy_blit(width, height, pitch, xdir, pitch, src, dest, PAT_COPY_ROP); 
}
void i810_accel_clear(struct vc_data *conp, struct display *p, int sy, int sx,
		      int height, int width)
{
	int mult;
	u32 bgx = 0, dest = 0;
	
	if (not_safe()) return;
	if (!lring_valid) 
		i810fb_init_ringbuffer();
	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	height *= fontheight(p);
	sy *= fontheight(p);

	switch(mult) {
	case 1:
		bgx = (u32) attr_bgcol_ec(p, conp);
		break;
	case 2:
		bgx = (int) ((u16 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
		break;
	case 3:
		bgx = ((int *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
	}	         	         

	width *= fontwidth(p) * mult;
	sx *= fontwidth(p) * mult;
	dest = (sy * p->next_line) + sx;
	color_blit(width, height, p->next_line, dest, COLOR_COPY_ROP, bgx);
}

void i810_accel_putc(struct vc_data *conp, struct display *p, int c, int yy, int xx)
{
	u8 *cdat;
	int dheight, dwidth, dpitch, fg = 0, bg = 0, dest, fontwidth, mult;
	register int skip, i, j, times, pattern = 0;

	if (not_safe()) return;
	if (!lring_valid) 
		i810fb_init_ringbuffer();
	/* 8x8 pattern blit */
	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	dheight = 8;              
	dwidth = 8 * mult;
	dpitch = p->next_line;
	switch (mult) {
	case 1:
		fg = (int) attr_fgcol(p,c);
		bg = (int) attr_bgcol(p,c);
		break;
	case 2:
		bg = (int) ((u16 *)p->dispsw_data)[attr_bgcol(p, c)];
		fg = (int) ((u16 *)p->dispsw_data)[attr_fgcol(p, c)];
		break;
	case 3:
		fg = ((int *) p->dispsw_data)[attr_fgcol(p, c)];
		bg = ((int *) p->dispsw_data)[attr_bgcol(p, c)];
	}
	times = fontheight(p) >> 3;
	if(fontwidth(p) <= 8) {
		fontwidth = 8;
		cdat = p->fontdata + ((c & p->charmask) * fontheight(p));
		skip = 1;
	}
	else {
		fontwidth = fontwidth(p);
		cdat = p->fontdata + ((c & p->charmask) * (fontwidth >> 3) * fontheight(p));
		skip = fontwidth >> 3;
	}
	dest = (yy * fontheight(p) * p->next_line) + 
		xx * (fontwidth * mult);

	for (i = skip; i--; ) {
		for (j = 0; j < times; j++) {
			pattern = j << (skip + 2);
			mono_pat_blit(dpitch, dheight, dwidth, dest + (j * (dpitch << 3)),
				      fg, bg, COLOR_COPY_ROP, 
				      *(u32 *)(cdat + pattern), 
				      *(u32 *)(cdat + pattern + (skip << 2)));

		}
		cdat += 4;
		dest += dwidth;
	}
}

inline int i810_accel8_getbg(struct display *p, int c)
{
	return attr_bgcol(p, c);
}
inline int i810_accel8_getfg(struct display *p, int c)
{
	return attr_fgcol(p, c);
}
inline int i810_accel16_getbg(struct display *p, int c)
{
	return (int) ((u16 *) p->dispsw_data)[attr_bgcol(p, c)];
}
inline int i810_accel16_getfg(struct display *p, int c)
{
	return (int) ((u16 *) p->dispsw_data)[attr_fgcol(p, c)];
}
inline int i810_accel24_getbg(struct display *p, int c)
{
	return ((int *) p->dispsw_data)[attr_bgcol(p, c)];
}
inline int i810_accel24_getfg(struct display *p, int c)
{
	return ((int *) p->dispsw_data)[attr_fgcol(p, c)];
}

void i810_accel_putcs(struct vc_data *conp, struct display *p,
			const unsigned short *s, int count, int yy, int xx)
{
	u8 *cdat;
	int c, chars, dpitch, dheight, dwidth, dest, fg = 0, bg = 0, offset, mult, fontwidth;
	register int skip, times, i, j, k, pattern = 0;
	int (*get_bg)(struct display *p, int c) = NULL;
	int (*get_fg)(struct display *p, int c) = NULL;
	
	if (not_safe()) return;
	if (!lring_valid) 
		i810fb_init_ringbuffer();
	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	
	dwidth = 8 * mult;
	dheight = 8;
	dpitch = p->next_line;
	times = fontheight(p) >> 3;
	if(fontwidth(p) <= 8) {
		skip = 1;
		fontwidth = 8;
	}
	else {
		skip = fontwidth(p) >> 3;
		fontwidth = fontwidth(p);
	}
	dest = (yy * fontheight(p) * p->next_line) + 
		       xx * (fontwidth * mult);
	chars = p->next_line/fontwidth;
	offset = skip * fontheight(p);
	switch (mult) {
	case 1:
		get_bg = i810_accel8_getbg;
		get_fg = i810_accel8_getfg;
		break;
	case 2:
		get_bg = i810_accel16_getbg;
		get_fg = i810_accel16_getfg;
		break;
	case 3:	
		get_bg = i810_accel24_getbg;
		get_fg = i810_accel24_getfg;
	}
	while(count--) {
		c = scr_readw(s++);
		cdat = p->fontdata + ((c & p->charmask) * offset);
		fg = get_fg(p, c);
		bg = get_bg(p, c);

		for (i = skip; i--; ) {
			k = dest;
			for (j = 0; j < times; j++) {
				pattern = j << (skip + 2);
				mono_pat_blit(dpitch, dheight, dwidth, k, 
					      fg, bg, COLOR_COPY_ROP, 
					      *(u32 *)(cdat + pattern), 
				      	      *(u32 *)(cdat + pattern + (skip << 2)));
				k += dpitch << 3; 				      	      
			}
			dest += dwidth;
			cdat += 4;
		}
		if(++xx > chars) {
			xx = 0;
			dest += p->next_line * fontheight(p);
		}
		
	}	 
}

void i810_accel_revc(struct display *p, int xx, int yy)
{
	int width, height, dest, pitch, mult;
	if (not_safe()) return;
	if (!lring_valid) 
		i810fb_init_ringbuffer();
	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;
	
	dest = (yy *fontheight(p) * p->next_line) + 
		xx * (fontwidth(p) * mult);
	pitch = p->next_line;
	if (fontwidth(p) <= 8)
		width = 8;
	else
		width = fontwidth(p);
	height = fontheight(p);
	color_blit(width, height, pitch, dest, INVERT_ROP, 0x0F);
}

void i810_accel_clear_margins(struct vc_data *conp, struct display *p,
			       int bottom_only)
{
	int bytes = p->next_line;
	u32 bgx = 0, mult;
	unsigned int right_start;
	unsigned int bottom_start = (conp->vc_rows)*fontheight(p);
	unsigned int right_width, bottom_width;

	if (not_safe()) return;
	if (!lring_valid) 
		i810fb_init_ringbuffer();
	mult = (p->var.bits_per_pixel >> 3);
	if (!mult) mult = 1;

	switch(mult) {
	case 1:
		bgx = (u32) attr_bgcol_ec(p, conp);
		break;
	case 2:
		bgx = (int) ((u16 *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
		break;
	case 3:
		bgx = ((int *)p->dispsw_data)[attr_bgcol_ec(p, conp)];
	}	         	         

	right_start = (conp->vc_cols)*fontwidth(p)*mult;
	if (!bottom_only && (right_width = (p->var.xres*mult)-right_start))
		color_blit(right_width, p->var.yres_virtual, bytes, right_start, COLOR_COPY_ROP, bgx); 
	if ((bottom_width = p->var.yres-bottom_start))
		color_blit(right_start, bottom_width, bytes, (p->var.yoffset+bottom_start)*bytes, 
			   COLOR_COPY_ROP, bgx);
}

void i810_cursor(struct display *p, int mode, int xx, int yy)
{
	int temp;
	
	yy -= p->yscroll;
	yy *= fontheight(p);
	if (fontwidthlog(p)) 
		xx <<= fontwidthlog(p);
	else 
		xx *= fontwidth(p);
	i810fb_enable_cursor(OFF);
	temp = xx & 0xFFFF;
	temp |= yy << 16;
	i810_writel(CURPOS, temp);
	
	switch (mode) {
		case CM_ERASE:
			i810_info->cursor.cursor_enable = 0;
			break;
		case CM_MOVE:
		case CM_DRAW:
			i810fb_enable_cursor(ON);
			i810_info->cursor.blink_count = i810_info->cursor.blink_rate;
			i810_info->cursor.cursor_enable = 1;
			i810_info->cursor.cursor_show = 1;
				
	}
} 
 
/* 
 * The following are wrappers for the generic framebuffer operations.  Each
 * is preceeded by a call to 'not_safe' before calling the actual generic 
 * operation.
 */
 
void i810_noaccel8_bmove(struct display *p, int sy, int sx, int dy, int dx,
                       int height, int width)
{
	if (not_safe()) return;
	fbcon_cfb8_bmove(p, sy, sx, dy, dx, height, width);
}

void i810_noaccel8_clear(struct vc_data *conp, struct display *p, int sy, int sx,
                      int height, int width)
{
	if (not_safe()) return;
	fbcon_cfb8_clear(conp, p, sy, sx, height, width);
}

void i810_noaccel8_putc(struct vc_data *conp, struct display *p, int c, int yy, int xx)
{	    
	if (not_safe()) return;
	fbcon_cfb8_putc(conp, p, c, yy, xx);
}

void i810_noaccel8_putcs(struct vc_data *conp, struct display *p,
                        const unsigned short *s, int count, int yy, int xx)
{
	if (not_safe()) return;
	fbcon_cfb8_putcs(conp, p, s, count, yy, xx);
}

void i810_noaccel8_revc(struct display *p, int xx, int yy)
{	                        	                  	                        
	if (not_safe()) return;
	fbcon_cfb8_revc(p, xx, yy);
}

void i810_noaccel8_clear_margins(struct vc_data *conp, struct display *p,
                               int bottom_only)
{           
	if (not_safe()) return;
	fbcon_cfb8_clear_margins(conp, p, bottom_only);
}

void i810_noaccel16_bmove(struct display *p, int sy, int sx, int dy, int dx,
                       int height, int width)
{
	if (not_safe()) return;
	fbcon_cfb16_bmove(p, sy, sx, dy, dx, height, width);
}

void i810_noaccel16_clear(struct vc_data *conp, struct display *p, int sy, int sx,
                      int height, int width)
{
	if (not_safe()) return;
	fbcon_cfb16_clear(conp, p, sy, sx, height, width);
}

void i810_noaccel16_putc(struct vc_data *conp, struct display *p, int c, int yy, int xx)
{	    
	if (not_safe()) return;
	fbcon_cfb16_putc(conp, p, c, yy, xx);
}

void i810_noaccel16_putcs(struct vc_data *conp, struct display *p,
                        const unsigned short *s, int count, int yy, int xx)
{
	if (not_safe()) return;
	fbcon_cfb16_putcs(conp, p, s, count, yy, xx);
}

void i810_noaccel16_revc(struct display *p, int xx, int yy)
{	                        	                  	                        
	if (not_safe()) return;
	fbcon_cfb16_revc(p, xx, yy);
}

void i810_noaccel16_clear_margins(struct vc_data *conp, struct display *p,
                               int bottom_only)
{           
	if (not_safe()) return;
	fbcon_cfb16_clear_margins(conp, p, bottom_only);
}

void i810_noaccel24_bmove(struct display *p, int sy, int sx, int dy, int dx,
                       int height, int width)
{
	if (not_safe()) return;
	fbcon_cfb24_bmove(p, sy, sx, dy, dx, height, width);
}

void i810_noaccel24_clear(struct vc_data *conp, struct display *p, int sy, int sx,
                      int height, int width)
{
	if (not_safe()) return;
	fbcon_cfb24_clear(conp, p, sy, sx, height, width);
}

void i810_noaccel24_putc(struct vc_data *conp, struct display *p, int c, int yy, int xx)
{	    
	if (not_safe()) return;
	fbcon_cfb24_putc(conp, p, c, yy, xx);
}

void i810_noaccel24_putcs(struct vc_data *conp, struct display *p,
                        const unsigned short *s, int count, int yy, int xx)
{
	if (not_safe()) return;
	fbcon_cfb24_putcs(conp, p, s, count, yy, xx);
}

void i810_noaccel24_revc(struct display *p, int xx, int yy)
{	                        	                  	                        
	if (not_safe()) return;
	fbcon_cfb24_revc(p, xx, yy);
}

void i810_noaccel24_clear_margins(struct vc_data *conp, struct display *p,
                               int bottom_only)
{           
	if (not_safe()) return;
	fbcon_cfb24_clear_margins(conp, p, bottom_only);
}

struct display_switch i810_noaccel8 = {
	fbcon_cfb8_setup,
	i810_noaccel8_bmove,
	i810_noaccel8_clear,
	i810_noaccel8_putc,
	i810_noaccel8_putcs,
	i810_noaccel8_revc,
	NULL,
	NULL,
	i810_noaccel8_clear_margins,
	FONTWIDTH(4) | FONTWIDTH(8) | FONTWIDTH(12) | FONTWIDTH(16)
};

struct display_switch i810_noaccel16 = {	
	fbcon_cfb16_setup,
	i810_noaccel16_bmove,
	i810_noaccel16_clear,
	i810_noaccel16_putc,
	i810_noaccel16_putcs,
	i810_noaccel16_revc,
	NULL,
	NULL,
	i810_noaccel16_clear_margins,
	FONTWIDTH(4) | FONTWIDTH(8) | FONTWIDTH(12) | FONTWIDTH(16)
};

struct display_switch i810_noaccel24 = {	
	fbcon_cfb24_setup,
	i810_noaccel24_bmove,
	i810_noaccel24_clear,
	i810_noaccel24_putc,
	i810_noaccel24_putcs,
	i810_noaccel24_revc,
	NULL,
	NULL,
	i810_noaccel24_clear_margins,
	FONTWIDTH(4) | FONTWIDTH(8) | FONTWIDTH(12) | FONTWIDTH(16)
};

struct display_switch i810_accel = {

	i810_accel_setup,
	i810_accel_bmove,
	i810_accel_clear,
	i810_accel_putc,
	i810_accel_putcs,
	i810_accel_revc,
	NULL,
	NULL,
	i810_accel_clear_margins,
	FONTWIDTH(4) | FONTWIDTH(8) | FONTWIDTH(12) | FONTWIDTH(16)
};

/* 
 * IOCTL operations
 */

/**
 * i810fb_allocate_surface - gets GART memory 
 * @var - pointer to 32-bit var containing size to allocate and offset from GART upon exit
 *
 * DESCRIPTION:
 * Allocates memory from GART
 *
 * RETURNS:
 * nonzero if fail
 */
/*static int i810fb_allocate_surface(u32 *var)
{
	u32 pgsize = 0;
	int offset;

	pgsize = (*var + I810_PAGESIZE - 1) >> 12;
	offset = i810_find_chunk(pgsize);
	if (offset < 0)
		return -ENOMEM;
	if (!(i810_allocate_memory(&i810_info, i810_pgtbl_virtual, pgsize, AGP_NORMAL, offset))) 
		return -ENOMEM;
	*var = (u32)(offset << 12);
	return 0;
}
*/
/**
 * i810fb_free_surface - free GART memory
 * @var: offset from GART start to free
 *
 * DESCRIPTION:
 * Frees GART memory
 * 
 * RETURNS:
 * nonzero if fail
 */
/*
static int i810fb_free_surface(u32 *var)
{
	u32 offset;
	
	offset = *var << 12;
	if (!i810_free_memory(&i810_info, i810_pgtbl_virtual, AGP_NORMAL, offset))
		return -EFAULT;
	return 0;
}
*/
	
static int i810fb_request_fifo(struct request_fifo_struct *fifo)
{
	if (fb_display[currcon].var.accel_flags)
		return -EINVAL;
	fifo->size = RINGBUFFER_SIZE;
	fifo->offset = i810_find_chunk(fifo->size >> 12);
	if (fifo->offset < 0)
		return -ENOMEM;
	if (!(i810_allocate_memory(i810_info->i810_gtt.i810_pgtbl_virtual, fifo->size >> 12, AGP_NORMAL, 
				   fifo->offset)))
		return -ENOMEM;
	fifo->offset <<= 12;
	fifo->lring_addr_phys = i810_info->fb_start_phys + (fifo->offset);
	++fifo_count;
	return 0;
}

static int i810fb_release_fifo(u32 offset)
{
	int err = 0;

	i810fb_init_ringbuffer();
	i810_free_memory(i810_info->i810_gtt.i810_pgtbl_virtual, AGP_NORMAL, offset >> 12);
	if (!fifo_count) 
		err = -EINVAL;
	if (fifo_count)
		--fifo_count;
	return err;
}

static struct fb_ops i810fb_ops = {
    THIS_MODULE, i810fb_open, i810fb_release, i810fb_get_fix, i810fb_get_var, i810fb_set_var, 
    i810fb_get_cmap, i810fb_set_cmap, i810fb_pan_display, i810fb_ioctl, i810fb_mmap
};


    /*
     *  Open/Release the frame buffer device
     */

static int i810fb_open(struct fb_info *info, int user)
{
	/*                                                                     
	 *  Nothing, only a usage count for the moment                          
	 */                                                                    

	MOD_INC_USE_COUNT;
	return(0);                              
}
        
static int i810fb_release(struct fb_info *info, int user)
{
	MOD_DEC_USE_COUNT;
	return(0);                                                    
}


    /*
     *  Get the Fixed Part of the Display
     */

static int i810fb_get_fix(struct fb_fix_screeninfo *fix, int con,
		       struct fb_info *info)
{
	struct fb_var_screeninfo *var;
	if (con == -1)
		var = &i810fb_default;
	else
		var = &fb_display[con].var;
	i810fb_encode_fix(fix, var);
	return 0;
}


    /*
     *  Get the User Defined Part of the Display
     */

static int i810fb_get_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info)
{
	if (con == -1)
		*var = i810fb_default;
	else
		*var = fb_display[con].var;
	set_color_bitfields(var);
	return 0;
}

static int i810fb_check_limits(struct fb_var_screeninfo *var, struct fb_info *info)
{
	u32 line_length;
	/*
	 *  Memory limit
	 */
	line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
	if (line_length*var->yres_virtual > i810_info->fb_size) 
		return -ENOMEM;
		
	/*
	 * Monitor limit
	 */
	 if (!i810fb_calc_monitor_limits(info, var->xres, var->yres)) 
	 	return -EINVAL;
	return 0;
}	

static void i810fb_round_off(struct fb_var_screeninfo *var)
{
	/*
	 *  Presently supports only these configurations 
	 */
	if (var->xres <= 640) 
		var->xres = 640;
#ifndef CONFIG_FB_I810_NONSTD
	if (var->xres < 800)
		var->xres = 640;
	if (var->xres < 1024 && var->xres >= 800)
		var->xres = 800;
	if (var->xres < 1152 && var->xres >= 1024)
		var->xres = 1024;
	if (var->xres < 1280 && var->xres >= 1152)
		var->xres = 1152;
	if (var->xres < 1600 && var->xres >= 1280)
		var->xres = 1280;
#endif /* CONFIG_FB_I810_NONSTD */		
	if (var ->xres >= 1600) 
		var->xres = 1600;
	var->xres = (var->xres + 7) & ~0x07;
	var->yres = (3 * var->xres) >> 2;
	if (var->xres > var->xres_virtual)
		var->xres_virtual = var->xres;
	if (var->yres > var->yres_virtual)
		var->yres_virtual = var->yres;
	if (var->bits_per_pixel <= 8)
		var->bits_per_pixel = 8;
	else if (var->bits_per_pixel <= 16)
		var->bits_per_pixel = 16;
	else 
		var->bits_per_pixel = 24;
}	
	

     /*
     *  Set the User Defined Part of the Display
     */

static int i810fb_set_var(struct fb_var_screeninfo *var, int con,
		       struct fb_info *info)
{
	int err, activate = var->activate;
	int oldxres, oldyres, oldvxres, oldvyres, oldbpp, oldaccel;
	
	struct display *display;
	if (con >= 0)
		display = &fb_display[con];
	else
	    display = &i810_info->disp;	/* used during initialization */
	
	/*
	 *  FB_VMODE_CONUPDATE and FB_VMODE_SMOOTH_XPAN are equal!
	 *  as FB_VMODE_SMOOTH_XPAN is only used internally
	 */

	if (var->vmode & FB_VMODE_CONUPDATE) {
		var->vmode |= FB_VMODE_YWRAP;
		var->xoffset = display->var.xoffset;
		var->yoffset = display->var.yoffset;
	}
	i810fb_round_off(var);
	if ((err = i810fb_check_limits(var, info))) 
		return err;		
	set_color_bitfields(var);
	if ((activate & FB_ACTIVATE_MASK) == FB_ACTIVATE_NOW) {
		oldxres = display->var.xres;
		oldyres = display->var.yres;
		oldvxres = display->var.xres_virtual;
		oldvyres = display->var.yres_virtual;
		oldbpp = display->var.bits_per_pixel;
		oldaccel = display->var.accel_flags;
		display->var = *var;

		if (oldxres != var->xres || oldyres != var->yres ||
		    oldvxres != var->xres_virtual || oldvyres != var->yres_virtual ||
		    oldbpp != var->bits_per_pixel || oldaccel != var->accel_flags) {
			struct fb_fix_screeninfo fix;
			
			i810fb_encode_fix(&fix, var);
			display->screen_base = (char *) i810_info->fb_start_virtual;
			display->visual = fix.visual;
			display->type = fix.type;
			display->type_aux = fix.type_aux;
			display->ypanstep = fix.ypanstep;
			display->ywrapstep = fix.ywrapstep;
			/* Choose the most efficient scrolling method */
			if (var->yres_virtual > var->yres)
				display->scrollmode = SCROLL_YNOMOVE;
			else if (var->accel_flags && var->bits_per_pixel == 8)
				display->scrollmode = 0;
			else
				display->scrollmode = SCROLL_YREDRAW;
			display->line_length = fix.line_length;
			display->can_soft_blank = 1;
			display->inverse = 0;
			switch (var->bits_per_pixel) {
#ifdef FBCON_HAS_CFB8
			case 8:
				display->dispsw = (accel && var->accel_flags) 
				                   ? &i810_accel : &i810_noaccel8; 
				break;
#endif
#ifdef FBCON_HAS_CFB16
			case 16:
				display->dispsw = (accel && var->accel_flags)
				                   ? &i810_accel : &i810_noaccel16;
				display->dispsw_data = i810_cfb16;
				break;
#endif
#ifdef FBCON_HAS_CFB24
			case 24:
				display->dispsw = (accel && var->accel_flags)
				                   ? &i810_accel : &i810_noaccel24;
				display->dispsw_data = i810_cfb24;
				break;
#endif
				
			default:
				display->dispsw = &fbcon_dummy;
				break;
			}
			if (i810_info->fb_info.changevar)
				(*i810_info->fb_info.changevar)(con);
		}
		if (con < 0)
			return 0;
		if (oldbpp != var->bits_per_pixel) {
			if ((err = fb_alloc_cmap(&display->cmap, 0, 0))) 
				return err;
			do_install_cmap(con, info);
		}
		if (!i810fb_calc_mode_limits(info, display))
			return -EINVAL;
		i810fb_calc_vga(info, display);
		if(!i810fb_enable)
			i810fb_load_regs(display);
		if (hwcur) {
			i810fb_init_cursor();
			i810fb_set_cursor(display);
		}
		i810fb_set_disp(display, con, info);
	}
	return 0;
}




    /*
     *  Pan or Wrap the Display
     *
     *  This call looks only at xoffset, yoffset and the FB_VMODE_YWRAP flag
     */

static int i810fb_pan_display(struct fb_var_screeninfo *var, int con,
			   struct fb_info *info)
{
	if (var->vmode & FB_VMODE_YWRAP) {
		if (var->yoffset < 0 ||
		    var->yoffset >= fb_display[con].var.yres_virtual ||
		    var->xoffset)
			return -EINVAL;
	} else {
		if (var->xoffset+fb_display[con].var.xres >
		    fb_display[con].var.xres_virtual ||
		    var->yoffset+fb_display[con].var.yres >
		    fb_display[con].var.yres_virtual)
			return -EINVAL;
	}
	fb_display[con].var.xoffset = var->xoffset;
	fb_display[con].var.yoffset = var->yoffset;
	i810fbcon_updatevar(con, info);
	if (var->vmode & FB_VMODE_YWRAP)
		fb_display[con].var.vmode |= FB_VMODE_YWRAP;
	else
		fb_display[con].var.vmode &= ~FB_VMODE_YWRAP;
	return 0;
}

    /*
     *  Get the Colormap
     */

static int i810fb_get_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info)
{
	if (con == currcon) /* current console? */
		return fb_get_cmap(cmap, kspc, i810fb_getcolreg, info);
	else if (fb_display[con].cmap.len) /* non default colormap? */
		fb_copy_cmap(&fb_display[con].cmap, cmap, kspc ? 0 : 2);
	else
		fb_copy_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel),
			     cmap, kspc ? 0 : 2);
	return 0;
}

    /*
     *  Set the Colormap
     */

static int i810fb_set_cmap(struct fb_cmap *cmap, int kspc, int con,
			struct fb_info *info)
{
	int err;
	
	if (!fb_display[con].cmap.len) {	/* no colormap allocated? */
		if ((err = fb_alloc_cmap(&fb_display[con].cmap,
					 1<<fb_display[con].var.bits_per_pixel, 0)))
			return err;
	}
	if (con == currcon)			/* current console? */
		return fb_set_cmap(cmap, kspc, i810fb_setcolreg, info);
	else
		fb_copy_cmap(cmap, &fb_display[con].cmap, kspc ? 0 : 1);
	return 0;
}

    /*
     *  Virtual Frame Buffer Specific ioctls
     */

static int i810fb_ioctl(struct inode *inode, struct file *file, u_int cmd,
		     u_long arg, int con, struct fb_info *info)
{
	int i;
	struct request_fifo_struct fifo;

	switch(cmd) {
	case I810FB_IOC_INSTRUCTFIFO:
		return i810fb_issue_client_instruction(arg);
	case I810FB_IOC_REQUESTFIFO:
		if(copy_from_user(&fifo, (void *) arg, sizeof(fifo)))
			return -EFAULT;
		if ((i = i810fb_request_fifo(&fifo)))
			return i;
		if(copy_to_user((void *) arg, &fifo, sizeof(fifo)))
			return -EFAULT;
		return 0;
	case I810FB_IOC_RELEASEFIFO:
		return i810fb_release_fifo((u32) arg);
	}

	return -EINVAL;
}

/**
 * i810fb_mmap - mmap framebuffer, mmio and off-screen surface
 * @info - pointer to fb_info
 * @file - file descriptor
 * @vma - virtual memory area structure
 *
 * DESCRIPTION:
 * This is a specialized mmap service for the i810fb.  Aside from memory mapping
 * the framebuffer and MMIO space, it also allows mapping of off-screen surfaces for
 * use as DMA/FIFO buffers which have been previously allocated via I810FB_IOC_REQUESTFIFO ioctl.  
 * The format to map the off-screen surface is fix->smem_len + fix->mmio_len + surface offset.  
 * The "surface offset" is returned by the previous ioctl call.
 */
static int i810fb_mmap(struct fb_info *info, struct file *file, struct vm_area_struct *vma)
{
	unsigned long off, start;
	u32 len;

	if (vma->vm_pgoff > (~0UL >> PAGE_SHIFT))
		return -EINVAL;
	off = vma->vm_pgoff << PAGE_SHIFT;
	start =i810_info->fb_start_phys;
	len = PAGE_ALIGN((start & ~PAGE_MASK) + i810_info->fb_size);
	if (off >= len + MMIO_SIZE) { 
		/* Client off-screen memory */
		off -= len + MMIO_SIZE;
		if (info->var.accel_flags)
			return -EINVAL;
		start = i810_info->fb_start_phys + off;
		len = PAGE_ALIGN((start & ~PAGE_MASK) + RINGBUFFER_SIZE);
		off = 0;
	}
	if (off >= len && off <= len + MMIO_SIZE) {
		/* memory mapped io */
		off -= len;
		if (info->var.accel_flags)
			return -EINVAL;
		start = i810_info->mmio_start_phys;
		len = PAGE_ALIGN((start & ~PAGE_MASK) + MMIO_SIZE);
	}
	start &= PAGE_MASK;
	if ((vma->vm_end - vma->vm_start + off) > len)
		return -EINVAL;
	off += start;
	vma->vm_pgoff = off >> PAGE_SHIFT;
#if defined(__i386__) || defined(__x86_64__)
 	if (boot_cpu_data.x86 > 3)
		pgprot_val(vma->vm_page_prot) |= _PAGE_PCD;
#endif 
	if (io_remap_page_range(vma->vm_start, off, vma->vm_end - vma->vm_start, vma->vm_page_prot))
		return -EAGAIN;
	return 0;
}

int __init i810fb_setup(char *options)
{
	char *this_opt;
	
	fontname[0] = '\0';
	i810fb_enable = 1;
	if (!options || !*options)
		return 0;
	
	for (this_opt = strtok(options, ":"); this_opt;
	     this_opt = strtok(NULL, ":")) {
		if (!strncmp(this_opt, "font=", 5))
			strcpy(fontname, this_opt+5);
#ifdef CONFIG_MTRR
		else if (!strncmp(this_opt, "mtrr", 4))
			mtrr = 1;
#endif
		else if (!strncmp(this_opt, "accel", 5))			
			accel = 1;
		else if (!strncmp(this_opt, "ext_vga", 7))
			ext_vga = 1;
		else if (!strncmp(this_opt, "vram=", 5))
			vram = (simple_strtoul(this_opt+5, NULL, 0));
		else if (!strncmp(this_opt, "xres=", 5))
			xres = simple_strtoul(this_opt+5, NULL, 0);
		else if (!strncmp(this_opt, "yres=", 5))
			yres = simple_strtoul(this_opt+5, NULL, 0);
		else if (!strncmp(this_opt, "vyres=", 6))
			vyres = simple_strtoul(this_opt+6, NULL, 0);
		else if (!strncmp(this_opt, "bpp=", 4))
			bpp = simple_strtoul(this_opt+4, NULL, 0);
		else if (!strncmp(this_opt, "rr=", 3))
			rr = simple_strtoul(this_opt+3, NULL, 0);
		else if (!strncmp(this_opt, "hsync1=", 7))
			hsync1 = simple_strtoul(this_opt+7, NULL, 0);
		else if (!strncmp(this_opt, "hsync2=", 7))
			hsync2 = simple_strtoul(this_opt+7, NULL, 0);
		else if (!strncmp(this_opt, "vsync1=", 7))
			vsync1 = simple_strtoul(this_opt+7, NULL, 0);
		else if (!strncmp(this_opt, "vsync2=", 7))
			vsync2 = simple_strtoul(this_opt+7, NULL, 0);
		else if (!strncmp(this_opt, "xcon=", 5))
			xcon = simple_strtoul(this_opt+5, NULL, 0);	
		else if (!strncmp(this_opt, "hwcur", 5))
			hwcur = 1;
	}
	return 0;
}


/* Intenal  routines */

/* 
 * Helper inline functions
 */
static inline u8 i810_readb(u32 where)
{
        return readb(i810_info->mmio_start_virtual + where);
}

static inline u16 i810_readw(u32 where)
{
	return readw(i810_info->mmio_start_virtual + where);
}

static inline u32 i810_readl(u32 where)
{
	return readl(i810_info->mmio_start_virtual + where);
}

static inline void i810_writeb(u32 where, u8 val)
{
	writeb(val, i810_info->mmio_start_virtual + where);
}

static inline void i810_writew(u32 where, u16 val)
{
	writew(val, i810_info->mmio_start_virtual + where);
}

static inline void i810_writel(u32 where, u32 val)
{
	writel(val, i810_info->mmio_start_virtual + where);
}

static u8 is_cga(void)
{	
	u8 val;
	val = i810_readb(MSR_READ);
	return val &= 1; 
}

static inline void i810_wait_for_scan_start(void)
{
	while((i810_readw(DISP_SL) & 0xFFF));
}

static inline void vsync_is_active(u32 port)
{
	while ((i810_readb(port) & 0x08));
}

static inline void vsync_is_inactive(u32 port)
{
	while (!(i810_readb(port) & 0x08));
}

static inline void i810_wait_for_hsync(void)
{
	while((i810_readw(DISP_SL) & 0xFFF) < 5);
}
	
static inline void i810_wait_for_vsync(int count)
{
	u32 port;	

	if (!count) return;

	if (is_cga) port = ST01_CGA;
	else port =  ST01_MDA;
		
	while(count--) {
		 vsync_is_active(port);
		 vsync_is_inactive(port);
	}
}       

/**
 * i810fb_screen_off - turns off/on display
 * @mode: on or off
 *
 * DESCRIPTION:
 * Blanks/unblanks the display
 */
static void i810fb_screen_off(u8 mode)
{
	u8 val;
	i810_writeb(SR_INDEX, SR01);
	val = i810_readb(SR_DATA);
	if (mode == OFF) 
		val |= SCR_OFF;
	else
		val &= ~SCR_OFF;

	i810_wait_for_scan_start(); 
	i810_writeb(SR_INDEX, SR01);
	i810_writeb(SR_DATA, val);
}


/**
 * i810fb_dram_off - turns off/on dram refresh
 * @mode: on or off
 *
 * DESCRIPTION:
 * Turns off DRAM refresh.  Must be off for only 2 vsyncs
 * before data becomes corrupt
 */
static void i810fb_dram_off(u8 mode)
{
	u8 val;
	val = i810_readb(DRAMCH);
	if (mode == OFF)
		val &= DRAM_OFF;
	else {
		val &= DRAM_OFF;
		val |= DRAM_ON;
	}
	i810_writeb(DRAMCH, val);
}

/**
 * i810fb_protect_regs - allows rw/ro mode of certain VGA registers
 * @mode: protect/unprotect
 *
 * DESCRIPTION:
 * The IBM VGA standard allows protection of certain VGA registers.  
 * This will  protect or unprotect them. 
 */
static void i810fb_protect_regs(int mode)
{
	u32 dataport;
	u32 indexport;
	u8 reg;

	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}
	i810_writeb(indexport, CR11);
	reg = i810_readb(dataport);
	if (mode == OFF)
		reg &= ~0x80;
	else 
		reg |= 0x80;
 		
	i810_writeb(indexport, CR11);
	i810_writeb(dataport, reg);
}

/**
 * i810fb_get_mem_freq - get RAM BUS frequency
 *
 * DESCRIPTION:
 * Determines if RAM is 100 or 133. Writes the result
 * to info->mem_freq
 */
static void i810fb_get_mem_freq(void)
{
	u8 reg;
	pci_read_config_byte(i810_info->dev, 0x50, &reg);
	reg &= FREQ_MASK;
	if (reg)
		i810_info->mem_freq = 133;
	else
		i810_info->mem_freq = 100;
}

/**
 * i810fb_lring_enable - enables/disables the ringbuffer
 * @mode: enable or disable
 *
 * DESCRIPTION:
 * Enables or disables the ringbuffer, effective enabling or
 * disabling the instruction/acceleration engine.
 */
static inline void i810fb_lring_enable(u32 mode)
{
	u32 tmp;
	tmp = i810_readl(LRING + 12);
	if (mode == OFF) 
		tmp &= ~1;
	else 
		tmp |= 1;
	wait_for_engine_idle();
	flush_cache();
	i810_writel(LRING + 12, tmp);
}       

/**
 * i810fb_init_ringbuffer - initialize the ringbuffer
 *
 * DESCRIPTION:
 * Initializes the ringbuffer by telling the device the
 * size and location of the ringbuffer.  It also sets 
 * the head and tail pointers = 0
 */
static void i810fb_init_ringbuffer(void)
{
	u32 tmp1, tmp2;
	
	wait_for_engine_idle();
	i810fb_lring_enable(OFF);
	i810_writel(LRING, 0);
	i810_writel(LRING + 4, 0);

	tmp2 = i810_readl(LRING + 8) & ~RBUFFER_START_MASK; 
	tmp1 = i810_info->lring_start_phys;
	i810_writel(LRING + 8, tmp2 | tmp1);

	tmp1 = i810_readl(LRING + 12);
	tmp1 &= ~RBUFFER_SIZE_MASK;
	tmp2 = (RINGBUFFER_SIZE - I810_PAGESIZE) & RBUFFER_SIZE_MASK;
	i810_writel(LRING + 12, tmp1 | tmp2);
	i810fb_lring_enable(ON);
}


/**
 * i810fb_restore_ringbuffer - restores saved ringbuffers
 * @lring: pointer to the ringbuffe structure
 */
static void i810fb_restore_ringbuffer(struct ringbuffer *lring)
{
	u32 tmp1, tmp2;
	wait_for_engine_idle();
	i810fb_lring_enable(OFF);
	
	i810_writel(LRING, 0);
	i810_writel(LRING + 4, 0);
	
	tmp1 = i810_readl(LRING + 8);
	tmp1 &= ~RBUFFER_START_MASK;
	tmp2 = lring->start;
	tmp2 &= RBUFFER_START_MASK;
	i810_writel(LRING + 8, tmp1 | tmp2);

	tmp1 = i810_readl(LRING + 12);
	tmp1 &= ~RBUFFER_SIZE_MASK;
	tmp2 = lring->size;
	tmp2 &= RBUFFER_SIZE_MASK;
	i810_writel(LRING + 12, tmp1 | tmp2);
	
	tmp1 = lring->size;
       	i810fb_lring_enable(tmp1 & 1);
}

/* Best to avoid floating point calculations in the kernel ...	*/
/**
 * i810_calc_dclk - calculates the P, M, and N values or a pixelclock value
 * @params: pointer to structure of video registervalues
 * @freq: pixclock to calculate
 *
 * DESCRIPTION:
 * Based on the formula Freq_actual = (4*M*Freq_ref)/(N^P)
 * Repeatedly computes the Freq until the actual Freq is equal to
 * the target Freq or until the loop count is zero.  In the latter
 * case, the actual frequency nearest the target will be used.
 * M, N, P registers are write to the params structure.
 */
static void i810_calc_dclk(struct mode_registers *params, u32 freq)
{
	u32 m_reg, n_reg, p_divisor, n_target_max;
	u32 m_target, n_target, p_target, val_min, f_best, n_best, m_best;
	u32 f_out, ref_freq, target_freq, diff, target_val;

	n_best = m_best = f_best = m_target = f_out = 0;
	ref_freq = 24000;
	/* Make target frequency divisible by 125000 */
	target_freq =  ((((freq + 999)/1000) + 124)/125) * 125;
	n_target_max = 50;
	target_val = 0xFFFF;
	val_min = 0xFFFF;

        /* Let's get a P divisor such that (16 * ref_freq)/(target_freq * p)  < 1 */
	p_divisor = 1;
	p_target = 0;
	while(((16 * ref_freq)/(target_freq * p_divisor)) && p_divisor < 32) {
		p_divisor <<= 1;
		++p_target;
	}

	n_reg = m_reg = n_target = 3;	
	while ((target_val) && (n_target < n_target_max)) {
		f_out = (4 * m_reg * ref_freq)/(n_reg * p_divisor);
		m_target = m_reg;
		n_target = n_reg;
		if (f_out <= target_freq) {
			++m_reg;
			diff = target_freq - f_out;
		}
		else {
			++n_reg;
			diff = f_out - target_freq;
		}
		if (diff > 10000) 
			target_val = 0xFFFF;
		else 
			target_val = (diff*10000)/target_freq;
		if (val_min > target_val) {
			val_min = target_val;
			f_best = f_out;
			n_best = n_target;
			m_best = m_target;
		}		 

	} 
	params->M = (m_best - 2) & 0x3FF;
	params->N = (n_best - 2) & 0x3FF;
	params->P = (p_target << 4);
	params->pixclock = f_out;

}

/**
 * I810fb_calc_monitor_limits - check if res is valid based on monitor specs
 * @info: pointer to info structure
 * @xres: horizontal resolution
 * @yres: vertical resolution
 *
 * DESCRIPTION:
 * Given xres, determine the vertical frequency. If vertical frequency is
 * lower than the specs of the monitor, exit with an error.
 *
 * RETURNS:
 * zero if invalid
 */ 
static int i810fb_calc_monitor_limits(struct fb_info *info, int xres, int yres)
{
        u32 htotal, vtotal, vfreq;
        
        htotal = 5 * xres / 4;
        vtotal = 21 * yres / 20;
        vfreq = info->monspecs.hfmax / vtotal;
        if (vfreq < info->monspecs.vfmin)
                return 0;
        return 1;
}

/**
 * i810fb_calc_mode_limits - determines pixelclock of a display mode
 * @info: pointer to info structure
 * @disp: pointer to display structure
 *
 * DESCRIPTION:
 * Calculates the required pixelclock. 
 *
 * USES:
 * i810_calc_dclk
 */                                                                                                                         
static int i810fb_calc_mode_limits(struct fb_info *info, struct display *disp)
{
	
	u32 htotal, vtotal, hfreq, vfreq, dclk, temp, rr_temp;

	htotal = 5 * (disp->var.xres) / 4;
	vtotal = 21 * (disp->var.yres) / 20;
	vfreq = info->monspecs.hfmax / vtotal;
	if (vfreq > info->monspecs.vfmax) 
		vfreq = info->monspecs.vfmax;
	hfreq = vfreq * vtotal;
	if (hfreq > info->monspecs.hfmax)
		hfreq = info->monspecs.hfmax;
       	if (vfreq < info->monspecs.vfmin)
	         return 0;
	dclk = hfreq * htotal;
	rr_temp = rr;
	if((rr_temp)) {  
		if (rr_temp < info->monspecs.vfmin)
			rr_temp = info->monspecs.vfmin; 
		if (rr_temp > vtotal * hfreq) 
			rr_temp = vtotal * hfreq;
		temp = rr_temp * htotal * vtotal;
		if ((hsync2) && (vsync2) && (temp > dclk) && 
		     hsync2 != hsync1) /* fixed frequency monitors */
			temp = dclk;
		dclk = temp;
	}

	i810_calc_dclk(&i810_info->mode_params, dclk); 
	return 1;
}
#ifdef CONFIG_FB_I810_NONSTD
/**
 * i810fb_calc_vga - calculates values for graphics registers
 * @info: pointer to info structure
 * @disp: pointer to display structure
 *
 * htotal: 1.25 x xres
 * vtotal: 1.10 x vres
 * hsync pulse: 3.5 - 4.0 % of horizontal scan time
 * vsync pulse: 150 divided by vertical tick time
 * left margin : 1/3 of hsync pulse
 * right margin: htotal - (xres + guard + hsync)
 * upper margin: 3
 * lower margin: vtotal - (yres + guard + vsync)
  
 * DESCRIPTION: 
 * The difference between htotal and xres should be greater than or equal than the sum of
 * the sync pulse and 2*gt.  If not, will use sync_min, and if
 * it still does not satisfy the criteria, will just increase htotal.  
 * To maintain the refresh rate, the pixclock can be recalculated in proportion
 * to the increase of the htotal. The hsync and vsync 
 * estimates are based on XFree86-Video-Timings-HOWTO by Eric S. Raymond.
 */
static void i810fb_calc_vga(struct fb_info *info, struct display *disp)
{
	int n, htotal, xres, yres, vtotal, htotal_new;
	int sync_min, sync_max, dclk;
	int temp, vert_tick, gt, blank_s, blank_e;
	struct mode_registers *params = &i810_info->mode_params;


	xres = disp->var.xres;
	yres = disp->var.yres;
	htotal =  5 * xres/4;
	vtotal = 21 * yres/20;

	/* Horizontal */
	sync_min = ((params->pixclock/1000) * 7) >> 1;
	sync_max = (params->pixclock/1000) << 2;
	sync_min = (sync_min + 7) & ~7;
	sync_max = sync_max & ~7;
	if (sync_max > 63 << 3)
		sync_max = 63 << 3;
	temp = ((params->pixclock/1000) * 19)/5;
	temp = (temp+7)  & ~7;
	if (temp < sync_min) temp = sync_min;
	if (temp > sync_max) temp = sync_max;	
	gt = (temp>>2) & ~7;
	if (htotal < xres + (gt << 1) + temp) {
		htotal_new = xres + (gt << 1) + temp;
		dclk = (params->pixclock * htotal_new) / htotal;
		i810_calc_dclk(params, dclk * 1000);
		htotal = htotal_new;
	}		
	/* htotal */
	n = (htotal >> 3) - 5;
	params->cr00 =  (u8) n;
	params->cr35 = (u8) ((n >> 8) & 1);
	
	/* xres */
	params->cr01 = (u8) ((xres >> 3) - 1);

	/* FIXME: hblank  -- presently unset */
	blank_s = ~1;
	blank_e = ~1;
/*	if (blank_e - blank_s > 63*8)
		blank_s = blank_e - 63*8;
*/	params->cr02 = (u8) (blank_s >> 3);
	n = blank_e >> 3;
	params->cr03 = (u8) (n & 0x1F);
	params->cr05 = (u8) ((n & ~0x20) << 2);
	params->cr39 = (u8) ((n >> 6) & 1);

	/* hsync */
	params->cr04 = (u8) ((xres + gt) >> 3);
	n = (int) params->cr05;
	n |= ((xres + gt + temp) >> 3) & 0x1F;
	params->cr05 = (u8) n;
	
	/* Setup var_screeninfo */
	disp->var.pixclock = 1000000000/params->pixclock;
	disp->var.left_margin = (htotal - (xres + temp + gt));
	disp->var.right_margin = gt;
	disp->var.hsync_len = temp;
	disp->var.sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT | FB_SYNC_ON_GREEN;
	disp->var.vmode = FB_VMODE_NONINTERLACED;
	
       	/* Vertical */
       	vert_tick = htotal/(params->pixclock/1000);
	if (!(vert_tick))
		vert_tick = 1;
	temp = 150/vert_tick;
	if (temp > 15)
		temp = 15;
	if (!temp)
		temp = 1;

	/* vtotal */
	n = vtotal - 2;
	gt = 3;
	params->cr06 = (u8) (n & 0xFF);
	params->cr30 = (u8) ((n >> 8) & 0x0F);
	params->cr07 = 0; /* not used */	

	/* vsync */ 
	n = yres + gt;
	params->cr10 = (u8) (n & 0xFF);
	params->cr32 = (u8) ((n >> 8) & 0x0F);
	params->cr11 = (u8) ((yres + gt + temp) & 0x0F);

	/* yres */
	n = yres - 1;
	params->cr12 = (u8) (n & 0xFF);
	params->cr31 = (u8) ((n >> 8) & 0x0F);
	
	/* vblank */
	blank_s = yres;
	blank_e = yres + gt;
	if (blank_e - blank_s > 127)
		blank_s = blank_e - 127;
	n = blank_s;
	params->cr15 = (u8) (n & 0xFF);
	params->cr33 = (u8) ((n >> 8) & 0x0F);

	n = blank_e;
	params->cr16 = (u8) (n & 0xFF);
	params->cr09 = 0;	

	/* var_screeninfo */
	disp->var.upper_margin = vtotal - (yres + gt + temp);
	disp->var.lower_margin = gt;
	disp->var.vsync_len = temp;
}	

#else /* standard video modes */
static void i810fb_calc_vga(struct fb_info *info, struct display *disp)
{ 
	struct mode_registers *params = &i810_info->mode_params;
	struct mode_registers *std_params = std_modes;

	u32 diff = 0, diff_best = 0xFFFFFFFF, i = 0, i_best = 0; 
	u32 res, total;
	u8 xres;
	
	xres = (u8) ((disp->var.xres >> 3) - 1);
	while  (std_params[i].pixclock) { 
		if (std_params[i].cr01 == xres) { 
			diff = (std_params[i].pixclock > params->pixclock) ?
			       std_params[i].pixclock - params->pixclock :
			       params->pixclock - std_params[i].pixclock;
			       
			if (diff < diff_best) {	 
		    		i_best = i;
		    		diff_best = diff;
			}
		}
		++i;
	}
	*params = std_params[i_best];
	/* Setup var_screeninfo */
	total = ((params->cr00 | (params->cr35 & 1) << 8) + 3) << 3;
	res = disp->var.xres;

	disp->var.pixclock = 1000000000/params->pixclock;
	disp->var.right_margin = (params->cr04 << 3) - res;
	disp->var.hsync_len = ((params->cr05 & 0x1F) - (params->cr04 & 0x1F)) << 3;
	disp->var.left_margin = (total - (res + disp->var.right_margin + disp->var.hsync_len));
	disp->var.sync = FB_SYNC_HOR_HIGH_ACT | FB_SYNC_VERT_HIGH_ACT | FB_SYNC_ON_GREEN;
	disp->var.vmode = FB_VMODE_NONINTERLACED;

	total = ((params->cr06 | (params->cr30 & 0x0F)  << 8)) + 2;
	res = disp->var.yres;
	disp->var.lower_margin = (params->cr10 | (params->cr32 & 0x0F) << 8) - res;
	disp->var.vsync_len = (params->cr11 & 0x0F) - (disp->var.lower_margin & 0x0F);
	disp->var.upper_margin = total - (res + disp->var.lower_margin + disp->var.vsync_len);
	
}
#endif /* CONFIG_FB_I810_NONSTD */

/**
 * i810fb_load_pll
 * @disp: pointer to display structure
 *
 * DESCRIPTION:
 * Loads the P, M, and N registers.  
 * USES:
 * info->mode_params
 */
static void i810fb_load_pll(struct display *disp)
{
 	u32 tmp1, tmp2;
	
	tmp1 = i810_info->mode_params.M | i810_info->mode_params.N << 16;
	tmp2 = i810_readl(DCLK_2D);
	tmp2 &= ~MN_MASK;
	i810_writel(DCLK_2D, tmp1 | tmp2);
	
	tmp1 = i810_info->mode_params.P;
	tmp2 = i810_readl(DCLK_0DS);
	tmp2 &= ~(P_OR << 16);
	i810_writel(DCLK_0DS, (tmp1 << 16) | tmp2);

	i810_writeb(MSR_WRITE, 0xC8);

}

/**
 * i810fb_load_vga - load standard VGA registers
 *
 * DESCRIPTION:
 * Load values to VGA registers
 */
static void i810fb_load_vga(void)
{	
	u32 indexport, dataport;
      
	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}
	i810_writeb(indexport, CR00);
	i810_writeb(dataport, i810_info->mode_params.cr00);
	i810_writeb(indexport, CR01);
	i810_writeb(dataport, i810_info->mode_params.cr01);
	i810_writeb(indexport, CR02);
	i810_writeb(dataport, i810_info->mode_params.cr02);
	i810_writeb(indexport, CR03);
	i810_writeb(dataport, i810_info->mode_params.cr03);
	i810_writeb(indexport, CR04);
	i810_writeb(dataport, i810_info->mode_params.cr04);
	i810_writeb(indexport, CR05);
	i810_writeb(dataport, i810_info->mode_params.cr05);
	i810_writeb(indexport, CR06);
	i810_writeb(dataport, i810_info->mode_params.cr06);
	i810_writeb(indexport, CR09);
	i810_writeb(dataport, i810_info->mode_params.cr09);
	i810_writeb(indexport, CR10);
	i810_writeb(dataport, i810_info->mode_params.cr10);
	i810_writeb(indexport, CR11);
	i810_writeb(dataport, i810_info->mode_params.cr11);
	i810_writeb(indexport, CR12);
	i810_writeb(dataport, i810_info->mode_params.cr12);
	i810_writeb(indexport, CR15);
	i810_writeb(dataport, i810_info->mode_params.cr15);
	i810_writeb(indexport, CR16);
	i810_writeb(dataport, i810_info->mode_params.cr16);
}


/**
 * i810fb_load_vgax - load extended VGA registers
 *
 * DESCRIPTION:
 * Load values to extended VGA registers
 */

static void i810fb_load_vgax(void)
{
	u32 indexport, dataport;

	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}
	i810_writeb(indexport, CR30);
	i810_writeb(dataport, i810_info->mode_params.cr30);
	i810_writeb(indexport, CR31);
	i810_writeb(dataport, i810_info->mode_params.cr31);
	i810_writeb(indexport, CR32);
	i810_writeb(dataport, i810_info->mode_params.cr32);
	i810_writeb(indexport, CR33);
	i810_writeb(dataport, i810_info->mode_params.cr33);
	i810_writeb(indexport, CR35);
	i810_writeb(dataport, i810_info->mode_params.cr35);
	i810_writeb(indexport, CR39);
	i810_writeb(dataport, i810_info->mode_params.cr39);
}
#ifdef CONFIG_FB_I810_NONSTD
/**
 * i810fb_get_watermark - gets watermark
 * @var: pointer to fb_var_screeninfo
 *
 * DESCRIPTION:
 * Load values to graphics registers
 * 
 * RETURNS:
 * watermark
 */

static u32 i810fb_get_watermark(struct fb_var_screeninfo *var)
{
	struct wm_info *wmark = 0;
	u32 i, pixclock, wm_best = 0, min, diff;

	if (i810_info->mem_freq == 100) {
		switch (var->bits_per_pixel) { 
		case 8:
			wmark = i810_wm_8_100;
			break;
		case 16:
			wmark = i810_wm_16_100;
			break;
		case 24:
			wmark = i810_wm_24_100;
		}
	}	
	else {
		switch(var->bits_per_pixel) {
		case 8:
			wmark = i810_wm_8_133;
			break;
		case 16:
			wmark = i810_wm_16_133;
			break;
		case 24:
			wmark = i810_wm_24_133;
		}
	}

	pixclock = i810_info->mode_params.pixclock/1000;
	min = ~0;
	for (i = 0; wmark[i].freq; i++) {
		if (pixclock <= wmark[i].freq) 
			diff = wmark[i].freq - pixclock;
		else 
			diff = pixclock - wmark[i].freq;
		if (diff < min) {
			wm_best = wmark[i].wm;
			min = diff;
		}
	}
	return wm_best;		
}	

#else
static u32 i810fb_get_watermark(struct fb_var_screeninfo *var)
{
	struct mode_registers *params = &i810_info->mode_params;
	u32 wmark = 0;
	
	if (i810_info->mem_freq == 100) {
		switch (var->bits_per_pixel) {
		case 8:
			wmark = params->bpp8_100;
			break;
		case 16:
			wmark = params->bpp16_100;
			break;
		case 24:
			wmark = params->bpp24_100;
		}
	}
	else {					
		switch (var->bits_per_pixel) {
		case 8:
			wmark = params->bpp8_133;
			break;
		case 16:
			wmark = params->bpp16_133;
			break;
		case 24:
			wmark = params->bpp24_133;
		}
	}
	return wmark;
}	

#endif /* CONFIG_FB_I810_NONSTD */

/**
 * i810fb_load_2d - load grahics registers
 * @wm: watermark
 *
 * DESCRIPTION:
 * Load values to graphics registers
 */

static void i810fb_load_2d(u32 wm)
{
	u32 tmp;

  	i810_writel(FW_BLC, wm); 
	tmp = i810_readl(PIXCONF);
	tmp |= 1;
	i810_writel(PIXCONF, tmp);
}	

/**
 * i810fb_hires - enables high resolution mode
 */
static void i810fb_hires(void)
{
	u32 indexport, dataport;
	u8 val;
	
	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}


	i810_writeb(indexport, CR80);
	val = i810_readb(dataport);
	i810_writeb(indexport, CR80);
	i810_writeb(dataport, val | 1);
}

/**
 * i810fb_load_fbstart - loads pointer of the framebuffer
 * @offset: where to start display from the start of framebuffer
 */
static void i810fb_load_fbstart(int offset)
{
	u32 addr, indexport, dataport, vidmem;
	u8 val;

	vidmem = i810_info->fb_start_phys + offset;
	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}
	
	addr = ((vidmem + 2) >> 2) & 0xFF;
	i810_writeb(indexport, CR0D);
	i810_writeb(dataport, (u8) addr);
	
	addr = (vidmem >> 10) & 0xFF;
	i810_writeb(indexport, CR0C);
	i810_writeb(dataport, (u8) addr);
	
	addr = (vidmem >> 18) & 0x3F;
	i810_writeb(indexport, CR40);
	val = i810_readb(dataport) & ~0x3F;
	i810_writeb(indexport, CR40);
	i810_writeb(dataport, (u8) addr | val);
	
	addr = vidmem >> 24;
	i810_writeb(indexport, CR42);
	i810_writeb(dataport, (u8) addr);
	
	i810_writeb(indexport, CR40);
	val = i810_readb(dataport) | 0x80;
	i810_writeb(indexport, CR40);
	i810_writeb(dataport, val);
}
	
/**
 * i810fb_load_pitch - loads the characters per line of the display
 * @var: pointer to fb_var_screeninfo
 *
 * DESCRIPTION:
 * Loads the characters per line
 */	
static void i810fb_load_pitch(struct fb_var_screeninfo *var)
{

	u32 tmp, indexport, dataport;
	u32 line_length = get_line_length(var->xres_virtual, var->bits_per_pixel) >> 3;
	u8 val;
			
	if (is_cga()) {
		indexport = CR_INDEX_CGA;
		dataport = CR_DATA_CGA;
	}
	else {
		indexport = CR_INDEX_MDA;
		dataport = CR_DATA_MDA;
	}

	i810_writeb(SR_INDEX, SR01);
	val = i810_readb(SR_DATA);
	val &= 0xE0;
	val |= 0x01;
	i810_writeb(SR_INDEX, SR01);
	i810_writeb(SR_DATA, val);

	tmp = line_length & 0xFF;
	i810_writeb(indexport, CR13);
	i810_writeb(dataport, (u8) tmp);
	
	tmp = line_length >> 8;
	i810_writeb(indexport, CR41);
	val = i810_readb(dataport) & ~0x0F;
	i810_writeb(indexport, CR41);
	i810_writeb(dataport, (u8) tmp | val);
}

/**
 * i810fb_load_color - loads the color depth of the display
 * @var: pointer to fb_var_screeninfo
 *
 * DESCRIPTION:
 * Loads the color depth of the display and the graphics engine
 */
static void i810fb_load_color(struct fb_var_screeninfo *var)
{
	u32 reg1;
	u16 reg2;
	reg1 = i810_readl(PIXCONF) & ~0xF0000;
	reg2 = i810_readw(BLTCNTL) & ~0x30;
	switch(var->bits_per_pixel) {
	case 1:
	case 2:
	case 4:
		break;
	case 8:
		reg1 |= 0x20000;
		break;
	case 16: 
		reg1 |= 0x50000;
		reg2 |= 0x10;
		break;
	case 24:
		reg1 |= 0x60000;
		reg2 |= 0x20;
	}
	reg1 |= 0x8000;  
	i810_writel(PIXCONF, reg1);
	i810_writew(BLTCNTL, reg2);
}


/**
 * i810fb_load_regs - loads all registers for the mode
 * @disp: pointer to display structure
 * 
 * DESCRIPTION:
 * Loads registers
 */
 
static void i810fb_load_regs(struct display *disp)
{
	i810fb_lring_enable(OFF);
	i810fb_screen_off(OFF);
	i810fb_protect_regs(OFF);
	i810fb_dram_off(OFF);
	i810fb_load_pll(disp);
	i810fb_load_vga();
	i810fb_load_vgax();
	i810fb_dram_off(ON);	
	i810fb_load_2d(i810fb_get_watermark(&disp->var));
	i810fb_hires();
	i810fb_screen_off(ON);
	i810fb_protect_regs(ON);
	i810fb_lring_enable(ON);
}


/**
 * i810fb_set_disp - initialize display
 * @disp: pointer to display structure
 */
static void i810fb_set_disp(struct display *disp, int con, struct fb_info *info)
{
	i810fb_load_pitch(&disp->var); 
	i810fbcon_updatevar(con, info);	
	
}
/*
 * Hardware Cursor Routines
 */

/**
 * i810fb_load_cursor_image - create cursor bitmap
 * @p: pointer to display structure
 * 
 * DESCRIPTION:
 * Creates the cursor bitmap using the 2bpp, 2plane
 * AND/XOR format of the i810.  For each line of cursor 
 * data, the first half is composed of the monochrome bitmap
 * image, and the second half is for the XOR/AND bit.
 * The cursor is a rectangular block, a fontwidth wide and
 * 2 scanlines length.  
 */
static void i810fb_load_cursor_image(struct display *p)
{
 	int i, j, h, w;
 	u8 *addr;
 	
	h = fontheight(p);
	w = fontwidth(p) >> 3;
	(u32)addr = i810_info->cursor_start_virtual;
	for (i = 64; i--; ) {
		for (j = 0; j < 8; j++) {
			addr[j] = 0xFF;               /* transparent - yes */
			addr[j+8] = 0x00;             /* use background */
		}	
		addr +=16;
	}
	
	(u32)addr = i810_info->cursor_start_virtual + (16 * (h-1));	
	for (i = 0; i < 2; i++) {
		addr[0] = 0x00;                      /* transparent - no */
		addr[8] = 0xFF;                      /* use foreground */
		addr +=16;
	}
}		 			
 			 	
/**
 * i810fb_set_cursor_color - set the cursor CLUT
 * @p: pointer to display structure
 *
 * DESCRIPTION:
 * The i810 has two DACS, the standard 256-index CLUT and the
 * alternate 8-index CLUT. The alternate CLUT is where the
 * cursor gets the color information.
 */
static void i810fb_set_cursor_color(struct display *p)
{
	u8 temp, r, g, b;
	
	temp = i810_readb(PIXCONF1);
	i810_writeb(PIXCONF1, temp | EXTENDED_PALETTE);

	i810_writeb(DACMASK, 0xFF); 
        i810_writeb(CLUT_INDEX_WRITE, 0x04);	

        r = (u8) ((i810_cfb24[p->bgcol] >> 16) & 0xFF);
	g = (u8) ((i810_cfb24[p->bgcol] >> 8) & 0xFF);
	b = (u8) ((i810_cfb24[p->bgcol]) & 0xFF);
	i810_writeb(CLUT_DATA, r);
	i810_writeb(CLUT_DATA, g);
	i810_writeb(CLUT_DATA, b);

	r = (u8) ((i810_cfb24[p->fgcol] >> 16) & 0xFF);
	g = (u8) ((i810_cfb24[p->fgcol] >> 8) & 0xFF);
	b = (u8) ((i810_cfb24[p->fgcol]) & 0xFF);
	i810_writeb(CLUT_DATA, r);
	i810_writeb(CLUT_DATA, g);
	i810_writeb(CLUT_DATA, b);

	temp = i810_readb(PIXCONF1);
	i810_writeb(PIXCONF1, temp & ~EXTENDED_PALETTE);
	
}		 
static void i810fb_set_cursor(struct display *disp)
{
	i810fb_set_cursor_color(disp); 
	i810fb_load_cursor_image(disp);
}	
	
/**
 * i810fb_enable_cursor - show or hide the hardware cursor
 * @mode: show (1) or hide (0)
 *
 * Description:
 * Shows or hides the hardware cursor
 */
static void i810fb_enable_cursor(int mode)
{
	u32 temp;
	
	temp = i810_readl(PIXCONF);
	if (mode == ON)
		temp |= CURSOR_ENABLE_MASK;
	else
		temp &= ~CURSOR_ENABLE_MASK;
	i810_writel(PIXCONF, temp);
}
			
/**
 * i810_cursor_timer_handler - cursor timer handler
 * @data: arbitrary data which points to cursor structure
 *
 * DESCRIPTION:
 * The cursor timer handler.  This handles the blinking
 * of the cursor
 */
static void i810_cursor_timer_handler(unsigned long data)
{ 
	struct cursor_data *cur;
	(u32) cur = data;
	if (cur->cursor_enable && !cur->blink_count) {
		if (cur->cursor_show) {
			i810fb_enable_cursor(OFF);
			cur->cursor_show = 0; 
		}
		else {
			i810fb_enable_cursor(ON);
			cur->cursor_show = 1;
		}
		cur->blink_count = cur->blink_rate;
	}	
	--cur->blink_count;
	cur->timer->expires = jiffies + (HZ/100);
	add_timer(cur->timer);
}				


/**
 * i810fb_init_cursor - initializes the cursor
 *
 * DESCRIPTION:
 * Initializes the cursor registers
 */
static void i810fb_init_cursor(void)
{
	i810fb_enable_cursor(OFF);
	i810_writel(CURBASE, i810_info->cursor_start_phys);
	i810_writew(CURCNTR, COORD_ACTIVE | CURSOR_MODE_64_TRANS);
}	


/* 
 * GART services
 */

static inline void flush_cache(void)
{
#if defined(__i386__)
	asm volatile ("wbinvd":::"memory");
#endif	
}

/**
 * i810_update_flag - updates usage of AGP
 *
 * DESCRIPTION:
 * Updates the usage count and other parameters
 * of the AGP
 *
 * RETURNS:
 * 0 if failed (maximum usage count);	
 */
static int i810_update_flag(void)
{
	struct pgtbl_record *i810_master = i810_info->i810_gtt.i810_master;
	int i, flag_start;
	u32 record_size;
	i = 0;
	flag_start = curr_pg_flag;
	++curr_pg_flag;
	if (curr_pg_flag == 256) 
		curr_pg_flag = 1;

	record_size = 1024 * (1 << i810_info->page_order);
	
	while ((flag_start != curr_pg_flag) && (i < record_size) && 
		(i810_master[i].flag != (u8) curr_pg_flag)) {
		if (i810_master[i].flag == (u8) curr_pg_flag) {
			++curr_pg_flag;
			i = 0;
		}
		else if (curr_pg_flag >= 256) { 
			curr_pg_flag = 1;
			i = 0;
		}	       
	        ++i;
	}
	if (flag_start == curr_pg_flag) {
		curr_pg_flag = -1;
		return 0;
	}
	return 1;
}

/**
 * i810_find_chunk - find free contiguous space in the GART
 * @size: size of space in 4096K bytes
 *
 * DESCRIPTION:
 * Finds a free contiguous space in the GART
 *
 * RETURNS:
 * offset to the GART, 0 if fail
 */
static int i810_find_chunk(u32 size)
{
	struct pgtbl_record *i810_master = i810_info->i810_gtt.i810_master;
	u32 offset = 0, n = 0, cur_size = 0, rec_size; 
	rec_size = 1024 * (1 << i810_info->page_order);
	do {	
		offset = n;
		while ((cur_size < size) && (n < rec_size) && (!(i810_master[n].flag))) 
			++cur_size;
	        if (cur_size < size) 
	        	cur_size = 0;
		n++;
	} while ((cur_size < size) && (n < rec_size));	
	if (cur_size < size) return -1;
	return (int) offset;	
}

/**
 * i810_allocate_page - get a free page from the kernel
 *
 * RETURNS:
 * pointer to new page, 0 if fail
 */
static unsigned long i810_allocate_page(void)
{
	void *new;
	
	new = (void *) __get_free_page(GFP_KERNEL);
	if (new == NULL) return 0;
	atomic_inc(&virt_to_page(new)->count);
	set_bit(PG_locked, &virt_to_page(new)->flags);
	return (unsigned long) new;
}

/**
 * i810_free_page - frees a page
 * @ptr: pointer to page
 */
static void i810_free_page(void * ptr)
{
	if (ptr == NULL) return;
	atomic_dec(&virt_to_page(ptr)->count);
	clear_bit(PG_locked, &virt_to_page(ptr)->flags);
        wake_up(&virt_to_page(ptr)->wait);
      	free_page((unsigned long) ptr);
}
	
/**
 * i810_free_memory - frees GART memory
 * @pgtbl: pointer to the GART table
 * @type: type of AGP memory
 * @offset: offset of memory to free
 *
 * DESCRIPTION:
 * Frees a continugous segment of memory from the GART
 * 
 * RETURNS:
 * nonzero if success, zero if fail
 */ 
static int i810_free_memory(u32 *pgtbl, u32 type, u32 offset)
{
	struct pgtbl_record *i810_master = i810_info->i810_gtt.i810_master;
	int i, flag;
	void* ptr;
	if (type == AGP_CACHE)
		return 1;
	if(!(flag = (u32) i810_master[offset].flag)) return 0;
	flush_cache();
	for (i = offset; i810_master[i].flag == flag && i < 16 * 1024; i++) {
		ptr = (void *) phys_to_virt(pgtbl[i]);
		i810_free_page(ptr);
		i810_writel(GTT + (i << 2), i810_info->i810_gtt.pgtbl_scratch | type | PG_ENABLE_MASK);
		i810_master[i].flag = 0;
	}
	flush_cache();
	return 1;
}

/**
 * i810_allocate_memory - allocate GART memory
 * @pgtbl: pointer to the GART table
 * @pgsize: size of memory to allocate, in pages
 * @type: type of AGP memory
 * @offset: offset of memory to free
 * DESCRIPTION:
 * Allocates a continugous segment of memory from the GART
 * 
 * RETURNS:
 * nonzero if success, pointer to memory if type is 
 * AGP_PHYSICAL, zero if fail, or if type is AGP_CACHE
 */ 
static unsigned long i810_allocate_memory(u32 *pgtbl, u32 pgsize, u32 type, u32 offset)
{
	struct pgtbl_record *i810_master = i810_info->i810_gtt.i810_master;
	void * new;
	u32 i, *temp, physical;

	if (type == AGP_CACHE) return 0;
	if (type == AGP_PHYSICAL && pgsize != 1) return 0; 

	/* create temporary storage */
	if(!(temp = (u32 *) vmalloc(pgsize << 2)))
		return 0;
	physical = 1;
	for (i = 0; i < pgsize; i++) {
		new = (void *) i810_allocate_page();
		if ((new == NULL) || i >= (16*1024 - offset) || (i810_master[offset + i].flag != 0)) {
			i810_free_memory(pgtbl, offset, type);
			vfree(temp);
			return 0;
		}
		i810_master[offset + i].flag = (u8) curr_pg_flag;
		temp[i] = (u32)  virt_to_phys((void *) new); 
	}
	flush_cache();
	for (i = 0; i < pgsize; i++) 
		i810_writel(GTT + ((offset + i) << 2), temp[i] | PG_ENABLE_MASK);
	flush_cache();	
	if (!i810_update_flag()) {
		i810_free_memory(pgtbl, offset, type);
		vfree(temp);
		return 0;
	}	
	if (type == AGP_PHYSICAL)
		physical = temp[0]; 
	vfree(temp);
	return physical;	
}

/**
 * i810_create_pgtbl - creates a GART table
 *
 * DESCRIPTION:
 * Create an i810 specific GART table
 *
 * RETURNS:
 * pointer to table, or zero if fail
 */
static  u32 i810_create_pgtbl(void)
{
	char *tbl; 
	char *tbl_end;
	struct page *page;
	int i;
	
	tbl = (char *) __get_free_pages(GFP_KERNEL, 4);
	if (tbl == NULL) {
		return 0;
	}
	tbl_end = tbl + GTT_SIZE - 1;
	for (page = virt_to_page(tbl); page <= virt_to_page(tbl_end); page++) 
		set_bit(PG_reserved, &page->flags);
	flush_cache();
	
	for (i = 0; i < (GTT_SIZE >> 12); i++) 
		((u32 *) tbl)[i] = i810_info->i810_gtt.pgtbl_scratch | PG_ENABLE_MASK; 
	return (u32) tbl;	
}
	
/**
 * i810_destroy_pgtbl - destroy a GART table
 * @pgtbl: pointer to table to destroy
 *
 * DESCRIPTION:
 * Destroy an i810 specific GART table
 *
 */
static void i810_destroy_pgtbl(char* pgtbl)
{
	struct page *page;
	char* tbl_end;
	int i;
	
	for (i = 0; i < (GTT_SIZE >> 12); i++)
		((u32 *) pgtbl)[i] = 0;
	tbl_end = pgtbl + GTT_SIZE - 1;
	for (page = virt_to_page(pgtbl); page <= virt_to_page(tbl_end); page++) 
	        clear_bit(PG_reserved, &page->flags);
        free_pages((unsigned long) pgtbl, 4);
}
	
static void i810_enable_pgtbl(void)
{
	u32 temp;
	temp = i810_readl(PGTBL_CTL);
	temp |= 1;
	i810_writel(PGTBL_CTL, temp);
}

/**
 * i810_disable_pgtbl - disables the GART table
 *
 * DESCRIPTION:
 * Unsets the enable bit of the Control Register
 */
static void i810_disable_pgtbl(void)
{
	u32 temp;
	temp = i810_readl(PGTBL_CTL);
	temp &= ~1;
	wait_for_engine_idle();
	i810_writel(PGTBL_CTL, temp);
}

/**
 * i810_bind_pgtbl - writes GART table pointer to Control Register
 * @pgtbl_phys: pointer to physical address of table
 * @pgtbl_virtual: pointer to virtual address of table
 *
 * DESCRIPTION:
 * Writes the address to the Control Register, effectively 
 * enabling GTT usage of the i810
 */
static void i810_bind_pgtbl(u32* pgtbl_phys, u32* pgtbl_virtual)
{
	if ((i810_readl(PGTBL_CTL) & 1))
		i810_disable_pgtbl();
	i810_writel(PGTBL_CTL, (u32) pgtbl_phys);
	i810_enable_pgtbl();
	flush_cache();
}

/**
 * i810_unbind_pgtbl - writes a zero to Control Register
 *
 * DESCRIPTION:
 * Writes a zero to the Control Register, effectively 
 * disabling GTT usage of the i810
 */
static void i810_unbind_pgtbl(void)
{
	if ((i810_readl(PGTBL_CTL) & 1))
		i810_disable_pgtbl();
	i810_writel(PGTBL_CTL, 0);
}


/**
 * not_safe - determines if it's safe to use the graphics device
 *
 * DESCRIPTION:
 * Checks if the current GART table pointer is equal to the 
 * valid GART table pointer
 *
 * RETURNS:
 * a nonzero if true
 */
static inline int not_safe(void)
{
	u32 temp1, temp2;
	temp1 = i810_readl(PGTBL_CTL) & ~0xFFF;
	temp2 = (u32) i810_info->i810_gtt.i810_pgtbl_phys & ~0xFFF;
	if (temp1 == temp2)
		return 0;
	return 1;
}		 


/**
 * find_agpgart - detects if the agpgart module is loaded
 *
 * DESCRIPTION:  
 * This is ugly! Check if agpgart is loaded by asking the kernel for unique
 * inter_module data of agpgart, "drm_agp". Since, this will increase
 * the module use count of agpgart, we need to release it before returning
 *
 * RETURNS:
 * nonzero if true	
 */
static inline int find_agpgart(void)
{
	if ((int)(inter_module_get("drm_agp"))) {
		/* release data, we don't need it */
		inter_module_put("drm_agp");
		return 1;
	}	
	return 0;
}		

static int i810fbcon_switch(int con, struct fb_info *info)
{
	u32 temp1, temp2;
	if (con < xcon) {
		temp1 = i810_readl(PGTBL_CTL) & ~0xFFF;
		temp2 = (u32) i810_info->i810_gtt.i810_pgtbl_phys & ~0xFFF;
		if (temp1 != temp2) {
			if (find_agpgart())
				xgart = i810_readl(PGTBL_CTL);
			else
				xgart = 0;
			if (accel) i810fb_lring_enable(OFF);		
			i810_bind_pgtbl(i810_info->i810_gtt.i810_pgtbl_phys,
					i810_info->i810_gtt.i810_pgtbl_virtual);
			if (accel) {
				xring_state.head = i810_readl(LRING);
				xring_state.tail = i810_readl(LRING + 4);
				xring_state.start = i810_readl(LRING + 8);
				xring_state.size = i810_readl(LRING + 12);
				i810fb_init_ringbuffer();
			}
		}		
	}
	else if (xgart)  {
		if (accel) {
			i810fb_lring_enable(OFF);
			i810fb_restore_ringbuffer(&xring_state);
		}	
		i810_writel(PGTBL_CTL, xgart);
		wait_for_engine_idle();
		flush_cache();
		currcon = con;
		return 0;
	}
	if (!lring_valid)
		i810fb_init_ringbuffer();
 	i810fb_set_var(&fb_display[con].var, con, info);
	currcon = con;
    	return 0;
}

    /*
     *  Update the `var' structure (called by fbcon.c)
     */
static int i810fbcon_updatevar(int con, struct fb_info *info)
{
  	struct display *p = &fb_display[con];
	int yoffset, xoffset, depth, total = 0;
	
	yoffset = p->var.yoffset;
	xoffset = p->var.xoffset;
	depth = p->var.bits_per_pixel >> 3;

	if (p->var.xres + xoffset <= p->var.xres_virtual)
		total += xoffset * fontwidth(p) * depth;
	if (p->var.yres + yoffset <= p->var.yres_virtual)
		total += yoffset * p->next_line;
	i810fb_load_fbstart(total);
 
	return 0;
}

    /*
     *  Blank the display.
     */
static void i810fbcon_blank(int blank, struct fb_info *info)
{
	int mode = 0, temp;
	
	temp = i810_readl(HVSYNC);
	
	switch(blank) {
	case VESA_NO_BLANKING:
		mode = POWERON;
		break;
	case VESA_VSYNC_SUSPEND:
		mode = STANDBY;
		break;
	case VESA_HSYNC_SUSPEND:
		mode = SUSPEND;
		break;
	case VESA_POWERDOWN:
		mode = POWERDOWN;
	}
	i810_writel(HVSYNC, mode);
	
}

static u_long get_line_length(int xres_virtual, int bpp)
{
	u_long length;
	
	length = xres_virtual*bpp;
	length = (length+31)&-32;
	length >>= 3;
	return(length);
}

static void i810fb_encode_fix(struct fb_fix_screeninfo *fix,
			   struct fb_var_screeninfo *var)
{
    	memset(fix, 0, sizeof(struct fb_fix_screeninfo));
    	strcpy(fix->id, i810fb_name);
    	fix->smem_start = i810_info->fb_start_phys;
    	fix->smem_len = i810_info->fb_size;
    	fix->type = FB_TYPE_PACKED_PIXELS;
    	fix->type_aux = 0;
    	switch (var->bits_per_pixel) {
	case 1:
	    	fix->visual = FB_VISUAL_MONO01;
	    	break;
	case 2:
	case 4:
	case 8:
	    	fix->visual = FB_VISUAL_PSEUDOCOLOR;
	    	break;
	case 16:
	case 24:
	case 32:
	    	fix->visual = FB_VISUAL_TRUECOLOR;
	    	break;
    }
    	fix->ywrapstep = 0;
    	fix->xpanstep = 0;
    	fix->ypanstep = 1;
    	fix->line_length = get_line_length(var->xres_virtual, var->bits_per_pixel);
	fix->mmio_start = i810_info->mmio_start_phys;
	fix->mmio_len = MMIO_SIZE;
	fix->accel = FB_ACCEL_I810;
		
}

static void set_color_bitfields(struct fb_var_screeninfo *var)
{
	switch (var->bits_per_pixel) {
	case 1:
	case 8:
		var->red.offset = 0;
		var->red.length = 8;
		var->green.offset = 0;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 16:	/* RGB 565 */
		var->red.offset = 11;
		var->red.length = 5;
		var->green.offset = 5;
		var->green.length = 6;
		var->blue.offset = 0;
		var->blue.length = 5;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 24:	/* RGB 888 */
		var->red.offset = 16;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 0;
		var->blue.length = 8;
		var->transp.offset = 0;
		var->transp.length = 0;
		break;
	case 32:	/* RGBA 8888 */
		var->red.offset = 0;
		var->red.length = 8;
		var->green.offset = 8;
		var->green.length = 8;
		var->blue.offset = 16;
		var->blue.length = 8;
		var->transp.offset = 24;
		var->transp.length = 8;
		break;
	}
	var->red.msb_right = 0;
	var->green.msb_right = 0;
	var->blue.msb_right = 0;
	var->transp.msb_right = 0;
	i810fb_load_color(var);
}


    /*
     *  Read a single color register and split it into
     *  colors/transparent. Return != 0 for invalid regno.
     */

static int i810fb_getcolreg(u_int regno, u_int *red, u_int *green, u_int *blue,
                         u_int *transp, struct fb_info *info)
{
	if (regno > 255)
		return 1;
 	i810_writeb(CLUT_INDEX_READ, (u8) regno);
 	*red = (u32) i810_readb(CLUT_DATA);
 	*green = (u32) i810_readb(CLUT_DATA);
 	*blue = (u32) i810_readb(CLUT_DATA);
    	*transp = 0;
    	return 0;
}


    /*
     *  Set a single color register. The values supplied are already
     *  rounded down to the hardware's capabilities (according to the
     *  entries in the var structure). Return != 0 for invalid regno.
     */

static int i810fb_setcolreg(u_int regno, u_int red, u_int green, u_int blue,
                         u_int transp, struct fb_info *info)
{
    	if (regno > 255)
		return 1;
	i810_writeb(CLUT_INDEX_WRITE, (u8) regno);
	i810_writeb(CLUT_DATA, (u8) red);
	i810_writeb(CLUT_DATA, (u8) green);
	i810_writeb(CLUT_DATA, (u8) blue); 	

      	return 0;
}


static void do_install_cmap(int con, struct fb_info *info)
{
	if (con != currcon)
		return;
	if (fb_display[con].cmap.len)
		fb_set_cmap(&fb_display[con].cmap, 1, i810fb_setcolreg, info);
	else
		fb_set_cmap(fb_default_cmap(1<<fb_display[con].var.bits_per_pixel), 1,
			    i810fb_setcolreg, info);
}


/**
 * i810fb_create_cursor - creates the cursor structure
 *
 * DESCRIPTION:
 * Creates the hardware cursor structure and starts the 
 * cursor timer
 *
 * RETURNS: nonzero if success
 */
static int __devinit i810fb_create_cursor(void) {
	i810_info->cursor.timer = kmalloc(sizeof(struct timer_list), GFP_KERNEL); 
	if (!i810_info->cursor.timer) return 0;
	init_timer(i810_info->cursor.timer);
	i810_info->cursor.blink_rate = 40;
	i810_info->cursor.blink_count = i810_info->cursor.blink_rate; 
	i810_info->cursor.cursor_show = 0;
	i810_info->cursor.cursor_enable = 0;
	i810_info->cursor.timer->data = (u32) &i810_info->cursor;
	i810_info->cursor.timer->expires = jiffies + (HZ/100);
	i810_info->cursor.timer->function = i810_cursor_timer_handler;
	return 1;
}

/**
 * i810fb_init_monspecs
 * @fb_info: pointer to device specific info structure
 *
 * DESCRIPTION:
 * Sets the the user monitor's horizontal and vertical
 * frequency limits
 */

static void __devinit i810fb_init_monspecs(struct fb_info *fb_info)
{
	if (!hsync1 && !hsync2) {
		hsync2 = 31;
		hsync1 = 31;
	}	
	fb_info->monspecs.hfmax = hsync2 * 1000;
	fb_info->monspecs.hfmin = hsync1 * 1000;
	if (hsync2 < hsync1) {
		if (hsync2)
			fb_info->monspecs.hfmin = hsync2 * 1000;
		else
			fb_info->monspecs.hfmax = hsync1 * 1000;
	}			
	if (!vsync2 && !vsync1) {
		vsync2 = 60;
		vsync1 = VFREQ_MIN;
	}	
	if (vsync2 > VFREQ_MAX)
		fb_info->monspecs.vfmax = VFREQ_MAX;
	else
		fb_info->monspecs.vfmax = vsync2;
#ifndef CONFIG_FB_I810_NONSTD
	if (vsync1 < VFREQ_MIN)
		fb_info->monspecs.vfmin = VFREQ_MIN;
	else
#endif
		fb_info->monspecs.vfmin = vsync1;		
	if (vsync2 < vsync1) {
		if (vsync2)
			fb_info->monspecs.vfmin = vsync2;
		else								 
			fb_info->monspecs.vfmax = vsync1;
	}		
}

/**
 * i810fb_init_defaults - initializes default values to use
 */
static void __devinit i810fb_init_defaults(void)
{
	i810fb_default.xres = xres;
	i810fb_default.yres = yres;
	i810fb_default.yres_virtual = vyres;
	i810fb_default.bits_per_pixel = bpp;
	i810fb_init_monspecs(&i810_info->fb_info);
	if (accel) i810fb_default.accel_flags = 1;
	if (hwcur) {
		i810_accel.cursor = i810_cursor;
		i810_noaccel8.cursor = i810_cursor;
		i810_noaccel16.cursor = i810_cursor;
		i810_noaccel24.cursor = i810_cursor;
	}	
}


/**
 * i810fb_save_vgax - save extended register states
 */
static void __devinit i810fb_save_vgax(void)
{
	u32 indexport, dataport;
	u8 i;

	if (is_cga()) {
		dataport = CR_DATA_CGA;
		indexport = CR_INDEX_CGA;
	}
	else {
		dataport = CR_DATA_MDA;
		indexport = CR_INDEX_MDA;
	}
	
	for (i = 0; i < 4; i++) {
		i810_writeb(indexport, CR30 + i);
		*(&(i810_info->hw_state.cr30) + i) = i810_readb(dataport);
	}
	i810_writeb(indexport, CR35);
	i810_info->hw_state.cr35 = i810_readb(dataport);
	i810_writeb(indexport, CR39);
	i810_info->hw_state.cr39 = i810_readb(dataport);
	i810_writeb(indexport, CR41);
	i810_info->hw_state.cr41 = i810_readb(dataport);
	i810_writeb(indexport, CR70);
	i810_info->hw_state.cr70 = i810_readb(dataport);	
	i810_info->hw_state.msr = i810_readb(MSR_READ);
	i810_writeb(indexport, CR80);
	i810_info->hw_state.cr80 = i810_readb(dataport);
	i810_writeb(SR_INDEX, SR01);
	i810_info->hw_state.sr01 = i810_readb(SR_DATA);
}

/**
 * i810fb_save_vgax - save standard register states
 */
static void __devinit i810fb_save_vga(void)
{
	u32 indexport, dataport;
	u8 i;
	if (is_cga()) {
		dataport = CR_DATA_CGA;
		indexport = CR_INDEX_CGA;
	}
	else {
		dataport = CR_DATA_MDA;
		indexport = CR_INDEX_MDA;
	}	
	for (i = 0; i < 10; i++) {
		i810_writeb(indexport, CR00 + i);
		*((&i810_info->hw_state.cr00) + i) = i810_readb(dataport);
	}
	for (i = 0; i < 8; i++) {
		i810_writeb(indexport, CR10 + i);
		*((&i810_info->hw_state.cr10) + i) = i810_readb(dataport);	
	}
}

/**
 * i810fb_restore_pll - restores saved PLL register
 */
static void __devexit i810fb_restore_pll(void)
{
	u32 tmp1, tmp2;
	
	tmp1 = i810_info->hw_state.dclk_2d;
	tmp2 = i810_readl(DCLK_2D);
	tmp1 &= ~MN_MASK;
	tmp2 &= MN_MASK;
	i810_writel(DCLK_2D, tmp1 | tmp2);

	tmp1 = i810_info->hw_state.dclk_1d;
	tmp2 = i810_readl(DCLK_1D);
	tmp1 &= ~MN_MASK;
	tmp2 &= MN_MASK;
	i810_writel(DCLK_1D, tmp1 | tmp2);

	i810_writel(DCLK_0DS, i810_info->hw_state.dclk_0ds);


}
/**
 * i810fb_restore_dac - restores saved DAC register
 */

static void __devexit i810fb_restore_dac(void)
{
	u32 tmp1, tmp2;

	tmp1 = i810_info->hw_state.pixconf;
	tmp2 = i810_readl(PIXCONF);
	tmp1 &= DAC_BIT;
	tmp2 &= ~DAC_BIT;
	i810_writel(PIXCONF, tmp1 | tmp2);
}

/**
 * i810fb_restore_vgax - restores saved extended VGA registers
 */
static void __devexit i810fb_restore_vgax(void)
{
	u32 indexport, dataport;
	u8 i, j;
	
	if (is_cga()) {
		dataport = CR_DATA_CGA;
		indexport = CR_INDEX_CGA;
	}
	else {
		dataport = CR_DATA_MDA;
		indexport = CR_INDEX_MDA;
	}
			
	for (i = 0; i < 4; i++) {
		i810_writeb(indexport, CR30+i);
		i810_writeb(dataport, *(&(i810_info->hw_state.cr30) + i));
	}
	i810_writeb(indexport, CR35);
	i810_writeb(dataport, i810_info->hw_state.cr35);
	i810_writeb(indexport, CR39);
	i810_writeb(dataport, i810_info->hw_state.cr39);
	i810_writeb(indexport, CR41);
	i810_writeb(dataport, i810_info->hw_state.cr39);

/*restore interlace*/
	i810_writeb(indexport, CR70);
	i = i810_info->hw_state.cr70;
	i &= INTERLACE_BIT;
	j = i810_readb(dataport);
	i810_writeb(indexport, CR70);
	i810_writeb(dataport, j | i);

	i810_writeb(indexport, CR80);
	i810_writeb(dataport, i810_info->hw_state.cr80);
	i810_writeb(MSR_WRITE, i810_info->hw_state.msr);
	i810_writeb(SR_INDEX, SR01);
	i = (i810_info->hw_state.sr01) & ~0xE0 ;
	j = i810_readb(SR_DATA) & 0xE0;
	i810_writeb(SR_INDEX, SR01);
	i810_writeb(SR_DATA, i | j);
}

/**
 * i810fb_restore_vga - restores saved standard VGA registers
 */
static void __devexit i810fb_restore_vga(void)
{
	u32 indexport, dataport;
	u8 i;
	
	if (is_cga()) {
		dataport = CR_DATA_CGA;
		indexport = CR_INDEX_CGA;
	}
	else {
		dataport = CR_DATA_MDA;
		indexport = CR_INDEX_MDA;
	}
	for (i = 0; i < 10; i++) {
		i810_writeb(indexport, CR00 + i);
		i810_writeb(dataport, *((&i810_info->hw_state.cr00) + i));
	}
	for (i = 0; i < 8; i++) {
		i810_writeb(indexport, CR10 + i);
		i810_writeb(dataport, *((&i810_info->hw_state.cr10) + i));
	}
}

/**
 * i810fb_restore_addr_map - restores saved address registers
 */
static void __devexit i810fb_restore_addr_map(void)
{
	u8 tmp;
	i810_writeb(GR_INDEX, GR10);
	tmp = i810_readb(GR_DATA);
	tmp &= ADDR_MAP_MASK;
	tmp |= i810_info->hw_state.gr10;
	i810_writeb(GR_INDEX, GR10);
	i810_writeb(GR_DATA, tmp);
}

/**
 * i810fb_restore_2D - restores saved graphics registers
 */
static void __devexit i810fb_restore_2d(void)
{
	u32 tmp_long;
	u16 tmp_word;

	tmp_word = i810_readw(BLTCNTL);
	tmp_word &= ~(3 << 4); 
	tmp_word |= i810_info->hw_state.bltcntl;
	i810_writew(BLTCNTL, tmp_word);
       
	i810_wait_for_hsync(); 
	i810fb_dram_off(OFF);
	i810_writel(PIXCONF, i810_info->hw_state.pixconf);
	i810fb_dram_off(ON);

	tmp_word = i810_readw(HWSTAM);
	tmp_word &= 3 << 13;
	tmp_word |= i810_info->hw_state.hwstam;
	i810_writew(HWSTAM, tmp_word);

	tmp_word = i810_readw(IER);
	tmp_word &= 3 << 13;
	tmp_word |= i810_info->hw_state.ier;
	i810_writew(IER, tmp_word);
	
	tmp_word = i810_readw(IMR);
	tmp_word &= 3 << 13;
	tmp_word |= i810_info->hw_state.imr;
	i810_writew(IMR, tmp_word);

	tmp_word = i810_readw(EMR);
	tmp_word &= ~0x3F;
	tmp_word |= i810_info->hw_state.ier;
	i810_writew(IER, tmp_word);

	tmp_long = i810_readl(FW_BLC);
	tmp_long &= FW_BLC_MASK;
	tmp_long |= i810_info->hw_state.fw_blc;
	i810_writel(FW_BLC, tmp_long);

	i810_writel(PGTBL_CTL, i810_info->hw_state.pgtbl_ctl);
}
	
/**
 * i810fb_save_2d - save graphics register states
 */
static void __devinit i810fb_save_2d(void)
{
	i810_info->hw_state.lring_state.head = i810_readl(LRING);
	i810_info->hw_state.lring_state.tail = i810_readl(LRING + 4);
	i810_info->hw_state.lring_state.start = i810_readl(LRING + 8);
	i810_info->hw_state.lring_state.size = i810_readl(LRING + 12);
	i810_info->hw_state.dclk_2d = i810_readl(DCLK_2D);
	i810_info->hw_state.dclk_1d = i810_readl(DCLK_1D);
	i810_info->hw_state.dclk_0ds = i810_readl(DCLK_0DS);
	i810_info->hw_state.pixconf = i810_readl(PIXCONF);
	i810_info->hw_state.fw_blc = i810_readl(FW_BLC);
	i810_info->hw_state.bltcntl = i810_readw(BLTCNTL);
	i810_info->hw_state.hwstam = i810_readw(HWSTAM);
	i810_info->hw_state.ier = i810_readw(IER);
	i810_info->hw_state.iir = i810_readw(IIR);
	i810_info->hw_state.imr = i810_readw(IMR);
	i810_info->hw_state.pgtbl_ctl = i810_readl(PGTBL_CTL);
}
	     
/**
 * i810fb_save_regs - saves the register state
 * 
 * DESCRIPTION:
 * Saves the ALL the registers' state
 */

static void __devinit i810fb_save_regs(void)
{
	i810fb_save_vga();
	i810fb_save_vgax();
	i810fb_save_2d();
}

	
/**
 * i810b_init_device - initialize device
 */
static void __devinit i810fb_init_device(void)
{
	i810fb_save_regs();
	i810fb_init_ringbuffer(); 
	if (hwcur) { 
		i810fb_init_cursor();
		add_timer(i810_info->cursor.timer);
	}
	/* mvo: enable external vga-connector (for laptops) */
	if(ext_vga) {
		i810_writel(HVSYNC, 0);
		i810_writel(PWR_CLKC, 3);
	}
  	i810fb_get_mem_freq();
}

/**
 * i810fb_restore_regs - loads the saved register state
 * 
 * DESCRIPTION:
 * Restores ALL of the registers' save state
 */
static void __devexit i810fb_restore_regs(void)
{
	i810fb_screen_off(OFF);
	i810fb_protect_regs(OFF);
	i810fb_dram_off(OFF);
	i810fb_restore_pll();
	i810fb_restore_dac();
	i810fb_restore_vga();
	i810fb_restore_vgax();
	i810fb_restore_addr_map();
	i810fb_dram_off(ON);
	i810fb_restore_2d();
	i810fb_restore_ringbuffer(&i810_info->hw_state.lring_state);
	i810fb_screen_off(ON);
	i810fb_protect_regs(ON);
}

/**
 * i810_create_master - create a master GART allocation record
 *
 * DESCRIPTION:
 * Creates table where allocation records are stored
 *
 * RETURNS:
 * pointer to table, zero if fail
 */
static struct pgtbl_record* __devinit i810_create_master(void)
{
	unsigned long size, master;
	size = sizeof(struct pgtbl_record) * 1024 * 16;
	master = (unsigned long) vmalloc(size);
	if (!master) 
		return 0;
	memset((void *) master, 0, size);
	return (struct pgtbl_record *) master;
}

/**
 * i810_destroy_master - destroy master record table
 */
static void __devexit i810_destroy_master(struct pgtbl_record *i810_master)
{
	vfree(i810_master);
}

/**
 * i810_enable_pgtbl - enables the GART table
 *
 * DESCRIPTION:
 * Sets the enable bit of the Control register
 */

static int __devinit i810fb_init_agp(void)
{
	int offset;
	u32 size;

	if(!(i810_info->i810_gtt.pgtbl_scratch = i810_allocate_page()))
		return 0;
	i810_info->i810_gtt.pgtbl_scratch = (virt_to_phys((void *) i810_info->i810_gtt.pgtbl_scratch)); 
	(u32) i810_info->i810_gtt.i810_pgtbl_virtual = i810_create_pgtbl();
	if(!i810_info->i810_gtt.i810_pgtbl_virtual)
		return 0;
        if (find_agpgart)
        	xgart = i810_readl(PGTBL_CTL);
	i810_info->i810_gtt.i810_pgtbl_phys = (u32 *) virt_to_phys((void *) i810_info->i810_gtt.i810_pgtbl_virtual);	
	i810_bind_pgtbl(i810_info->i810_gtt.i810_pgtbl_phys, i810_info->i810_gtt.i810_pgtbl_virtual);

	/* allocate framebuffer memory */
	size = i810_info->fb_size >> 12;	
	offset = i810_find_chunk(size);
	if (offset < 0)
		return 0;
	i810_info->fb_offset = offset;	
	if (!(i810_allocate_memory(i810_info->i810_gtt.i810_pgtbl_virtual, size, AGP_NORMAL, offset))) {	
		i810_info->fb_start_phys = 0;
		printk("i810fb_init:  cannot allocate framebuffer memory\n");
		return 0;
	}	
	
	/* allocate ringbuffer memory */
	size = RINGBUFFER_SIZE >> 12;
	offset = i810_find_chunk(size);
	if (offset < 0)
		return 0;
	i810_info->lring_offset = offset;	
	if (!(i810_allocate_memory(i810_info->i810_gtt.i810_pgtbl_virtual, size, AGP_NORMAL, offset))) {	
		i810_info->lring_start_phys = 0;
		printk("i810fb_init:  cannot allocate ringbuffer memory\n");
		return 0;
	}	

	/* allocate cursor memory */
	if (hwcur) {
		size = 1;
		offset = i810_find_chunk(size);
		if (offset < 0)
			return 0;
		i810_info->cursor_offset = offset;
		if (!(i810_info->cursor_start_phys = i810_allocate_memory(i810_info->i810_gtt.i810_pgtbl_virtual, 
							       size, AGP_PHYSICAL, offset))) {	
			i810_info->cursor_start_phys = 0;
			printk("i810fb_init:  cannot allocate cursor memory\n");
			return 0;
		}
	}		
	return 1;

}

    /*
     *  Initialisation
     */

static int __devinit i810fb_init_pci (struct pci_dev *dev, 
				      const struct pci_device_id *entry)
{
	struct resource *res;
	int err;
	u32 sysmem, extra;

#ifdef CONFIG_FB_I810_NONSTD
	u8 std = 1;
#else
	u8 std = 0;
#endif	
	if (!i810fb_enable)
		return -EINVAL;
	if(!(i810_info = kmalloc(sizeof(struct i810_fbinfo), GFP_KERNEL)))
		return -ENOMEM;
	memset(i810_info, 0, sizeof(struct i810_fbinfo));
	i810_info->dev = dev;
	if ((err = pci_enable_device(i810_info->dev))) {
		i810fb_release_resource();
		return err;		
	}

        /* Compute max framebuffer memory */
	sysmem = virt_to_phys(high_memory) >> 20;
	if (sysmem < 32) {
		printk("i810fb_init: system memory to small\n");
		kfree(i810_info);
		return -ENOMEM;
	}
	else if (sysmem >= 32 && sysmem < 64)
		i810_info->fb_max = 4 << 20;
	else if (sysmem >= 64 && sysmem < 128)
		i810_info->fb_max = 28 << 20;
	else
		i810_info->fb_max = 64 << 20;

	res = &i810_info->dev->resource[0];
	i810_info->fb_base_phys = res->start;
	i810_info->aper_size = res->end - res->start + 1;
	if (i810_info->fb_max > i810_info->aper_size)
		i810_info->fb_max = i810_info->aper_size;
	extra = (RINGBUFFER_SIZE + CURSOR_SIZE + (1 << 20) - 1) >> 20;
	if (vram < 1 || vram > (i810_info->fb_max >> 20) - extra) {
		printk("i810fb_init: vram exceeds boundaries( 1 - %d MB)\n", (int) ((i810_info->fb_max >> 20) - extra));
		kfree(i810_info);
		return -ENOMEM;
	}	
	i810_info->fb_size = vram << 20;

	if (!request_mem_region(i810_info->fb_base_phys, i810_info->aper_size, "Intel 810 Framebuffer Driver")) {
		printk("i810fb_init: cannot request framebuffer region\n");
		kfree(i810_info);
		return -ENODEV;
	}
	res = &i810_info->dev->resource[1];
	i810_info->mmio_start_phys = res->start;
	if (!request_mem_region(i810_info->mmio_start_phys, MMIO_SIZE, "Intel 810 Framebuffer Driver")) {
		printk("i810fb_init: cannot request mmio region\n");
		release_mem_region(i810_info->fb_base_phys, i810_info->aper_size);
		kfree(i810_info);
		return -ENODEV;
	}
        i810_info->mmio_start_virtual = (unsigned long) ioremap_nocache(i810_info->mmio_start_phys, MMIO_SIZE);

	i810_info->page_order = 3;
	if (i810_info->fb_max > 32 * 1024)
		i810_info->page_order = 4;
	
	if (!(i810_info->i810_gtt.i810_master = i810_create_master())) {
		printk("i810fb_init: cannot allocate master record table\n");
		release_mem_region(i810_info->fb_base_phys, i810_info->aper_size);
		release_mem_region(i810_info->mmio_start_phys, MMIO_SIZE);
		kfree(i810_info);
		return -ENOMEM;
	}
	if (!(i810fb_init_agp())) {
		i810fb_release_resource();
		return -ENOMEM;
	}

      	i810_info->fb_start_phys = i810_info->fb_base_phys + (i810_info->fb_offset << 12);
	i810_info->fb_start_virtual = (unsigned long) ioremap_nocache(i810_info->fb_start_phys + 
								 (i810_info->fb_offset << 12), i810_info->fb_size);

	i810_info->lring_start_phys = i810_info->fb_base_phys + (i810_info->lring_offset << 12);
	i810_info->lring_start_virtual = (unsigned long) ioremap_nocache(i810_info->lring_start_phys, RINGBUFFER_SIZE);
	if (hwcur) {
		i810_info->cursor_start_virtual = (unsigned long) ioremap_nocache(i810_info->fb_base_phys + 
										  (i810_info->cursor_offset << 12), 
										  CURSOR_SIZE);
		if (!i810fb_create_cursor()) {
			i810fb_release_resource();
			return -ENOMEM;
		}
	}		
#ifdef CONFIG_MTRR
	if (mtrr)
		mtrr_add((unsigned long) i810_info->fb_start_phys, i810_info->aper_size, MTRR_TYPE_WRCOMB, 1);
#endif
	i810fb_init_device();        
	i810fb_init_defaults();
	strcpy(i810_info->fb_info.modename, i810fb_name);
	strcpy(i810_info->fb_info.fontname, fontname);
	i810_info->fb_info.changevar = NULL;
	i810_info->fb_info.node = -1;
	i810_info->fb_info.fbops = &i810fb_ops;
	i810_info->fb_info.disp = &i810_info->disp;
	i810_info->fb_info.switch_con = &i810fbcon_switch;
	i810_info->fb_info.updatevar = &i810fbcon_updatevar;
	i810_info->fb_info.blank = &i810fbcon_blank;
	i810_info->fb_info.flags = FBINFO_FLAG_DEFAULT;

	if((err = i810fb_set_var(&i810fb_default, -1, &i810_info->fb_info))) {
		i810fb_release_resource();
		return err;
	}

   	if((err = register_framebuffer(&i810_info->fb_info))) {
    		i810fb_release_resource(); 
    		return err;  
    	}   
	i810fb_load_regs(&i810_info->disp);
	i810fb_enable = 0;
	
      	printk("fb%d: Intel 810/815 framebuffer device v%d.%d.%d, Tony Daplas\n"
      	       "     Framebuffer using %dK of System RAM\n" 
	       "     Logical framebuffer starts at 0x%04x\n"
	       "     MMIO address starts at 0x%04x\n"
	       "     Mode is %dx%d @ %dbpp\n"
	       "     Acceleration is %sabled\n"
	       "     MTRR is %sabled\n"
	       "     External VGA is %sabled\n"
	       "     Hardware cursor is %sabled\n" 
	       "     Using %sstandard video timings\n",	
	       GET_FB_IDX(i810_info->fb_info.node), VERSION_MAJOR, VERSION_MINOR, VERSION_TEENIE,
	       (int) i810_info->fb_size>>10, (int) i810_info->fb_start_phys, (int) i810_info->mmio_start_phys,
	       i810fb_default.xres, i810fb_default.yres, i810fb_default.bits_per_pixel,
	       (accel) ? "en" : "dis", (mtrr) ? "en" : "dis", (ext_vga) ? "en" : "dis", (hwcur) ? "en" : "dis",
	       (std) ? "non" : "");
	
	return 0;
}


static void __devexit i810fb_release_resource(void)
{
  #ifdef CONFIG_MTRR 
  	if (mtrr) 
  		mtrr_del(0, (unsigned long) i810_info->fb_start_phys, i810_info->aper_size); 
  #endif 
	if (hwcur) {
		del_timer(i810_info->cursor.timer);
		kfree(i810_info->cursor.timer);
		if (i810_info->cursor_start_virtual) {
			iounmap((void *) i810_info->cursor_start_virtual);
			i810_info->cursor_start_virtual = 0;
			printk("i810fb_release_resource: cursor unmapped\n");
		}
		if (i810_info->cursor_start_phys) {
			i810_free_memory(i810_info->i810_gtt.i810_pgtbl_virtual, 
					 AGP_PHYSICAL, i810_info->cursor_offset);
			i810_info->cursor_start_phys = 0;
			printk("i810fb_release_resource: cursor released\n");
		}
	}	
	i810fb_lring_enable(OFF);
	if (find_agpgart)
		i810_writel(PGTBL_CTL, xgart);
	else
		i810_unbind_pgtbl();
	if (i810_info->lring_start_virtual) {
		iounmap((void *) i810_info->lring_start_virtual);
		i810_info->lring_start_virtual = 0;
		printk("i810fb_release_resource: lring unmapped\n");
	}
	if (i810_info->lring_start_phys) {
		i810_free_memory(i810_info->i810_gtt.i810_pgtbl_virtual, AGP_NORMAL, i810_info->lring_offset);
		i810_info->lring_start_phys = 0;
		printk("i810fb_release_resource: lring released\n");
	}
	if (i810_info->fb_start_virtual) {
		iounmap((void *) i810_info->fb_start_virtual);
		i810_info->fb_start_virtual = 0;
		printk("i810fb_release_resource: framebuffer unmapped\n");
	}
	if (i810_info->fb_start_phys) {
		i810_free_memory(i810_info->i810_gtt.i810_pgtbl_virtual, AGP_NORMAL, i810_info->fb_offset);
		i810_info->fb_start_phys = 0;
		printk("i810fb_release_resource: framebuffer released\n");
	}	
	if (i810_info->i810_gtt.i810_pgtbl_virtual) {
		i810_disable_pgtbl();
		i810_destroy_pgtbl((char *) i810_info->i810_gtt.i810_pgtbl_virtual);
		i810_info->i810_gtt.i810_pgtbl_virtual = 0;
		printk("i810fb_release_resource: page table released\n");
	}	
	if (i810_info->i810_gtt.i810_master) {
		i810_destroy_master(i810_info->i810_gtt.i810_master);
		i810_info->i810_gtt.i810_master = 0;
		printk("i810fb_release_resource: record table released\n");
	}	
	if (i810_info->i810_gtt.pgtbl_scratch) {
		i810_info->i810_gtt.pgtbl_scratch = (unsigned long) phys_to_virt(i810_info->i810_gtt.pgtbl_scratch);
		i810_free_page((void *) i810_info->i810_gtt.pgtbl_scratch);
		i810_info->i810_gtt.pgtbl_scratch = 0;
		printk("i810fb_release_resource: scratch page released\n");
	}
	if (i810_info->mmio_start_virtual) {
		iounmap((void *) i810_info->mmio_start_virtual);
		i810_info->mmio_start_virtual = 0;
		printk("i810fb_release_resource: MMIO address unmapped\n");
	}	
	release_mem_region(i810_info->fb_base_phys, i810_info->aper_size);
	printk("i810fb_release_resource: resource 0 released\n");
	release_mem_region(i810_info->mmio_start_phys, MMIO_SIZE);
	printk("i810fb_release_resource: resource 1 released\n");
	pci_disable_device(i810_info->dev);
	if (i810_info) {
		kfree(i810_info);
		printk("i810fb_release_resource: i810 private data released\n");
	}
}

static void __devexit i810fb_remove_pci(struct pci_dev *dev)
{
	i810fb_restore_regs();
	unregister_framebuffer(&i810_info->fb_info);  
	i810fb_release_resource();
	printk("cleanup_module:  unloaded i810 framebuffer device\n");
}                                                	


/* Modularization */

int __init i810fb_init(void)
{
	int err;
	i810fb_enable = 1;
	err = pci_module_init(&i810fb_driver);
	if (err)
		return err;
	return 0;
}

#ifdef MODULE

MODULE_PARM(vram, "i");
MODULE_PARM_DESC(vram, "System RAM to allocate to framebuffer in MiB (default=4)");
MODULE_PARM(bpp, "i");
MODULE_PARM_DESC(bpp, "Color depth for display in bits per pixel (default = 8)");
MODULE_PARM(xres, "i");
MODULE_PARM_DESC(xres, "Hozizontal resolution in pixels (default = 640)");
MODULE_PARM(yres, "i");
MODULE_PARM_DESC(yres, "Vertical resolution in scanlines (default = 480)");
MODULE_PARM(vyres, "i");
MODULE_PARM_DESC(vyres, "Virtual vertical resolution in scanlines (default = 480)");
MODULE_PARM(hsync1, "i");
MODULE_PARM_DESC(hsync1, "Mimimum horizontal frequency of monitor in KHz (default = 31)");
MODULE_PARM(hsync2, "i");
MODULE_PARM_DESC(hsync2, "Maximum horizontal frequency of monitor in KHz (default = 31)");
MODULE_PARM(vsync1, "i");
MODULE_PARM_DESC(vsync1, "Minimum vertical frequency of monitor in Hz(default = 50)");
MODULE_PARM(vsync2, "i");
MODULE_PARM_DESC(vsync2, "Maximum vertical frequency of monitor in Hz(default = 60)");
MODULE_PARM(rr, "i");
MODULE_PARM_DESC(rr, "Try to use this refresh rate (default = 60)");
MODULE_PARM(accel, "i");
MODULE_PARM_DESC(accel, "Use Acceleration (BLIT) engine (default = 0)");
MODULE_PARM(mtrr, "i");
MODULE_PARM_DESC(mtrr, "Use MTRR (default = 0)");
MODULE_PARM(ext_vga, "i");
MODULE_PARM_DESC(ext_vga, "Enable external VGA connector (default = 0)");
MODULE_PARM(xcon, "i");
MODULE_PARM_DESC(xcon, "console number used by XFree86 - zero-based (default = 6)");
MODULE_PARM(hwcur, "i");
MODULE_PARM_DESC(hwcur, "use hardware cursor (default = 0)");

MODULE_AUTHOR("Tony A. Daplas");
MODULE_DESCRIPTION("Framebuffer device for the Intel 810/815 and compatible cards");
MODULE_LICENSE("GPL"); 

static void __exit i810fb_exit(void)
{
	pci_unregister_driver(&i810fb_driver);
}
module_init(i810fb_init);
module_exit(i810fb_exit);

#endif

