/*
 *  linux/fs/supermount/file_operations.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/dir.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/dir.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2 & 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *  Rewriten for kernel 2.4. (C) 2001 MandrakeSoft Inc.
 *                                    Juan Quintela (quintela@mandrakesoft.com)
 *
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/sched.h>
#include <linux/file.h>
#include <linux/poll.h>
#include <linux/slab.h>

#include "supermount_i.h"

#define DEFAULT_POLLMASK (POLLIN | POLLOUT | POLLRDNORM | POLLWRNORM)

static loff_t
supermount_llseek(struct file *file, loff_t offset, int origin)
{
	struct file *subfile = get_subfile(file);
	loff_t retval;

	dump_file(file);
	retval = vfs_llseek(subfile, offset, origin);
	file->f_pos = subfile->f_pos;
	file->f_reada = subfile->f_reada;
	file->f_version = subfile->f_version;
	return retval;
}

static ssize_t
supermount_read(struct file *file, char *buf, size_t count, loff_t * ppos)
{
	struct file *subfile = get_subfile(file);
	int rc;

	dump_file(file);
	rc = -EINVAL;
	if (subfile->f_op && subfile->f_op->read)
		rc = subfile->f_op->read(subfile, buf, count, ppos);
	if (rc > 0)
		file->f_pos = subfile->f_pos = *ppos;
	return rc;
}

static ssize_t
supermount_write(struct file *file, const char *buf,
		 size_t count, loff_t * ppos)
{
	struct file *subfile = get_subfile(file);
	int rc = 0;

	dump_file(file);
	supermount_debug("count= %d, pos %Ld", count, *ppos);

	if (subfile->f_op && subfile->f_op->write)
		rc = subfile->f_op->write(subfile, buf, count, ppos);
	supermount_debug("rc: %d", rc);
	if (rc > 0) {
		struct inode *subinode;
		subinode = subfile->f_dentry->d_inode;

		supermount_debug("%Ld, %d, %Ld, %Ld", subinode->i_size,
				 count, *ppos, subfile->f_pos);
		file->f_pos = subfile->f_pos = *ppos;
		file->f_mode = subfile->f_mode;
		file->f_dentry->d_inode->i_size = subinode->i_size;
		file->f_dentry->d_inode->i_blocks = subinode->i_blocks;
		file->f_dentry->d_inode->i_mode = subinode->i_mode;
		supermount_debug("%Ld, %d, %Ld", subinode->i_size,
				 count, *ppos);
	}
	return rc;
}

int
supermount_readdir(struct file *file, void *buf, filldir_t fill_fn)
{
	struct file *subfile = get_subfile(file);
	int rc;

	dump_file(file);
	rc = vfs_readdir(subfile, fill_fn, buf);
	file->f_pos = subfile->f_pos;	/* update position, nfs needs it */
	UPDATE_ATIME(file->f_dentry->d_inode);
	return rc;
}

static unsigned int
supermount_poll(struct file *file, struct poll_table_struct *table)
{
	struct file *subfile = get_subfile(file);
	int rc = DEFAULT_POLLMASK;

	dump_file(file);
	if (subfile->f_op && subfile->f_op->poll)
		rc = subfile->f_op->poll(subfile, table);

	return rc;
}

static int
supermount_ioctl(struct inode *inode, struct file *file,
		 unsigned int cmd, unsigned long arg)
{
	struct file *subfile = get_subfile(file);
	struct inode *subinode;
	int rc = -ENOTTY;

	dump_inode(inode);
	subinode = subfile->f_dentry->d_inode;
	if (subfile->f_op && subfile->f_op->ioctl)
		rc = subfile->f_op->ioctl(subinode, subfile, cmd, arg);
	return rc;
}

int
supermount_open(struct inode *inode, struct file *file)
{
	struct dentry *subdent;
	struct file *subfile;
	struct supermount_sb_info *sbi = supermount_sbi(inode->i_sb);
	struct vfsmount *mnt;
	int rc;

	dump_inode(inode);

	rc = -EIO;
	if (subfs_go_online(inode->i_sb))
		goto out;

	subdent = get_subfs_dentry(file->f_dentry);
	rc = PTR_ERR(subdent);
	if (IS_ERR(subdent))
		goto out;

	if (!subdent->d_inode)
		BUG();

	subfs_get_read_access(inode->i_sb);

	if (file->f_mode & FMODE_WRITE) {
		rc = subfs_get_write_access(inode->i_sb);
		if (rc)
			goto dput_ret;
	}

	mnt = mntget(sbi->s_undermount);
	subfile = dentry_open(subdent, mnt, file->f_flags);
	rc = PTR_ERR(subfile);
	if (IS_ERR(subfile))
		goto close_ret;

	put_subfile(file, subfile);
	subfile->f_mode = file->f_mode;

	return 0;

close_ret:
	subfs_put_write_access(inode->i_sb);
dput_ret:
	subfs_put_read_access(inode->i_sb);
	dput(subdent);
out:
	subfs_go_inactive(inode->i_sb);
	return rc;
}

static int
supermount_flush(struct file *file)
{
	struct file *subfile = get_subfile(file);
	int rc = 0;

	dump_file(file);
	if (subfile->f_op && subfile->f_op->flush)
		rc = subfile->f_op->flush(subfile);

	return rc;
}

static int
supermount_release(struct inode *inode, struct file *file)
{
	struct file *subfile = get_subfile(file);
	struct inode *subinode;
	int write_on = 0;

	dump_inode(inode);
	dump_file(file);

	subinode = subfile->f_dentry->d_inode;
	if (!(file->f_mode & FMODE_WRITE) && subinode->i_nlink == 0) {
		if (!subfs_get_write_access(inode->i_sb))
			write_on = 1;

	}

	filp_close(subfile, current->files);
	if ((file->f_mode & FMODE_WRITE) || write_on) {
		mark_subfs_dirty(inode->i_sb);
		subfs_put_write_access(inode->i_sb);
	}
	subfs_put_read_access(inode->i_sb);
	subfs_go_inactive(inode->i_sb);
	return 0;
}

static int
supermount_fsync(struct file *file, struct dentry *dentry, int datasync)
{
	struct file *subfile = get_subfile(file);
	int rc = -EINVAL;

	dump_file(file);

	if (subfile->f_op && subfile->f_op->fsync)
		rc = subfile->f_op->fsync(subfile, subfile->f_dentry, datasync);

	return rc;
}

static int
supermount_fasync(int fd, struct file *file, int on)
{
	struct file *subfile = get_subfile(file);
	int rc = -EINVAL;

	dump_file(file);

	if (subfile->f_op && subfile->f_op->fasync)
		rc = subfile->f_op->fasync(fd, subfile, on);

	return rc;
}

static int
supermount_lock(struct file *file, int cmd, struct file_lock *fl)
{
	struct file *subfile = get_subfile(file);
	int rc = 0;

	dump_file(file);
	if (subfile->f_op && subfile->f_op->lock)
		rc = subfile->f_op->lock(subfile, cmd, fl);
	else if (cmd == F_GETLK)
		posix_test_lock(file, fl);
	return rc;
}

/* Fixme:
 * readv: easy, export churnk from vfs
 * writev: easy, export churnk from vfs
 * sendpage: only used for networking, not needed
 * get_unmmapped_area: only used for devices, not needed
 */

struct file_operations supermount_dir_operations = {
	llseek:supermount_llseek,
	read:supermount_read,
	readdir:supermount_readdir,
	ioctl:supermount_ioctl,
	open:supermount_open,
	flush:supermount_flush,
	release:supermount_release,
	fsync:supermount_fsync,
	fasync:supermount_fasync,
};

struct file_operations supermount_file_operations = {
	llseek:supermount_llseek,
	read:supermount_read,
	write:supermount_write,
	poll:supermount_poll,
	ioctl:supermount_ioctl,
	mmap:generic_file_mmap,
	open:supermount_open,
	flush:supermount_flush,
	release:supermount_release,
	fsync:supermount_fsync,
	fasync:supermount_fasync,
	lock:supermount_lock,
};
