/*
 *  linux/fs/supermount/super.c
 *
 *  Original version:
 *      Copyright (C) 1995, 1997
 *      Stephen Tweedie (sct@dcs.ed.ac.uk)
 *
 *      from
 *
 *      linux/fs/minix/inode.c
 *      Copyright (C) 1991, 1992  Linus Torvalds
 *
 *      and
 *
 *      linux/fs/ext2/super.c
 *      Copyright (C) 1992, 1993, 1994, 1995  Remy Card
 *
 *  Rewriten for kernel 2.2, 2.4. (C) 1999, 2000 Alexis Mikhailov
 *                                    (alexis@abc.cap.ru)
 *
 */

#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/list.h>
#include <linux/locks.h>

#include "supermount_i.h"

struct inode *
supermount_create_inode(struct super_block *s)
{
	struct inode *inode = new_inode(s);
	struct supermount_sb_info *sbi = supermount_sbi(s);

	if (!inode)
		return NULL;

	inode->i_mode = sbi->s_default_mode;
	if (s->s_root) {
		inode->i_uid = s->s_root->d_inode->i_uid;
		inode->i_gid = s->s_root->d_inode->i_gid;
	} else {
		inode->i_uid = 0;
		inode->i_gid = 0;
	}
	inode->i_blksize = inode->i_sb->s_blocksize;
	inode->i_rdev = 0;
	inode->i_atime = inode->i_mtime = inode->i_ctime = CURRENT_TIME;
	supermount_i(inode)->i_sb_version = sbi->s_version;
	list_add(&(supermount_i(inode)->i_fs), &sbi->s_inodes);

	return inode;
}

/* Check for media change, but without invalidating buffers and inodes */
static int
just_check_disk_change(dev_t dev)
{
	supermount_debug("beginning");
	if (!query_disk_change(dev))
		return 0;
	supermount_debug("Disk change detected on device %d/%d",
			 MAJOR(dev), MINOR(dev));
	return 1;
}

static void
supermount_clean_inodes(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct list_head *ptr;
	struct supermount_inode_info *info;

	list_for_each(ptr, &sbi->s_inodes) {
		info = list_entry(ptr, struct supermount_inode_info, i_fs);
		if (!info)
			BUG();
		if (!info->i_dentry) {
			supermount_debug("inode without dentry");
			continue;
		}
		supermount_debug("name: *%s*", info->i_dentry->d_name.name);
		dput(info->i_dentry);
		info->i_dentry = NULL;
	}
}

void
subfs_umount(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct vfsmount *mnt = sbi->s_undermount;
	struct inode *root_inode;

	dump_sbi(sbi);

	if (sb->s_root) {
		root_inode = sb->s_root->d_inode;
		supermount_debug("there is s_root");
		if (!supermount_i(root_inode)->i_dentry)
			BUG();
		dput(supermount_i(root_inode)->i_dentry);
		supermount_i(root_inode)->i_dentry = NULL;
	}
	supermount_debug("before shrink_dcache");
	shrink_dcache_sb(sb);
	supermount_clean_inodes(sb);
	supermount_debug("before freeing mnt, count %d",
			 atomic_read(&mnt->mnt_count));
	if (!mnt)
		supermount_panic(sb, "subfs_umount",
				 "Can't find mount point\n");
	mntput(mnt);
	supermount_debug("after freeing mnt");
	sbi->s_state = SUPERMOUNT_UNMOUNTED;
	sbi->s_undermount = NULL;
	sbi->s_version++;
	sb->s_flags = sbi->s_mflags;
//      subfs_put_read_access (sb);
}

static int
subfs_remount(struct super_block *sb, int flags)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct super_block *subsb = supermount_subfs_sb(sbi);
	int real_flags = (sb->s_flags | flags) & MS_RMT_MASK;
	int retval;

	supermount_debug("flags = %d", flags);
	retval = do_remount_sb(subsb, real_flags, NULL);
	if (flags & MS_RDONLY)
		fsync_dev(subsb->s_dev);
	return retval;
}

static inline int
subfs_remount_ro(struct super_block *sb)
{
	int retval = subfs_remount(sb, MS_RDONLY);

	if (retval) {
		supermount_debug("Error %d remounting subfs ro", retval);
		return retval;
	}
	unmark_subfs_rw(sb);
	return 0;
}

static inline int
subfs_remount_rw(struct super_block *sb)
{
	int retval = subfs_remount(sb, 0);

	if (retval) {
		supermount_debug("Error %d remounting subfs rw", retval);
		return retval;
	}
	mark_subfs_rw(sb);
	return 0;
}

/*
   return: 0 means success
           1 failure
*/

static int
subfs_needs_mount(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	kdev_t dev;

	dump_sbi(sbi);

	if (!subfs_is_mounted(sb))
		return 1;

	/* check for disk change or disk not present */
	dev = supermount_subfs_sb(sbi)->s_dev;
	if (!just_check_disk_change(dev))
		return 0;

	/* We have a disk change!  Unmount the subfs */
	subfs_umount(sb);
	/* The call to just_check_disk_change may clear the media-
	   changed flag on the device, so we need to force a media
	   invalidation.  We don't want to do this before unmounting
	   the subfs, naturally! */
	invalidate_media(dev);

	return 1;
}

static int
subfs_real_mount(struct super_block *sb, char *type)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct vfsmount *mnt;
	int retval;

	supermount_debug("type=%s, dev=%s, data=%s",
			 type, sbi->s_devname, sbi->s_data);

	if (sbi->s_data_copy)
		strcpy(sbi->s_data, sbi->s_data_copy);

	mnt = do_kern_mount(type,
			    sb->s_flags | MS_MGC_VAL,
			    sbi->s_devname, sbi->s_data);

	if (IS_ERR(mnt)) {
		retval = PTR_ERR(mnt);
		supermount_debug("Error mounting %d", retval);
	} else {
		retval = 0;
		supermount_debug("mount finished ok");
		sbi->s_undermount = mnt;
	}
	return retval;
}

static int
subfs_real_mount2(struct super_block *sb, char *type)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	struct inode *root_inode;
	int retval = 0;

	dump_sbi(sbi);
	sb->s_flags = sbi->s_mflags;

	retval = subfs_real_mount(sb, type);

	if (retval == -EROFS) {
		sb->s_flags |= MS_RDONLY;
		supermount_debug("mounting readonly");
		retval = subfs_real_mount(sb, type);
		if (!retval)
			goto assign_root_inode;
	}
	if (retval)
		goto out;

	if (!(sb->s_flags & MS_RDONLY)) {
		supermount_debug("RW filesystem, remounting RO");
		retval = subfs_remount_ro(sb);
		if (retval)
			goto out;
	}
assign_root_inode:
	root_inode = sb->s_root->d_inode;
	if (supermount_i(root_inode)->i_dentry)
		BUG();
	supermount_i(root_inode)->i_dentry = dget(sbi->s_undermount->mnt_root);
	dump_sbi(sbi);
out:
	return retval;
}

static char *types_in_order[] = {
	"udf",
	"iso9660",
	"vfat",
	"msdos",
	"ext2",
	NULL
};

static int
subfs_mount(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	int retval = -ENODEV;
	char **type = types_in_order;

	if (sbi->s_type && strcmp(sbi->s_type, "auto"))
		return subfs_real_mount2(sb, sbi->s_type);

	while (*type && retval) {
		retval = subfs_real_mount2(sb, *type);
		type++;
	}

	return retval;
}

/* Return 0 (OK) if medium is valid. */
int
subfs_go_online(struct super_block *sb)
{
	int retval = 0;
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	dump_sbi(sbi);

	/* If we are dismounting, never return a success value. */
	if (atomic_read(&sbi->s_dismount))
		return 1;

	/* If there are still files open, we never bother checking the
	   medium. */
	if (subfs_is_active(sb))
		return 0;

//      lock_super (sb);

	if (!subfs_needs_mount(sb))
		goto out;

	supermount_debug("we need to mount again");

	/* OK, we're unmounted now --- can we remount?  Please? */
	if (sbi->s_undermount)
		BUG();

	retval = subfs_mount(sb);
	if (retval) {
		supermount_debug("Mount failed, errno %d", retval);
		retval = 1;
		goto out;
	}
	supermount_debug("mount go ok");
	sbi->s_version++;
	supermount_i(sb->s_root->d_inode)->i_sb_version = sbi->s_version;
	unmark_subfs_rw(sb);
	sbi->s_state = SUPERMOUNT_SUSPENDED;
out:
//      unlock_super (sb);
	return retval;
}

/* 
 * Check the state of the supermount filesystem after it has been
 * released by some inode.  If it is no longer active, do the necessary
 * cleanup: fsync the disk, remount it readonly if appropriate, and
 * unlock the drive door if possible.
 */

void
subfs_go_inactive(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	dump_sbi(sbi);

	/* If we are dismounting, do nothing. */
	if (atomic_read(&sbi->s_dismount))
		return;

	if (!subfs_is_mounted(sb))
		return;

	if (subfs_is_active(sb))
		return;

	/*
	 * Do an initial fsync_dev.  We do this before locking the 
	 * superblock to minimise the duration of the lock. 
	 */
	if (subfs_is_dirty(sb)) {
		supermount_debug("subfs is dirty, sync super fs");
		fsync_dev(sb->s_dev);
		supermount_debug("superfs synced");
		mark_subfs_clean(sb);
	}

	/* 
	 * We need to recheck the state of the filesystem now.
	 * It may have become active again.  If not, we'll do
	 * some cleanup under the superblock lock to tidy things
	 * up.  */
//      lock_super (sb);

	if (!subfs_is_mounted(sb) || subfs_is_active(sb))
		goto out;

	/*
	 * First thing is, has the filesystem been made
	 * read-write since it became active?  If so, now is the
	 * time to retire it back to read-only state.
	 */

	if (subfs_is_rw(sb)) {
		int retval;

		/*
		 * Yup, it's read-write --- remount it.
		 */

		supermount_debug("remounting readonly");
		retval = subfs_remount_ro(sb);
		if (retval)
			supermount_panic(sb, "subfs_go_inactive",
					 "Error %d remounting "
					 "subfs readonly", retval);
	}

	/*
	 * We can now afford to unlock the drive door. 
	 */
out:
//      unlock_super (sb);
	dump_sbi(sbi);
}

/*
 * Try to put the filesystem into writable mode for the duration of
 * an open(O_RDWR) or an inode operation requiring write access.  
 */

int
subfs_get_write_access(struct super_block *sb)
{
	int retval;
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	dump_sbi(sbi);

	if (!subfs_is_online(sb))
		BUG();

//      lock_super (sb);

	retval = 0;
	if (!atomic_read(&sbi->s_writecount)) {
		/*
		 * We may already have write access to the filesystem
		 */
		if (!subfs_is_rw(sb)) {
			if (sb->s_flags & MS_RDONLY)
				retval = -EROFS;
			else {
				supermount_debug("remounting read/write");
				retval = subfs_remount_rw(sb);
			}
		}
	}

	if (!retval)
		atomic_inc(&sbi->s_writecount);

//      unlock_super (sb);
	dump_sbi(sbi);
	return retval;
}

/*
 * Once we're done, release write access.
 */
void
subfs_put_write_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	dump_sbi(sbi);

	if (!subfs_is_rw(sb)) {
		BUG();
		supermount_panic(sb, "subfs_put_write_access",
				 "filesystem not read/write");
	}
	if (atomic_read(&sbi->s_writecount) == 0) {
		supermount_panic(sb, "subfs_put_write_access",
				 "filesystem not write accessed");
	}
	atomic_dec(&sbi->s_writecount);
}

/* Supermount open/close inode.  These functions maintain the
   superblock reference counts of active inodes; the subfs is
   suspended when that count reaches zero.

   Files are opened on any significant reference, and are not closed
   until they become fully dereferenced (during the last iput).
   Directories, on the other hand, are always closed unless they are
   active and opened; a directory referenced as CWD is not open.
*/

void
subfs_get_read_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	dump_sbi(sbi);
	atomic_inc(&sbi->s_opencount);
	switch (sbi->s_state) {
	case SUPERMOUNT_UNMOUNTED:
		supermount_panic(sb, "subfs_get_read_access",
				 "opening inode on unmounted subfs");
	case SUPERMOUNT_SUSPENDED:
		supermount_lock_door(sb);
		sbi->s_state = SUPERMOUNT_ONLINE;
		supermount_debug("going online.");
		break;
	case SUPERMOUNT_ONLINE:
		/* No need to do anything */ ;
	}
	dump_sbi(sbi);
}

void
subfs_put_read_access(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	dump_sbi(sbi);
	atomic_dec(&sbi->s_opencount);
	if (atomic_read(&sbi->s_opencount) == 0) {
		supermount_unlock_door(sb);
		sbi->s_state = SUPERMOUNT_SUSPENDED;
		supermount_debug("going offline.");
	}
}
