#ifndef _SUPERMOUNT_I_H
#define _SUPERMOUNT_I_H

#define SUPERMOUNT_DEBUG

/*
 * The supermount superblock magic number
 */

#define SUPERMOUNT_SUPER_MAGIC	0x9fa1

static inline int
is_inode_supermounted(struct inode *inode)
{
	return (inode->i_sb &&
		(inode->i_sb->s_magic == SUPERMOUNT_SUPER_MAGIC));
}

static inline int
is_dentry_supermounted(struct dentry *dentry)
{
	return (dentry->d_inode && is_inode_supermounted(dentry->d_inode));
}

#define SUPERMOUNT_VERSION		"0.7"

/*
 * The difference between SUSPENDED and ONLINE is that in SUSPENDED
 * the door is unlocked, and in ONLINE it is locked.
 */

typedef enum {
	SUPERMOUNT_UNMOUNTED,	/* No media mounted */
	SUPERMOUNT_SUSPENDED,	/* Mounted but suspended because 
				   no files open */
	SUPERMOUNT_ONLINE,	/* Mounted and active */
} sm_state_t;

/*
 * supermount super-block data in memory
 */
struct supermount_sb_info {
	sm_state_t s_state;
	int s_version;		/* Used to indicate obsolete inodes */
	mode_t s_default_mode;	/* Default mode for supermount root */

	char *s_type;		/* Type of fs to be sub-mounted */
	char *s_devname;	/* Where to mount the subfs from */
	/* unused */
	int s_mflags;		/* Flags to pass when mounting subfs */
	char *s_data;		/* Data to pass when mounting subfs */
	char *s_data_copy;	/* copy of previous data.  FS
				   destroy that data, and we
				   need to mount several
				   times.  We could allocate
				   the memory in
				   supermount_mount, but then
				   we have more possible
				   errors */
	struct vfsmount *s_undermount;	/* Mount point for subfs */
	atomic_t s_opencount;	/* Refcount of opened inodes
				   (an inode in use as cwd
				   does not count as open) */
	atomic_t s_writecount;	/* Refcount of write accesses
				   on the filesystem */
	atomic_t s_dismount;	/* if != 0 we are unmounting */

	/* Flags */
	int s_dirty:1;		/* Do we need to fsync() the subfs? */
	int s_rw:1;		/* Is the subfs mounted read/write? */
	struct list_head s_inodes;
};

static inline struct supermount_inode_info *
supermount_i(struct inode *inode)
{
	return (&inode->u.supermount_i);
}

static inline struct supermount_sb_info *
supermount_sbi(struct super_block *sb)
{
	if (!sb)
		BUG();
	return (struct supermount_sb_info *) (sb->u.generic_sbp);
}

static inline struct super_block *
supermount_subfs_sb(struct supermount_sb_info *sbi)
{
	return sbi->s_undermount->mnt_sb;
}

/*
 * Function prototypes
 */

static inline struct file *
get_subfile(struct file *file)
{
	struct file *subfile = (struct file *) file->private_data;
	if (!subfile)
		BUG();
	return subfile;
}

static inline void
put_subfile(struct file *file, struct file *subfile)
{
	file->private_data = (void *) subfile;
}

/* How to test if an inode is obsolete: */
static inline int
is_inode_obsolete(struct inode *inode)
{
	struct supermount_sb_info *sbi = supermount_sbi(inode->i_sb);

	return (supermount_i(inode)->i_sb_version != sbi->s_version);
}

/* Manage the subfs dirty flag */
static inline void
mark_subfs_dirty(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	sbi->s_dirty = 1;
}
static inline void
mark_subfs_clean(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	sbi->s_dirty = 0;
}
static inline int
subfs_is_dirty(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	return (sbi->s_dirty);
}

/* Manage the subfs rw flag */
static inline void
mark_subfs_rw(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	sbi->s_rw = 1;
}
static inline void
unmark_subfs_rw(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	sbi->s_rw = 0;
}
static inline int
subfs_is_rw(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	return (sbi->s_rw);
}

static inline int
subfs_is_online(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	return (sbi->s_state == SUPERMOUNT_ONLINE);
}

static inline int
subfs_is_mounted(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);

	return (sbi->s_state != SUPERMOUNT_UNMOUNTED);
}

/* How to test if a supermount filesystem is active or not: */

static inline int
subfs_is_active(struct super_block *sb)
{
	struct supermount_sb_info *sbi = supermount_sbi(sb);
	int write_mounted = atomic_read(&sbi->s_writecount) > 0;
	int read_mounted = atomic_read(&sbi->s_opencount) > 1;
	/*
	 * The subfs is always active if it is write-accessed.  (We
	 * don't resort to inode counts in that case since we can be
	 * write-active even if we are only accessing directories, in
	 * the case of such operations as renames.)  Otherwise, it is
	 * deemed inactive iff only the root is open, and its i_count is
	 * precisely one.  
	 */
	return write_mounted || (subfs_is_mounted(sb) && read_mounted);
}

/*
 * Debug code
 */
extern char supermount_debug_enable;

#ifdef SUPERMOUNT_DEBUG

#define dump(args...) \
do { \
	if (supermount_debug_enable) { \
		printk (KERN_DEBUG "[%s:%s:%d] ", \
			__FILE__, __FUNCTION__, __LINE__); \
		printk (args); \
		printk ("\n"); \
	} \
} while (0)

static inline void
_dump_inode(char *msg, int line, struct inode *i)
{
	if (!supermount_debug_enable)
		return;
	if (!i) {
		printk(KERN_DEBUG "%s: no inode\n", msg);
		return;
	}
	printk(KERN_DEBUG
	       "%s:%d inode addr=%p inode=%ld count=%d mode=%06o%s%s%s\n", msg,
	       line, i, i->i_ino, atomic_read(&i->i_count), i->i_mode,
	       is_inode_supermounted(i) ? " supermounted" : " normal",
	       (i->i_state & I_DIRTY) ? " dirty" : "", (is_inode_supermounted(i)
							&& is_inode_obsolete(i))
	       ? " obsolete" : "");
}

#define dump_inode(i)    _dump_inode (__FUNCTION__, __LINE__, i)

static inline void
_dump_dentry(char *msg, int line, struct dentry *d)
{
	if (!supermount_debug_enable)
		return;
	if (!d) {
		printk(KERN_DEBUG "%s:%d no dentry\n", msg, line);
		return;
	}
	printk(KERN_DEBUG
	       "%s:%d dentry addr=%p name=%s count=%d flags=%d parent=%p"
	       " sb=%p inode=%p\n", msg, line, d, d->d_name.name,
	       atomic_read(&d->d_count), d->d_flags, d->d_parent, d->d_sb,
	       d->d_inode);
	_dump_inode(msg, line, d->d_inode);
}

#define dump_dentry(d)    _dump_dentry (__FUNCTION__, __LINE__, d)

static inline void
_dump_file(char *msg, int line, struct file *file)
{
	struct file *subfile = get_subfile(file);
	if (!supermount_debug_enable)
		return;

	_dump_dentry(msg, line, file->f_dentry);
	if (!subfile)
		printk(KERN_DEBUG "%s:%d no subfile\n", msg, line);
	else
		_dump_dentry(msg, line, subfile->f_dentry);
}

#define dump_file(f)    _dump_file (__FUNCTION__, __LINE__, f)

static inline void
_dump_sbi(char *msg, int line, struct supermount_sb_info *sbi)
{
	char *state[] = { "unmounted", "suspended", "online" };
	if (!supermount_debug_enable)
		return;

	if (!sbi) {
		printk(KERN_DEBUG "%s:%d, no superblock\n", msg, line);
		return;
	}

	printk(KERN_DEBUG
	       "%s:%d sbi state=%s version=%d mode=%08o mnt=%p mnt_sb=%p open=%d write=%d%s%s%s\n",
	       msg, line, state[sbi->s_state], sbi->s_version,
	       sbi->s_default_mode, sbi->s_undermount,
	       (sbi->s_undermount ? sbi->s_undermount->mnt_sb : NULL),
	       atomic_read(&sbi->s_opencount), atomic_read(&sbi->s_writecount),
	       (sbi->s_dirty) ? ", dirty" : "", (sbi->s_rw) ? ", rw" : "",
	       atomic_read(&sbi->s_dismount) ? ", dismount" : "");
}

#define dump_sbi(s)    _dump_sbi (__FUNCTION__, __LINE__, s)

static inline void
_dump_operation(char *msg, int line, char *operation, struct inode *dir, struct
		dentry *dentry)
{
	if (!supermount_debug_enable)
		return;
	printk(KERN_DEBUG "%s:%d %s\n", msg, line, operation);
	_dump_inode(msg, line, dir);
	_dump_dentry(msg, line, dentry);
}

#define dump_operation(op, dir, dentry)  \
        _dump_operation(__FUNCTION__, __LINE__, op, dir, dentry)

static inline void
_dump_lowlevel(char *msg, int line, char *operation, struct dentry *dir, struct
	       dentry *dentry)
{
	if (!supermount_debug_enable)
		return;
	printk(KERN_DEBUG "%s:%d lowlevel %s\n", msg, line, operation);
	_dump_dentry(msg, line, dir);
	_dump_dentry(msg, line, dentry);
}

#define dump_lowlevel(op, dir, dentry)  \
        _dump_lowlevel(__FUNCTION__, __LINE__, op, dir, dentry)

#define supermount_debug dump
#else
#define supermount_debug(f, a...) /**/
#define dump_dentry(d)
#define dump_inode(i)
#define dump_file(f)
#define dump_sbi(s)
#define dump_operation(o, d, d2)
#define dump_lowlevel(o, d, d2)
#endif				/* DEBUG */
/* debug.c */
extern void supermount_error(struct super_block *, const char *,
			     const char *, ...)
    __attribute__ ((format(printf, 3, 4)));
extern NORET_TYPE void supermount_panic(struct super_block *, const char *,
					const char *, ...)
    __attribute__ ((NORET_AND format(printf, 3, 4)));
extern void supermount_warning(struct super_block *, const char *,
			       const char *, ...)
    __attribute__ ((format(printf, 3, 4)));
/* super.c */
extern struct inode *supermount_create_inode(struct super_block *s);
extern void subfs_umount(struct super_block *sb);
extern int subfs_go_online(struct super_block *);
extern void subfs_go_inactive(struct super_block *);
extern int subfs_get_write_access(struct super_block *);
extern void subfs_put_write_access(struct super_block *);
extern void subfs_get_read_access(struct super_block *);
extern void subfs_put_read_access(struct super_block *);
/* translate.c */
extern struct dentry *get_subfs_dentry(struct dentry *dentry);
extern struct dentry *get_subfs_dentry_by_inode(struct inode *inode);
/* mediactl.c */
extern void supermount_handle_door(struct super_block *, int);
static inline void
supermount_lock_door(struct super_block *sb)
{
	supermount_handle_door(sb, MEDIA_LOCK);
}
static inline void
supermount_unlock_door(struct super_block *sb)
{
	supermount_handle_door(sb, MEDIA_UNLOCK);
}

/* inode_operations.c */
extern struct inode_operations supermount_root_iops;
extern struct inode_operations supermount_dir_iops;
extern struct inode_operations supermount_file_iops;
extern struct inode_operations supermount_symlink_iops;
/* super_operations.c */
extern struct super_block *supermount_read_super(struct super_block *,
						 void *, int);
/* file_operations.c */
extern struct file_operations supermount_dir_operations;
extern struct file_operations supermount_file_operations;
/* dentry_operations.c */
extern struct dentry_operations supermount_file_dops;
extern struct dentry_operations supermount_dir_dops;

#endif				/* _SUPERMOUNT_I_H */
