/*+M*************************************************************************
 * Symbol Technologies Trilogy IEEE PCMCIA device driver for Linux.
 *
 * Copyright (c) 2000 Symbol Technologies Inc. -- http://www.symbol.com
 * All rights reserved.
 *
 * Developed for Symbol Technologies Inc. by TriplePoint, Inc.
 *   http://www.triplepoint.com
 *
 * Abstract:
 *
 *     This is a part of the driver for the Trilogy 802.11 11 Mbit
 *     wireless MAC controller as used in Symbol Technologies systems.
 *
 * Author:
 *
 *     Tim Gardner
 *
 *---------------------------------------------------------------------------
 * This driver supports the following features:
 *   - Hot plug/unplug
 *   - Access Point and Ad-Hoc (peer-to-peer) communication
 *   - Card power management
 *
 *   Refer to the manual page for additional configuration, feature, and
 *   support information.
 *---------------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * ALTERNATIVELY, this driver may be distributed under the terms of
 * the following license, in which case the provisions of this license
 * are required INSTEAD OF the GNU General Public License. (This clause
 * is necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *-M*************************************************************************/

#ifndef _TrilogyHARDWARE_
#define _TrilogyHARDWARE_

#include <linux/config.h>
#if CONFIG_SMP
#ifndef __SMP__
#define __SMP__
#endif
#endif

#if MODULE
#ifndef _S24_INIT_
#define __NO_VERSION__
#endif
#include <linux/module.h>
#else
#define MOD_INC_USE_COUNT
#define MOD_DEC_USE_COUNT
#endif

#ifdef PCMCIA_DEBUG
#define DBG     1
#endif // PCMCIA_DEBUG

#include <pcmcia/config.h>
#include <pcmcia/k_compat.h>

#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/ptrace.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/timer.h>
#include <linux/interrupt.h>
#include <linux/in.h>
#include <linux/delay.h>
#include <asm/io.h>
#include <asm/system.h>
#include <asm/bitops.h>

#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/skbuff.h>
#include <linux/if_arp.h>
#include <linux/ioport.h>

#ifdef HAS_WIRELESS_EXTENSIONS
#include <linux/wireless.h>
#define IEEE_WIRELESS
#endif // HAS_WIRELESS_EXTENSIONS

#include <pcmcia/version.h>
#include <pcmcia/cs_types.h>
#include <pcmcia/cs.h>
#include <pcmcia/cistpl.h>
#include <pcmcia/cisreg.h>
#include <pcmcia/ciscode.h>
#include <pcmcia/ds.h>
#include "debug.h"

typedef struct net_device DEVICE;

#ifndef offsetof
#define offsetof(TYPE, MEMBER) ((size_t) &((TYPE *)0)->MEMBER)
#endif

#define FIELD_OFFSET(type, field)    offsetof(type,field)

#define NELEM(arr) (sizeof(arr) / sizeof(arr[0]))

// The buffer size associated with the LTV record (allowing for the header
// information).
#define MAX_LTV_BUF_SIZE (512 - (sizeof(__u16) * 2))

#define S24T_TALLIES_SIZE (sizeof(CFG_HERMES_TALLIES_STRCT) + \
    (sizeof(__u16) * 2))

/*
 * Uncomment this macro when the firmware supports keylengths other then 5.
 */
#define SPECTRUM24T_KEYLENGTH       1

#define S24T_MAX_MULTICAST           16
#define S24T_MAX_PACKET_SIZE         1500
#define S24T_MAX_NAME_LEN            32
#define S24T_NUM_IO_PORTS            0x40
#define S24T_TX_TIMEOUT              ((400 * HZ) / 1000)

#define S24T_MIN_COMM_QUALITY        0
#define S24T_MAX_COMM_QUALITY        92
#define S24T_MIN_SIGNAL_LEVEL        47
#define S24T_MAX_SIGNAL_LEVEL        138
#define S24T_MIN_NOISE_LEVEL         47
#define S24T_MAX_NOISE_LEVEL         138

#define S24T_EJECTED                 0xffff

// Host <==> Adapter conversions
#define atohl(v) cpu_to_le32(v)
#define atohs(v) cpu_to_le16(v)
#define htoal(v) le32_to_cpu(v)
#define htoas(v) le16_to_cpu(v)

#if (LINUX_VERSION_CODE < VERSION(2,1,0))
typedef struct {
    volatile unsigned int lock;
} spinlock_t;
#define spin_lock_init(l) do { } while(0)

#ifndef CAP_NET_ADMIN
#  define CAP_NET_ADMIN     12
#  define capable(x)        suser()
#endif // CAP_NET_ADMIN

// All pre-2.1.0 kernels only support integers (32-bit) values or
// normal strings and no MODULE_PARM_DESC().
#define MODULE_PARM_DESC(a,b)
#define p_u8    __u32
#define p_s8    __s32
#define p_u16   __u32
#define p_s16   __s32
#define p_u32   __u32
#define p_s32   __s32
#define p_char  char
#else // LINUX_VERSION_CODE
#define p_u8    __u8
#define p_s8    __s8
#define p_u16   __u16
#define p_s16   __s16
#define p_u32   __u32
#define p_s32   __s32
#define p_char  char
#endif // LINUX_VERSION_CODE

#define DRIVER_NAME     "Spectrum24t"

typedef enum {
    FALSE = 0,
    TRUE  = 1
} bool_t;

typedef struct {
    __u16   WordLen;
    __u16   Rid;
}  __attribute__ ((packed)) S24tRidType, *PS24tRidType;

#define MAX_INFO_REC_DATA (512-sizeof(S24tRidType))
typedef struct {
    S24tRidType Rec;
    __u8        Data[MAX_INFO_REC_DATA];
}  __attribute__ ((packed)) S24tInfoRecType, *PS24tInfoRecType;

#define S24T_IOCTL_UIL            SIOCDEVPRIVATE
#define   S24T_UIL_CONNECT          0x01
#define   S24T_UIL_DISCONNECT       0x02
#define   S24T_UIL_BLOCK            0x03
#define   S24T_UIL_UNBLOCK          0x04
#define   S24T_UIL_ACTION_TALLIES   0x05
#define   S24T_UIL_ACTION_SCAN      0x06
#define   S24T_UIL_ACTION_DIAG      0x07
#define   S24T_UIL_SEND_DIAG_MSG    0x08
#define   S24T_UIL_GET_INFO         0x09
#define   S24T_UIL_PUT_INFO         0x0A
#define SIOCSIWNETNAME              SIOCDEVPRIVATE+1
#define SIOCGIWNETNAME              SIOCDEVPRIVATE+2
#define SIOCSIWSTANAME              SIOCDEVPRIVATE+3
#define SIOCGIWSTANAME              SIOCDEVPRIVATE+4
#define SIOCSIWPORTTYPE             SIOCDEVPRIVATE+5
#define SIOCGIWPORTTYPE             SIOCDEVPRIVATE+6

struct S24T_private {
    dev_node_t                  node;
    DEVICE                      *dev;
    spinlock_t                  slock;
    struct net_device_stats     stats;
    volatile __u8               *Cor;
#ifdef WIRELESS_EXT
    struct iw_statistics        wstats;
#endif // WIRELESS_EXT
    u_long                      txBytes;
    __u16                       txFid;
    __u16                       rxFid;

    __u8                        MACAddress[ETH_ALEN];
    char                        NetworkName[S24T_MAX_NAME_LEN+1]; /* ESS ID */
    char                        StationName[S24T_MAX_NAME_LEN+1];
    __u16                       LinkStatus; /* 1 means linked. */

#   define                      S24_FW_VERSION_LEN 14
    char                        FwVersion[S24_FW_VERSION_LEN+1];
    char                        FwDate[S24_FW_VERSION_LEN+1];
    /*
     * All 16 bit setup values can be treated in a reasonably generic manner.
     */
#   define                      AntennaDiversity        0
#   define                      PortType                1
#   define                      Channel                 2
#   define                      APDensity               3
#   define                      MaxDataLength           4
#   define                      RtsThreshHold           5
#   define                      TxRateControl           6
#   define                      PromiscuousMode         7
#   define                      TickTime                8
#   define                      FragmentationThreshHold 9
#   define                      EncryptionKeyId         10
#   define                      EncryptionEnabled       11
#   define                      EncryptionAuthentication 12
#   define                      EncryptionKeyLen        13
#   define                      PowerSaveMode           14
#   define                      ReceiveAllMCast         15
#   define                      MaxSleepDuration        16
#   define                      MaxParam                17

    __u16                       Param[MaxParam];


#define AUTHENTICATION_ALGORITHM_OPEN_SYSTEM    1
#define AUTHENTICATION_ALGORITHM_SHARED_KEY     2
#define AUTHENTICATION_ALGORITHM_SHARED_KEY_128 3 /* AP only */

#define MAX_NUM_KEY 4
#define MAX_KEY_LEN 13
    __u8                        EncryptionKey[MAX_NUM_KEY][MAX_KEY_LEN];

    S24tInfoRecType             InfoRec;

} S24tPrivate;

/*--------------------------------------------------------------------------*/
#ifndef PCMCIA_DEBUG
#define PCMCIA_DEBUG    0
#endif // PCMCIA_DEBUG

#undef  DBG_TRAP
#define DBG_TRAP         {}

#define VALIDATE_PARAMS

#define NIC_INTEN_MASK  (IEN_RX|IEN_ALLOC|IEN_INFO|IEN_TXEXC|IEN_TX)

/////////////////////////////////////////////////////////////////////
//
// Offsets from Adapter->IoPAddr of the ports used to access the Trilogy
// NIC Registers.
//
// Command/Status Registers

#define NIC_COMMAND     0x00
#define NIC_PARAM0      0x02
#define NIC_PARAM1      0x04
#define NIC_PARAM2      0x06
#define NIC_STATUS      0x08
#define NIC_RESP0       0x0A
#define NIC_RESP1       0x0C
#define NIC_RESP2       0x0E

// FID Management Registers

#define NIC_INFO_FID    0x10
#define NIC_RX_FID      0x20
#define NIC_ALLOC_FID   0x22
#define NIC_TX_COMP_FID 0x24

// BAP 0

#define NIC_SELECT0     0x18    // FID/RID 0
#define NIC_OFFSET0     0x1C
#define NIC_DATA0       0x36

// BAP 1

#define NIC_SELECT1     0x1A    // FID/RID 1
#define NIC_OFFSET1     0x1E
#define NIC_DATA1       0x38

// Event

#define NIC_EVSTAT      0x30
#define NIC_INTEN       0x32
#define NIC_EVACK       0x34

// Control

#define NIC_CONTROL     0x14

// Host Software

#define NIC_SW_SUPPORT0 0x28
#define NIC_SW_SUPPORT1 0x2A
#define NIC_SW_SUPPORT2 0x2C
#define NIC_SW_SUPPORT3 0x2E    // reset reason code

// Auxiliary port

#define NIC_AUX_BASE    0x3A
#define NIC_AUX_OFFSET  0x3C
#define NIC_AUX_DATA    0x3E

/////////////////////////////////////////////////////////////////////
//
// Constants for the NIC_COMMAND register.
//
// Controller/State

#define CR_INITIALIZE   0x0000
#define CR_ENABLE       0x0001
#define CR_DISABLE      0x0002
#define CR_DIAGNOSE     0x0003

// Buffer Management

#define CR_ALLOCATE     0x000A
#define CR_TRANSMIT     0x000B
#define CR_RECLAIM      0x0100

// Regulate

#define CR_NOTIFY       0x0010
#define CR_INQUIRE      0x0011

// Configure

#define CR_ACCESS       0x0021
#define CR_ACCESS_WRITE 0x0121
#define CR_DOWNLOAD     0x0022
#define DL_MODE_DIABLE  0x0000  // Disables AUX port access & makes the
                                // downloaded firmware active
#define DL_MODE_ENABLE_VP 0x0100  // Enables download of firmware image via the
                                // AUX port registers directly into volatile
                                // memory.
#define DL_MODE_ENABLE_NVP 0x0200  // Enables download of firmware  via the AUX
                                // port registers into a temp download buffer
                                // in volatile memory.
#define DL_MODE_FLUSH   0x0300  // Writes the download buffer contents into
                                // the destination area in non-volatile
                                // memory.

#define COMMAND_REGISTER_BUSY   0x8000  // Command Busy bit mask

/////////////////////////////////////////////////////////////////////
//
// BAP register defines

//#define OFFSET_REGISTER_ERROR 0x4000  // Offset register error
#define OFFSET_REGISTER_BUSY    0x8000  // Offset register busy
#define DEAD_OR_EJECTED         0xFFFF  // Dead or ejected card


/////////////////////////////////////////////////////////////////////
//
// Constants for EvStat Register

#define EVS_RX          0x0001  // Frame is received
#define EVS_TX          0x0002  // TX completed
#define EVS_TXEXC       0x0004  // TX unsuccessfully completed
#define EVS_ALLOC       0x0008  // Alloc or reclaim of frame structure buffer
#define EVS_CMD         0x0010  // Command execution is completed
#define EVS_INFO        0x0080  // Information frame is generated
#define EVS_INFO_DROP   0x2000  // Unsolicited frame structure alloc error
#define EVS_WAIT_ERROR  0x4000  // Wait timeout error in the adapter
#define EVS_TICK        0x8000  // Auxiliary timer tick


/////////////////////////////////////////////////////////////////////
//
// Constants for IntEn Register
//
// Specifies what events cause interrupts

#define IEN_RX          0x0001  // Frame reception
#define IEN_TX          0x0002  // TX completed
#define IEN_TXEXC       0x0004  // TX unsuccessfully completed
#define IEN_ALLOC       0x0008  // Alloc or reclaim of frame structure buffer
#define IEN_CMD         0x0010  // Command execution completion
#define IEN_INFO        0x0080  // Information frame is generation
#define IEN_INFO_DROP   0x2000  // Unsolicited frame structure alloc error
#define IEN_WAIT_ERROR  0x4000  // Wait timeout error in the adapter
#define IEN_TICK        0x8000  // Auxiliary timer tick

/////////////////////////////////////////////////////////////////////
//
// Constants for EvAck Register
//
// Specifies what events cause interrupts

#define ACK_RX          0x0001  // Frame reception
#define ACK_TX          0x0002  // TX completed
#define ACK_TXEXC       0x0004  // TX unsuccessfully completed
#define ACK_ALLOC       0x0008  // Alloc or reclaim of frame structure buffer
#define ACK_CMD         0x0010  // Command execution completion
#define ACK_INFO        0x0080  // Information frame is generation
#define ACK_INFO_DROP   0x2000  // Unsolicited frame structure alloc error
#define ACK_WAIT_ERROR  0x4000  // Wait timeout error in the adapter
#define ACK_TICK        0x8000  // Auxiliary timer tick

/////////////////////////////////////////////////////////////////////
//
// Constant for Control Register

#define CTL_AUX_EN      0x8000  // Enable auxiliary port access

/////////////////////////////////////////////////////////////////////
//
// Constant for RID identification
typedef struct _SUP_RANGE {
    __u16  role;
    __u16  id;
    __u16  variant;
    __u16  bottom;
    __u16  top;
}  __attribute__ ((packed)) SUP_RANGE;

// From AP firmware to report association change of status of stations (0xF201)

typedef struct _ASSOC_STATUS {
    __u16      assocStatus;
    __u8       staAddr[6];
    __u8       oldApAddr[6];          // Optional: only if status is ASSOC_STATUS_STA_REASSOCIATED
}  __attribute__ ((packed)) ASSOC_STATUS, *PASSOC_STATUS;

#define ASSOC_STATUS_STA_ASSOCIATED     1
#define ASSOC_STATUS_STA_REASSOCIATED   2
#define ASSOC_STATUS_STA_DISASSOCIATED  3

#define EVENT_CONNECT       1
#define EVENT_DISCONNECT    2
#define EVENT_ROAM          3
#define EVENT_ASSOCIATE     4
#define EVENT_UNASSOCIATE   5

#define ENCRYPT_FLAG_ENABLE     3
#define ENCRYPT_FLAG_DISABLE    0

typedef struct {
    __u16   Len;
    __u8    String[1];
} S24tByteStringType, *PS24tByteStringType;

typedef struct _RID_STRUCT {
    __u16  length;
    __u16  rid;
    union {
        __u16           __u16;
        __u16          linkStatus;
        ASSOC_STATUS    assocStatus;
        SUP_RANGE       supRange;
        __u16           data[250];
    } u;
}  __attribute__ ((packed)) RID_STRUCT, *PRID_STRUCT;

#define RID_INFO_HANDOVER_ADDR      0xF000
#define RID_INFO_COMM_TALLIES       0xF100
#define RID_INFO_SCAN_RESULTS       0xF101
#define RID_INFO_LINK_STATUS        0xF200
#define RID_INFO_ASSOCIATION_STATUS 0xF201

#define RID_CFG_PORT_TYPE           0xFC00
#define RID_CFG_MAC_ADDRESS         0xFC01
#define RID_CFG_SSID                0xFC02
#define RID_CFG_CHANNEL             0xFC03
#define RID_CFG_OWN_SSID            0xFC04
#define RID_CFG_OWN_ATIM            0xFC05
#define RID_CFG_SYSTEM_SCALE        0xFC06      // AP Density
#define RID_CFG_MAX_DATA_LENGTH     0xFC07
#define RID_CFG_WDS_ADDRESS         0xFC08
#define RID_CFG_PM_ENABLE           0xFC09
#define RID_CFG_PM_EPS              0xFC0A
#define RID_CFG_MULTICAST_RCV       0xFC0B
#define RID_CFG_MAX_SLEEP_TIME      0xFC0C
#define RID_CFG_PM_HOLDOVER_TIME    0xFC0D
#define RID_CFG_NAME                0xFC0E
#define RID_CFG_MKK_FLAG            0xFC0F
#define RID_CFG_PREFERRED_BSS_ID    0xFC20
#define RID_CFG_MANDATORY_BSS_ID    0xFC21
#define RID_CFG_AUTHENTICATION      0xFC22      // Symbol
#define RID_CFG_ENCRYPT_INDEX       0xFC23      // Symbol
#define RID_CFG_KEY1                0xFC24      // Symbol
#define RID_CFG_KEY2                0xFC25      // Symbol
#define RID_CFG_KEY3                0xFC26      // Symbol
#define RID_CFG_KEY4                0xFC27      // Symbol
#define RID_CFG_ENCRYPT_FLAG        0xFC28      // Symbol
#define RID_CFG_AUTH_TYPE           0xFC2A      // Symbol
#define RID_CFG_KEY_LENGTH          0xFC2B      // Symbol

#define RID_GROUP_ADDR              0xFC80
#define RID_CREATE_IBSS             0xFC81
#define RID_FRAG_THRESHOLD          0xFC82
#define RID_RTS_THRESHOLD           0xFC83
#define RID_TX_RATE                 0xFC84
#define RID_PROMISCUOUS_MODE        0xFC85

#define RID_DIVERSITY               0xFC87      // Symbol
#define RID_OPTIONS                 0xFC88      // Symbol
#define RID_IF_FREQUENCY            0xFC89      // Symbol
#define RID_BASIC_RATES             0xFC8A      // Symbol

#define RID_TX_INHIBIT              0xFCA5      // Symbol
#define RID_CLEAR_STATS             0xFCA6      // Symbol

#define RID_TICK_TIME               0xFCE0

#define RID_MAX_DOWNLOAD_TIME       0xFD00
#define RID_DOWNLOAD_BUFFER         0xFD01
#define RID_PRI_IDENTITY            0xFD02
#define RID_PRI_SUP_RANGE           0xFD03
#define RID_PRIMARY_VERSION         0xFD05
#define RID_PRIMARY_DATE            0xFD06

#define RID_NIC_SERIAL              0xFD0A
#define RID_NIC_FW_ID               0xFD0B

#define RID_CHANNEL_LIST            0xFD10
#define RID_REGULATORY              0xFD11  // regulatory domains
#define RID_TEMP_TYPE               0xFD12  // temperature type
#define RID_CIS                     0xFD13
#define RID_COUNTRY_CODE            0xFD14
#define RID_COUNTRY_NAME            0xFD15
#define RID_MKK_CALL_SIGN           0xFD16
#define RID_MKK_RSSI                0xFD17
#define RID_ORDINALS                0xFD18

#define RID_STA_IDENTITY            0xFD20
#define RID_STA_SUP_RANGE           0xFD21
#define RID_MFI_ACT_RANGE           0xFD22
#define RID_CFI_ACT_RANGE           0xFD23
#define RID_SECONDARY_VERSION       0xFD24  // Symbol
#define RID_SECONDARY_DATE          0xFD25  // Symbol

#define RID_PORT_STATUS             0xFD40
#define RID_CURRENT_SSID            0xFD41
#define RID_CURRENT_BSSID           0xFD42
#define RID_COMMS_QUALITY           0xFD43
#define RID_CURR_TX_RATE            0xFD44

#define RID_SHORT_RETRY             0xFD48  // short retry limit (no RTS)
#define RID_LONG_RETRY              0xFD49  // long retry limit (RTS/CTS)

#define RID_PRIVACY                 0xFD4F  // privacy availability

#define RID_PHY_TYPE                0xFDC0  // phy type
#define RID_CURRENT_CHANNEL         0xFDC1  // current channel
#define RID_POWER_MODE              0xFDC2  // power save mode
#define RID_CCA_MODE                0xFDC3  // CCA mode

#define RID_SUPPORTED_DATA_RATES    0xFDC6  // supported rates

#define RID_FW_VERSION              0xFFFF  // file names

#define TX_RATE_1_MBIT              1
#define TX_RATE_2_MBIT              2
#define TX_RATE_5_5_MBIT            4
#define TX_RATE_11_MBIT             8

#define DEFAULT_CHANNEL             1
#define DEFAULT_IO_BASE_ADDRESS     0x300
#define DEFAULT_MEMORY_SIZE         0x1000
#define DEFAULT_MEMORY_WINDOW       0xd4000
#define DEFAULT_RTS_THRESHOLD       1600
#define DEFAULT_FRAG_THRESHOLD      0
#define DEFAULT_TX_RATE             0xf
#define DEFAULT_INTERRUPT_NUMBER    3
#define MAX_IOBASEADDR              0xFFFF
#define MIN_IOBASEADDR              0x0100
#define MAX_IRQ                     15
#define MIN_IRQ                     2
#define MAX_MULTICAST_LIST          32

#define OFFSET_COR                  0x3e0   // COR offset within Common Memory
#define CS_CIS_MFG_NAME_OFFSET      0x16    // Manufacturer Name string , null terminated
#define CS_CIS_MFG_NAME_LEN         32
#define COR_RESET                   0x80
#define COR_IO_MODE                 0x41
#define COR_CLEAR                   0x0


// Timeout definitions

#define COMMAND_BUSY_TIMEOUT        100 // * 10 = 1000 microseconds

// Constant for Control Register

#define CTL_AUX_EN      0x8000  // Enable auxiliary port access
#define CTL_AUX_DISABLE 0x4000  // Disable to auxiliary port access
#define CTL_AUX_OPEN    0xC000  // Auxiliary port is open if CTL reg
                                // has this value.
// AUX port access keys.
// The AUX port is locked and to unlock it you have to write the
// following values to the PARAM0-1 registers then write a
// CTL_AUX_EN to the NIC_CONTROL register. When the NIC_CONTROL
// register is a CTL_AUX_OPEN then access to the auxiliary port
// is enabled.

#define AUX_PARAM0_KEY      0xFE01
#define AUX_PARAM1_KEY      0xDC23
#define AUX_PARAM2_KEY      0xBA45

#define TX_CTL_PORT0    0x0000
#define TX_CTL_PORT1    0x0100
#define TX_CTL_PORT2    0x0200
#define TX_CTL_PORT3    0x0300
#define TX_CTL_PORT4    0x0400
#define TX_CTL_PORT5    0x0500
#define TX_CTL_PORT6    0x0600

#define TX_CTL_802_3    0x0000          // Use 802.3 portion of header
#define TX_CTL_802_11   0x0008          // Use 802.11 portion of header

#define TX_CTL_TX_INT   0x0002          // Generate event on successfull TX
#define TX_CTL_TX_EX    0x0004          // Generate event on TX error

typedef struct _Trilogy_TX_CTL
{
    __u16      usReserved0;
    __u16      usReserved1;
    __u32       ulSwSupport;
    __u16      usReserved2;
    __u16      usTxControl;          // Type of TX (bits defined below)
}  __attribute__ ((packed)) Trilogy_TX_CTL, *PTrilogy_TX_CTL;

typedef struct _Trilogy_RX_CTL
{
    __u32       ulTime;
    __u8       ucSignal;
    __u8       ucSilence;
    __u8       ucReserved0;
    __u8       ucRate;
    __u16      usReserved1;
    __u16      usReserved2;
}  __attribute__ ((packed)) Trilogy_RX_CTL, *PTrilogy_RX_CTL;

typedef struct _HDR_802_3
{
    __u8    ucDestAddr[6];
    __u8    ucSrcAddr[6];
    __u16   usLength;               // Big Endian format
}  __attribute__ ((packed)) HDR_802_3, *PHDR_802_3;

#define ETHERNET_HEADER_SIZE sizeof(HDR_802_3)
#define MAX_ETHERNET_BODY_SIZE 1500
#define RFC1042_SIZE            8

typedef struct _TRILOGY_MAC_HDR
{
#define TX_STATUS_FORM_ERR              0x0008
#define TX_STATUS_DISCON                0x0004
#define TX_STATUS_AGED_ERR              0x0002
#define TX_STATUS_RETRY_ERR             0x0001

#define RX_STATUS_TYPE_NORMAL           0x0000
#define RX_STATUS_TYPE_RFC1042          0x2000
#define RX_STATUS_TYPE_BRIDGE_TUNNEL    0x4000
#define RX_STATUS_TYPE_WMP              0x6000
#define RX_STATUS_RCV_PORT0             0x0000
#define RX_STATUS_RCV_PORT1             0x0100
#define RX_STATUS_RCV_PORT2             0x0200
#define RX_STATUS_RCV_PORT3             0x0300
#define RX_STATUS_RCV_PORT4             0x0400
#define RX_STATUS_RCV_PORT5             0x0500
#define RX_STATUS_RCV_PORT6             0x0600
#define RX_STATUS_UNDECRYPT_ERR         0x0002  // Only in promiscuous mode
#define RX_STATUS_FCS_ERR               0x0001  // Only in promiscuous mode

#define RX_TUNNELED (RX_STATUS_TYPE_RFC1042|RX_STATUS_TYPE_BRIDGE_TUNNEL)

    __u16   usStatus;               // TX/RX status

    union
    {
        Trilogy_TX_CTL tx;
        Trilogy_RX_CTL rx;
    } u;

    // 802.11 Format header information
#define RX_FCTL_DATA                    0x0008

    __u16   usFrameControl;
    __u16   usDuration_id;
    __u8    ucAddress1[6];
    __u8    ucAddress2[6];
    __u8    ucAddress3[6];

    __u16   usSequenceControl;

    __u8    ucAddress4[6];
    __u16   usDataLen;              // Little-Endian format

    // 802.3 Format header information

    __u8    ucDestAddr[6];
    __u8    ucSrcAddr[6];
    __u16   usLength;               // Big Endian format
} __attribute__ ((packed)) TRILOGY_MAC_HDR, *PTRILOGY_MAC_HDR;

extern int S24tInit(struct S24T_private *lp);
extern void S24tHalt(struct S24T_private *lp);
extern int S24tEnable( struct S24T_private *lp);
extern int S24tDisable( struct S24T_private *lp);
extern int S24tIrqEnable( struct S24T_private *lp);
extern int S24tIrqDisable( struct S24T_private *lp);
extern int S24tEvStat( struct S24T_private *lp,__u16 *EvStat);
extern int S24tTx( struct S24T_private *lp, __u8 *Buff, __u32 Len);
extern struct sk_buff * S24tRx( struct S24T_private *lp);
extern int S24tGetTxFid( struct S24T_private *lp);
extern int S24tEvAck(  struct S24T_private *lp, __u16 EvAck);
extern int S24tSetMulticast(struct S24T_private *lp,__u32 McBytes,__u8 *Mc);
extern int S24tPromiscuousMode(struct S24T_private *lp,__u16 Enabled);
extern int S24tCardInserted(struct S24T_private *lp);
extern int S24tGetInfoRec(struct S24T_private *lp);
extern int S24tSetEssId(struct S24T_private *lp);
extern int S24tGeneralInfo(struct S24T_private *lp);

#endif  // _TrilogyHARDWARE_

