/*+M*************************************************************************
 * Symbol Technologies Trilogy IEEE PCMCIA device driver for Linux.
 *
 * Copyright (c) 2000 Symbol Technologies Inc. -- http://www.symbol.com
 * All rights reserved.
 *
 * Developed for Symbol Technologies Inc. by TriplePoint, Inc.
 *   http://www.triplepoint.com
 *
 *---------------------------------------------------------------------------
 * This driver supports the following features:
 *   - Hot plug/unplug
 *   - Access Point and Ad-Hoc (peer-to-peer) communication
 *   - Card power management
 *
 *   Refer to the manual page for additional configuration, feature, and
 *   support information.
 *---------------------------------------------------------------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * ALTERNATIVELY, this driver may be distributed under the terms of
 * the following license, in which case the provisions of this license
 * are required INSTEAD OF the GNU General Public License. (This clause
 * is necessary due to a potential bad interaction between the GPL and
 * the restrictions contained in a BSD-style copyright.)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, and the entire permission notice in its entirety,
 *    including the disclaimer of warranties.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote
 *    products derived from this software without specific prior
 *    written permission.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT,
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *-M*************************************************************************/
#include "Spectrum24t.h"

typedef struct {
    __u16   Rid;
    __u32   ParamOffset;
    char    *Name;
} S24tGenricRidSetupType, *PS24tGenricRidSetupType;

#if DBG
static drv_info_t *DbgInfo = &S24tInfo;
STATIC char *S24tRidDebug(__u16 Rid);
#endif

STATIC __u16 S24tRead16(struct S24T_private *lp,__u16 offset);
STATIC void S24tReads(struct S24T_private *lp,__u8 *dest,__u16 offset,__u32 len);
STATIC void S24tWrite16(struct S24T_private *lp,__u16 offset,__u16 val);
STATIC void S24tWrites(struct S24T_private *lp,__u8 *src,__u16 offset,__u32 len);
STATIC int S24tCommand(struct S24T_private *lp,__u16 Command,__u16 *Result,__u16 *P0,__u16 *P1,__u16 *P2);
STATIC int S24tReadMacAddress(struct S24T_private *lp);
STATIC void S24tParseInfoRec(struct S24T_private *lp);
STATIC int S24tSetMUEncryptionAlgorithm(struct S24T_private *lp);
STATIC int S24tGenericInit(struct S24T_private *lp);
STATIC int S24tSetStationName(struct S24T_private *lp);
STATIC int S24tWriteRidSetupRec(struct S24T_private *lp,PS24tGenricRidSetupType p);
STATIC int S24tSetMacAddress(struct S24T_private *lp);
STATIC int S24tGetFirmwareVersion(struct S24T_private *lp);

int
S24tCardInserted(
    struct S24T_private *lp
    )
{
    if (lp->Cor && (lp->Cor[OFFSET_COR] == COR_IO_MODE))
    {
        return(1);
    }
    else
    {
        DBG_PRINT("Spectrum24t not inserted %02x\n",(uint)lp->Cor[OFFSET_COR]);
        return(0);
    }
}

int
S24tInit(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tInit")
    int Status;
    int i;
    __u8 Mfg[CS_CIS_MFG_NAME_LEN];

    DBG_ENTER(DbgInfo);
    DBG_ASSERT(lp->Cor);
    DBG_ASSERT(sizeof(TRILOGY_MAC_HDR) == 60);

    if (!S24tCardInserted(lp))
    {
        DBG_LEAVE(DbgInfo);
        return(-ENODEV);
    }

    Mfg[CS_CIS_MFG_NAME_LEN-1] = '\0';
    for (i=0; i<(CS_CIS_MFG_NAME_LEN-1); i++)
    {
        Mfg[i] = lp->Cor[(CS_CIS_MFG_NAME_OFFSET+i)*2];
    }

    DBG_NOTICE(DbgInfo, "MfgName '%s', Mode %02x\n",Mfg,lp->Cor[OFFSET_COR]);

    if ((Status = S24tCommand(lp,CR_INITIALIZE,NULL,NULL,NULL,NULL)))
    {
        DBG_PRINT("Failed to initialize.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }
    mdelay(200);
    S24tWrite16(lp,NIC_EVACK,0xffff);

    /*
     * Set the MAC address from the input options.
     */
    if (lp->MACAddress[0] | lp->MACAddress[1] | lp->MACAddress[2])
    {
        if ((Status = S24tSetMacAddress(lp)))
        {
            DBG_PRINT("Failed to set MAC address.\n");
            DBG_LEAVE(DbgInfo);
            return(Status);
        }
    }

    if ((Status=S24tReadMacAddress(lp)))
    {
        DBG_PRINT("Failed to read MAC address.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    if ((Status = S24tSetMUEncryptionAlgorithm(lp)))
    {
        DBG_PRINT("Failed to set encryption algorithm.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    if ((Status=S24tGenericInit(lp)))
    {
        DBG_PRINT("Failed generic init.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    if ((Status = S24tSetEssId(lp)))
    {
        DBG_PRINT("Failed to set ESS.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    if ((Status = S24tSetStationName(lp)))
    {
        DBG_PRINT("Failed to set station name.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    if ((Status = S24tGetFirmwareVersion(lp)))
    {
        DBG_PRINT("Failed to get firmware version.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }
    DBG_LEAVE(DbgInfo);
    return(0);
}

int
S24tEnable(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tEnable")
    int Status;
    __u16   P0;

    DBG_ENTER(DbgInfo);

    if ((Status = S24tCommand(lp,CR_ENABLE,NULL,NULL,NULL,NULL)))
    {
        DBG_PRINT("Failed to enable.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }
    mdelay(5);

    P0 = 1585;
    if ((Status = S24tCommand(lp,CR_ALLOCATE,NULL,&P0,NULL,NULL)))
    {
        DBG_PRINT("S24tEnable - failed to start CR_ALLOCATE.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    Status = S24tIrqEnable(lp);

    DBG_LEAVE(DbgInfo);
    return(Status);
}

int
S24tDisable(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tDisable")
    int Status;

    DBG_ENTER(DbgInfo);

    S24tIrqDisable(lp);
    if ((Status = S24tCommand(lp,CR_DISABLE,NULL,NULL,NULL,NULL)))
    {
        DBG_PRINT("S24tEnable - failed to disable.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    DBG_LEAVE(DbgInfo);
    return(Status);
}

int
S24tIrqEnable(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tIrqEnable")

    DBG_ENTER(DbgInfo);

    if (!S24tCardInserted(lp))
    {
        DBG_LEAVE(DbgInfo);
        return(-ENXIO);
    }
    S24tWrite16(lp,NIC_INTEN,NIC_INTEN_MASK);

    DBG_ENTER(DbgInfo);
    return(0);
}

int
S24tIrqDisable(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tIrqDisable")

    DBG_ENTER(DbgInfo);

    if (!S24tCardInserted(lp))
    {
        DBG_LEAVE(DbgInfo);
        return(-ENXIO);
    }
    S24tWrite16(lp,NIC_INTEN,0);

    DBG_LEAVE(DbgInfo);
    return(0);
}

int
S24tEvStat(
    struct S24T_private *lp,
    __u16               *EvStat
    )
{
    if (!S24tCardInserted(lp))
    {
        return(-ENXIO);
    }
    *EvStat = S24tRead16(lp,NIC_EVSTAT);

    return(0);
}

int
S24tEvAck(
    struct S24T_private *lp,
    __u16               EvAck
    )
{
    DBG_FUNC("S24tEvAck")

    DBG_ENTER(DbgInfo);

    if (!S24tCardInserted(lp))
    {
        DBG_LEAVE(DbgInfo);
        return(-ENXIO);
    }
    S24tWrite16(lp,NIC_EVACK,EvAck);

    DBG_LEAVE(DbgInfo);
    return(0);
}

int
S24tGetTxFid(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tGetTxFid")

    DBG_ENTER(DbgInfo);

    if (!S24tCardInserted(lp))
    {
        DBG_LEAVE(DbgInfo);
        return(-ENXIO);
    }

    lp->txFid = S24tRead16(lp,NIC_ALLOC_FID);

    DBG_TX(DbgInfo, "TxFid %04x\n",lp->txFid);

    DBG_LEAVE(DbgInfo);
    return(0);
}

STATIC int
S24tCopyDown(
    struct S24T_private *lp,
    __u16               Fid,
    __u8                *Buff1,
    __u32               Len1,
    __u8                *Buff2,
    __u32               Len2,
    __u8                *Buff3,
    __u32               Len3,
    __u32               Command
    )
{
    DBG_FUNC("S24tCopyDown")
    int Status = 0;
    __u16 Offset;

    DBG_ENTER(DbgInfo);

    Offset = S24tRead16(lp,NIC_OFFSET0);
    if (Offset & OFFSET_REGISTER_BUSY)
    {
        DBG_PRINT("S24tCopyDown offset reg busy.\n");
        DBG_LEAVE(DbgInfo);
        return(-EBUSY);
    }

    S24tWrite16(lp,NIC_SELECT0,Fid);
    S24tWrite16(lp,NIC_OFFSET0,0);

    do {
        Offset = S24tRead16(lp,NIC_OFFSET0);
    } while ((S24tCardInserted(lp)) && (Offset & OFFSET_REGISTER_BUSY));

    S24tWrites(lp,Buff1,NIC_DATA0,Len1);

    if (Buff2 && Len2)
    {
        DBG_ASSERT(!(Len1 & 1));
        S24tWrites(lp,Buff2,NIC_DATA0,Len2);

        if (Buff3 && Len3)
        {
            DBG_ASSERT(!(Len2 & 1));
            S24tWrites(lp,Buff3,NIC_DATA0,Len3);
        }
    }

    if (Command)
    {
        Status = S24tCommand(lp,Command,NULL,&Fid,NULL,NULL);
    }
    DBG_LEAVE(DbgInfo);
    return(Status);
}

STATIC int
S24tCopyUp(
    struct S24T_private *lp,
    __u16               Rid,
    __u8                *Buff,
    __u32               Len,
    __u16               Command
    )
{
    DBG_FUNC("S24tCopyUp")

    int Status;
    __u16 OffsetReg;
    __u16 tRid;

    DBG_ENTER(DbgInfo);

    DBG_ASSERT(!(Len & 1));

    if ((!Buff) || (!Len))
    {
        DBG_LEAVE(DbgInfo);
        return(0);
    }

    if (Command)
    {
        tRid = Rid;
        Status =
        S24tCommand(
            lp,
            Command,
            NULL,
            &tRid,
            NULL,
            NULL
        );
        if (Status)
        {
            DBG_PRINT("S24tCopyUp - CR_ACCESS failed.\n");
            DBG_LEAVE(DbgInfo);
            return(Status);
        }
    }

    OffsetReg = S24tRead16(lp,NIC_OFFSET1);
    if (OffsetReg & OFFSET_REGISTER_BUSY)
    {
        DBG_PRINT("S24tCopyUp - Offset reg busy.\n");
        DBG_LEAVE(DbgInfo);
        return(-EBUSY);
    }

    S24tWrite16(lp,NIC_SELECT1,Rid);
    S24tWrite16(lp,NIC_OFFSET1,0);

    do {
        OffsetReg = S24tRead16(lp,NIC_OFFSET1);
        if (!(OffsetReg & OFFSET_REGISTER_BUSY))
        {
            break;
        }
        udelay(10);
    } while (S24tCardInserted(lp));

    S24tReads(lp,Buff,NIC_DATA1,Len);

    DBG_LEAVE(DbgInfo);
    return(0);
}

STATIC void
S24tCopyUpSecond(
    struct S24T_private *lp,
    __u8                *Buff,
    __u32               Len
    )
{
    S24tReads(lp,Buff,NIC_DATA1,Len);
}

#define H802_11_SNAP_LEN 8
//STATIC __u8 SnapBridgeTunnel[H802_11_SNAP_LEN] = { 0x00, 0x00, 0xAA, 0xAA, 0x03, 0x00, 0x00, 0xF8 };
STATIC __u8 Snap802_1H[H802_11_SNAP_LEN] =       { 0x00, 0x00, 0xAA, 0xAA, 0x03, 0x00, 0x00, 0x00 };

int
S24tTx(
    struct S24T_private *lp,
    __u8                *Buff,
    __u32               Len
    )
{
    DBG_FUNC("S24tTx")
    PHDR_802_3 Hdr;
    __u16 TypeOrLength;
    __u8 *Buff1;
    __u32 Len1;
    __u8 *Buff2;
    __u32 Len2;
    __u8 *Buff3;
    __u32 Len3;

    TRILOGY_MAC_HDR TrilogyHeader;
    int Status;

    DBG_ENTER(DbgInfo);

    if (!S24tCardInserted(lp))
    {
        DBG_LEAVE(DbgInfo);
        return(-ENXIO);
    }

    if (!lp->txFid)
    {
        return(-EBUSY);
    }

    DBG_ASSERT(ETHERNET_HEADER_SIZE == 14);

    if (Len < ETHERNET_HEADER_SIZE)
    {
        return(-EINVAL);
    }

    Hdr = (PHDR_802_3)Buff;

    memset(&TrilogyHeader,0,sizeof(TRILOGY_MAC_HDR));

    /*
     * Tell the firmware to interrupt on completion or error.
     */
    TrilogyHeader.u.tx.usTxControl = TX_CTL_802_3|TX_CTL_TX_EX|TX_CTL_TX_INT;

    /*
     * Get the ethernet header.
     */
    memcpy(TrilogyHeader.ucDestAddr,Hdr,ETHERNET_HEADER_SIZE);

    /*
     * If the type length field is a type, then insert a SNAP header. Make
     * the header type length field a length.
     */
    TypeOrLength = __be16_to_cpu(Hdr->usLength);

    if (TypeOrLength > MAX_ETHERNET_BODY_SIZE)
    {
        __u16 PayloadLen;

        /*
         * The SNAP buffer contains the new payload.
         */
        Buff1 = (__u8 *)&TrilogyHeader;
        Len1 = sizeof(TRILOGY_MAC_HDR)-sizeof(__u16);

        /*
         * Compute a new payload length that includes the SNAP header.
         */
        PayloadLen = Len - ETHERNET_HEADER_SIZE;
        PayloadLen += H802_11_SNAP_LEN;

        *(__u16 *)Snap802_1H = __cpu_to_be16(PayloadLen);

        Buff2 = Snap802_1H;
        Len2 = H802_11_SNAP_LEN;

        /*
         * Point to the original packet type and payload.
         */
        Buff3 = Buff + (2*ETH_ALEN);
        Len3 = Len - (2*ETH_ALEN);

        /*
         * Make Len 3 an even number.
         */
        Len3++;
        Len3 &= (~1);
    }
    else
    {
        Buff1 = (__u8 *)&TrilogyHeader;
        Len1 = sizeof(TRILOGY_MAC_HDR);

        /*
         * Point to the original payload.
         */
        Buff2 = Buff + ETHERNET_HEADER_SIZE;
        Len2 = Len - ETHERNET_HEADER_SIZE;

        /*
         * Make Len 2 an even number.
         */
        Len2++;
        Len2 &= (~1);

        Buff3 = NULL;
        Len3 = 0;
    }

    Status =
    S24tCopyDown(
        lp,
        lp->txFid,
        Buff1,
        Len1,
        Buff2,
        Len2,
        Buff3,
        Len3,
        CR_TRANSMIT|CR_RECLAIM
    );

    /*
     * Can't transmit again until the interrupt handler recovers the TX FID.
     */
    lp->txFid = 0;

    DBG_LEAVE(DbgInfo);
    return(Status);
}

struct sk_buff *
S24tRx(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tRx")

    int Status;
    TRILOGY_MAC_HDR TrilogyHeader;
    struct sk_buff *Skb=NULL;
    __u32 SkbLen;

    DBG_ENTER(DbgInfo);

    lp->rxFid = S24tRead16(lp,NIC_RX_FID);

    Status =
    S24tCopyUp(
        lp,
        lp->rxFid,
        (__u8 *)&TrilogyHeader,
        sizeof(TRILOGY_MAC_HDR)-sizeof(HDR_802_3),
        0
    );
    if (Status)
    {
        DBG_PRINT("S24tRx - header read error.\n");
        DBG_LEAVE(DbgInfo);
        return(NULL);
    }

    if (TrilogyHeader.usDataLen > (MAX_ETHERNET_BODY_SIZE+H802_11_SNAP_LEN))
    {
        DBG_PRINT("Rx %u too large\n",(uint)TrilogyHeader.usDataLen);
        return(NULL);
    }
    SkbLen = TrilogyHeader.usDataLen+sizeof(HDR_802_3);

    if (!(Skb = ALLOC_SKB(SkbLen+2)))
    {
        DBG_PRINT("Could not allocate RX SKB len %u\n",(uint)SkbLen+2);
        DBG_LEAVE(DbgInfo);
        return(NULL);
    }
    skb_reserve(Skb, 2);

    if (TrilogyHeader.usStatus & RX_TUNNELED)
    {
        S24tCopyUpSecond(lp,Skb->data,2*ETH_ALEN);
        S24tCopyUpSecond(lp,Skb->data+(2*ETH_ALEN),8);
        S24tCopyUpSecond(lp,Skb->data+(2*ETH_ALEN),SkbLen - ((2*ETH_ALEN)+8) );

        SkbLen -= 8;
    }
    else
    {
        S24tCopyUpSecond(lp,Skb->data,SkbLen);
    }

    skb_put(Skb,SkbLen);

    DBG_LEAVE(DbgInfo);
    return(Skb);
}

STATIC int
S24tSetBinaryRec(
    struct S24T_private *lp,
    __u32               Rid,
    __u8                *Buff,
    __u32               Len
    )
{
    DBG_FUNC("S24tSetBinaryRec")

    int Status;
    S24tRidType Rec;

    DBG_ENTER(DbgInfo);

    memset(&Rec,0,sizeof(S24tRidType));

    Rec.WordLen = 1 + ((Len+1) / 2);
    Rec.Rid = Rid;

    Status =
    S24tCopyDown(
        lp,
        Rid,
        (__u8 *)&Rec,
        sizeof(S24tRidType),
        Buff,
        Len,
        NULL,
        0,
        CR_ACCESS_WRITE
    );

    DBG_LEAVE(DbgInfo);
    return(Status);
}

int
S24tSetMulticast(
    struct S24T_private *lp,
    __u32               McBytes,
    __u8                *Mc
    )
{
    DBG_FUNC("S24tSetMulticast")

    int Status;

    DBG_ENTER(DbgInfo);

    Status = S24tSetBinaryRec(lp,RID_GROUP_ADDR,Mc,McBytes);

    DBG_LEAVE(DbgInfo);
    return(Status);
}

STATIC int
S24tSetMacAddress(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tSetMacAddress")

    int Status;

    DBG_ENTER(DbgInfo);

    Status = S24tSetBinaryRec(lp,RID_CFG_MAC_ADDRESS,lp->MACAddress,ETH_ALEN);

    DBG_LEAVE(DbgInfo);
    return(Status);
}

int
S24tPromiscuousMode(
    struct S24T_private *lp,
    __u16               Enabled
    )
{
    DBG_FUNC("S24tPromiscuousMode")

    int Status;

    DBG_ENTER(DbgInfo);

    lp->Param[PromiscuousMode] = Enabled;

    Status = S24tSetBinaryRec(lp,RID_PROMISCUOUS_MODE,(__u8 *)&Enabled,sizeof(Enabled));

    DBG_LEAVE(DbgInfo);
    return(Status);
}

int
S24tGetInfoRec(
    struct S24T_private *lp
    )
{
    __u16 Rid;
    int Status;

    Rid = S24tRead16(lp,NIC_INFO_FID);

    Status =
    S24tCopyUp(
        lp,
        Rid,
        (__u8 *)&lp->InfoRec.Rec,
        sizeof(S24tRidType),
        0
    );
    if (Status)
    {
        DBG_PRINT("S24tGetInfoRec failed to get record header\n");
        return(Status);
    }

    if (
        (!lp->InfoRec.Rec.WordLen)
        ||
        ((lp->InfoRec.Rec.WordLen-1) > (MAX_INFO_REC_DATA*2))
       )
    {
        DBG_PRINT("Bogus info record length %u, type %04x\n",
            (uint)lp->InfoRec.Rec.WordLen,
            (uint)lp->InfoRec.Rec.Rid
        );
        return(-ENXIO);
    }

    S24tCopyUpSecond(
        lp,
        lp->InfoRec.Data,
        (lp->InfoRec.Rec.WordLen-1)*2
    );

    S24tParseInfoRec(lp);
    return(0);
}

STATIC int
S24tSetString(
    struct S24T_private *lp,
    __u16 Rid,
    __u16 RidLen,
    char *Str
    )
{
    DBG_FUNC("S24tSetString")

    int Status;
    PS24tByteStringType S;

    memset(lp->InfoRec.Data,0,MAX_INFO_REC_DATA);
    lp->InfoRec.Rec.WordLen = RidLen;
    lp->InfoRec.Rec.Rid = Rid;

    S = (PS24tByteStringType)lp->InfoRec.Data;

    S->Len = strlen(Str);

    if (S->Len)
    {
        strcpy(S->String,Str);
    }

    Status =
    S24tCopyDown(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        NULL,
        0,
        NULL,
        0,
        CR_ACCESS_WRITE
    );
    if (Status)
    {
        DBG_PRINT("Could not set %s to %s.\n",S24tRidDebug(Rid),Str);
        return(Status);
    }
    else
    {
        DBG_NOTICE(DbgInfo,"Set %s to %s\n",S24tRidDebug(Rid),Str);
    }

    if (!Status)
    {
        memset(lp->InfoRec.Data,0,MAX_INFO_REC_DATA);

        Status =
        S24tCopyUp(
            lp,
            Rid,
            (__u8 *)&lp->InfoRec,
            (RidLen+1)*2,
            CR_ACCESS
        );
        if (Status)
        {
            DBG_PRINT("Could not read %s\n",S24tRidDebug(Rid));
        }
        else
        {
            DBG_NOTICE(DbgInfo,"%s %u '%s'\n",S24tRidDebug(Rid),(uint)S->Len,S->String);
        }
    }

    return(Status);
}

STATIC int
S24tSetStationName(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tSetStationName")
    int Status;
    DBG_ENTER(DbgInfo);
    Status = S24tSetString(lp,RID_CFG_NAME,18,lp->StationName);
    DBG_LEAVE(DbgInfo);
    return(Status);
}

int
S24tSetEssId(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tSetEssId")
    int Status;
    DBG_ENTER(DbgInfo);
    Status = S24tSetString(lp,RID_CFG_SSID,18,lp->NetworkName);
    DBG_LEAVE(DbgInfo);
    return(Status);
}

STATIC S24tGenricRidSetupType RidSetup[] = {
    { RID_DIVERSITY,            AntennaDiversity,           "Antenna Diversity" },
    { RID_CFG_PORT_TYPE,        PortType,                   "Port Type" },
    { RID_CFG_CHANNEL,          Channel,                    "Channel" },
    { RID_CFG_SYSTEM_SCALE,     APDensity,                  "AP Density" },
    { RID_CFG_MAX_DATA_LENGTH,  MaxDataLength,              "MaxDataLength" },
    { RID_RTS_THRESHOLD,        RtsThreshHold,              "RtsThreshHold" },
    { RID_TX_RATE,              TxRateControl,              "TxRateControl" },
    { RID_PROMISCUOUS_MODE,     PromiscuousMode,            "PromiscuousMode" },
    { RID_TICK_TIME,            TickTime,                   "TickTime" },
    { RID_FRAG_THRESHOLD,       FragmentationThreshHold,    "FragmentationThreshHold" },
    /* { RID_CFG_PM_ENABLE,        PowerSaveMode,              "PowerSaveMode" }, */
    { RID_CFG_MULTICAST_RCV,    ReceiveAllMCast,            "ReceiveAllMCast" },
    { RID_CFG_MAX_SLEEP_TIME,   MaxSleepDuration,          "MaxSleepDuration" },
    { 0, 0, "" }
};

STATIC int
S24tWriteRidSetupRec(
    struct S24T_private     *lp,
    PS24tGenricRidSetupType p
    )
{
    DBG_FUNC("S24tWriteRidSetupRec")

    int Status=0;

    DBG_ENTER(DbgInfo);

    while (p->Rid)
    {
        DBG_ASSERT(p->ParamOffset < MaxParam);

        Status = S24tSetBinaryRec(lp,p->Rid,(__u8 *)&lp->Param[p->ParamOffset],sizeof(__u16));

        if (Status)
        {
            DBG_PRINT("Could not set %s.\n",p->Name);
            DBG_LEAVE(DbgInfo);
            return(Status);
        }
        else
        {
            DBG_NOTICE(DbgInfo,"Set %s to %u\n",p->Name,(uint)*(__u16 *)lp->InfoRec.Data);
        }
        p++;
    }
    DBG_LEAVE(DbgInfo);
    return(Status);
}

STATIC int
S24tGenericInit(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tGenericInit")
    int Status=0;

    DBG_ENTER(DbgInfo);
    Status = S24tWriteRidSetupRec(lp,RidSetup);
    DBG_LEAVE(DbgInfo);
    return(Status);
}

STATIC S24tGenricRidSetupType EncryptionEnabledSetup[] = {
    { RID_CFG_ENCRYPT_INDEX,    EncryptionKeyId,            "EncryptionKeyId" },
    // KeyLength is a read only field designed to inform the driver which keylengths are supported
    // Trying to set it will generate an error
    //   { RID_CFG_KEY_LENGTH,       EncryptionKeyLen,           "EncryptionKeyLen" },
    //
    { RID_CFG_ENCRYPT_FLAG,     EncryptionEnabled,          "EncryptionEnabled" },
    /*
     * Authentication must come after all other encryption setting.
     */
    { RID_CFG_AUTH_TYPE,        EncryptionAuthentication,   "EncryptionAuthentication" },
    { 0, 0, "" }
};

STATIC S24tGenricRidSetupType EncryptionDisabledSetup[] = {
    { RID_CFG_ENCRYPT_FLAG,     EncryptionEnabled,          "EncryptionEnabled" },
    { 0, 0, "" }
};

STATIC int
S24tSetMUEncryptionAlgorithm(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tSetMUEncryptionAlgorithm")

    int Status=0;
    __u32 i;

    DBG_ENTER(DbgInfo);

    if (lp->Param[EncryptionEnabled] == ENCRYPT_FLAG_DISABLE)
    {
        Status = S24tWriteRidSetupRec(lp,EncryptionDisabledSetup);
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    Status = S24tWriteRidSetupRec(lp,EncryptionEnabledSetup);
    if (Status)
    {
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    for (i=0; i<MAX_NUM_KEY; i++)
    {
        Status = S24tSetBinaryRec(lp,RID_CFG_KEY1+i,lp->EncryptionKey[i],lp->Param[EncryptionKeyLen]);
        if (Status)
        {
            DBG_PRINT("Could not set encryption key %u.\n",i);
        }
        else
        {
            DBG_NOTICE(DbgInfo,"Set encryption key %u to %s\n",i,lp->EncryptionKey[i]);
        }
    }


    DBG_LEAVE(DbgInfo);
    return(Status);
}


int
S24tGeneralInfo(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tGeneralInfo")

    int Status;
    PS24tByteStringType S;

    S = (PS24tByteStringType)lp->InfoRec.Data;

    memset(&lp->InfoRec,0,sizeof(S24tInfoRecType));

    lp->InfoRec.Rec.WordLen = 2;
    lp->InfoRec.Rec.Rid = RID_PORT_STATUS;

    Status =
    S24tCopyUp(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        CR_ACCESS
    );
    if (!Status)
    {
        DBG_NOTICE(DbgInfo,"Port Status %u\n", *(__u16 *)lp->InfoRec.Data);
    }
    else
    {
        return(Status);
    }

    memset(&lp->InfoRec,0,sizeof(S24tInfoRecType));

    lp->InfoRec.Rec.WordLen = 18;
    lp->InfoRec.Rec.Rid = RID_CURRENT_SSID;

    Status =
    S24tCopyUp(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        CR_ACCESS
    );
    if (!Status)
    {
        DBG_NOTICE(DbgInfo,"Current SSID len %u '%s'\n",(uint)S->Len,S->String);
    }
    else
    {
        return(Status);
    }

    memset(&lp->InfoRec,0,sizeof(S24tInfoRecType));

    lp->InfoRec.Rec.WordLen = 4;
    lp->InfoRec.Rec.Rid = RID_CURRENT_BSSID;

    Status =
    S24tCopyUp(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        CR_ACCESS
    );
    if (!Status)
    {
        DBG_NOTICE(DbgInfo,"Current BSSID %02x:%02x:%02x:%02x:%02x:%02x\n",
            lp->InfoRec.Data[0],
            lp->InfoRec.Data[1],
            lp->InfoRec.Data[2],
            lp->InfoRec.Data[3],
            lp->InfoRec.Data[4],
            lp->InfoRec.Data[5]
        );
    }
    else
    {
        return(Status);
    }

    memset(&lp->InfoRec,0,sizeof(S24tInfoRecType));

    lp->InfoRec.Rec.WordLen = 4;
    lp->InfoRec.Rec.Rid = RID_COMMS_QUALITY;

    Status =
    S24tCopyUp(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        CR_ACCESS
    );
    if (!Status)
    {
        DBG_NOTICE(DbgInfo,"Comms Quality CQ %u ASL %u ANL %u\n",
            (uint) *(__u16 *)&lp->InfoRec.Data[0],
            (uint) *(__u16 *)&lp->InfoRec.Data[2],
            (uint) *(__u16 *)&lp->InfoRec.Data[4]
        );
    }
    else
    {
        return(Status);
    }

    memset(&lp->InfoRec,0,sizeof(S24tInfoRecType));

    lp->InfoRec.Rec.WordLen = 2;
    lp->InfoRec.Rec.Rid = RID_CURR_TX_RATE;

    Status =
    S24tCopyUp(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        CR_ACCESS
    );
    if (!Status)
    {
        DBG_NOTICE(DbgInfo,"Current Tx Rate %u\n",
            (uint) *(__u16 *)&lp->InfoRec.Data[0]
        );
    }
    else
    {
        return(Status);
    }

    memset(&lp->InfoRec,0,sizeof(S24tInfoRecType));

    lp->InfoRec.Rec.WordLen = 2;
    lp->InfoRec.Rec.Rid = RID_PHY_TYPE;

    Status =
    S24tCopyUp(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        CR_ACCESS
    );
    if (!Status)
    {
        DBG_NOTICE(DbgInfo,"Phy type %u\n",
            (uint) *(__u16 *)&lp->InfoRec.Data[0]
        );
    }
    else
    {
        return(Status);
    }

    memset(&lp->InfoRec,0,sizeof(S24tInfoRecType));

    lp->InfoRec.Rec.WordLen = 2;
    lp->InfoRec.Rec.Rid = RID_CURRENT_CHANNEL;

    Status =
    S24tCopyUp(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        CR_ACCESS
    );
    if (!Status)
    {
        DBG_NOTICE(DbgInfo,"Current channel %u\n",
            (uint) *(__u16 *)&lp->InfoRec.Data[0]
        );
    }
    else
    {
        return(Status);
    }

    return(Status);
}


STATIC int S24tGetFirmwareVersion(
    struct S24T_private *lp
    )
{
    int Status;

    memset(&lp->InfoRec,0,sizeof(S24tInfoRecType));
    lp->InfoRec.Rec.WordLen = S24_FW_VERSION_LEN/2;
    lp->InfoRec.Rec.Rid = RID_SECONDARY_VERSION;

    Status =
    S24tCopyUp(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        CR_ACCESS
    );
    if (!Status)
    {
        memcpy(lp->FwVersion,lp->InfoRec.Data,S24_FW_VERSION_LEN);
        lp->FwVersion[S24_FW_VERSION_LEN] = '\0';
    }
    else
    {
        DBG_PRINT("GetFirmwareVersion - failed to get firmware version.\n");
        return(Status);
    }

    memset(&lp->InfoRec,0,sizeof(S24tInfoRecType));
    lp->InfoRec.Rec.WordLen = S24_FW_VERSION_LEN/2;
    lp->InfoRec.Rec.Rid = RID_SECONDARY_DATE;

    Status =
    S24tCopyUp(
        lp,
        lp->InfoRec.Rec.Rid,
        (__u8 *)&lp->InfoRec,
        (lp->InfoRec.Rec.WordLen+1)*2,
        CR_ACCESS
    );
    if (!Status)
    {
        memcpy(lp->FwDate,lp->InfoRec.Data,S24_FW_VERSION_LEN);
        lp->FwDate[S24_FW_VERSION_LEN] = '\0';
    }
    else
    {
        DBG_PRINT("GetFirmwareVersion - failed to get firmware date.\n");
        return(Status);
    }

    return(Status);
}


STATIC int
S24tWaitForCommandNotBusy(
    struct S24T_private *lp,
    __u16               *CmdStatus
    )
{
    DBG_FUNC("S24tWaitForCommandNotBusy")

    __u32 MsMaxDelay;

    MsMaxDelay = 50000;

    // Make sure the adapter has settled before reading.
    // LKA -- 02/04/01
    udelay(10);
    while (
           MsMaxDelay
           &&
           (((*CmdStatus) = S24tRead16(lp,NIC_COMMAND)) & COMMAND_REGISTER_BUSY)
          )
    {
        udelay(100);
        if ((--MsMaxDelay) == 0)
        {
            DBG_PRINT("CmdStatus timeout, last value %04x.\n",(*CmdStatus));
            DBG_PRINT("EvStat looks like %04x.\n", S24tRead16(lp,NIC_EVSTAT));
            DBG_ASSERT(0);
            return(-ENXIO);
        }
    }

    return(0);
}

STATIC int
S24tCommand(
    struct S24T_private *lp,
    __u16               Command,
    __u16               *Result,
    __u16               *P0,
    __u16               *P1,
    __u16               *P2
    )
{
    DBG_FUNC("S24tCommand")
    int Status;
    __u16 CmdStatus;
    __u16 EvStatus;
    __u32 MsMaxDelay;

    DBG_ENTER(DbgInfo);

    if (!S24tCardInserted(lp))
    {
        DBG_LEAVE(DbgInfo);
        return(-ENXIO);
    }

    if ((Status=S24tWaitForCommandNotBusy(lp,&CmdStatus)))
    {
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    if (P0) S24tWrite16(lp,NIC_PARAM0,*P0);
    if (P1) S24tWrite16(lp,NIC_PARAM1,*P1);
    if (P2) S24tWrite16(lp,NIC_PARAM2,*P2);


    // The windows driver checks the command register here as well, and I figure that it can't hurt
    // LKA -- 02/04/01
    if ((Status = S24tWaitForCommandNotBusy(lp, &CmdStatus)))
    {
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    S24tWrite16(lp,NIC_COMMAND,Command);
    udelay(10);

    MsMaxDelay = 100000;
    while (!(Status = S24tEvStat(lp,&EvStatus)))
    {
        if (EvStatus & EVS_CMD)
        {
            break;
        }

        udelay(10);
        if ((--MsMaxDelay) == 0)
        {
            if (Command == CR_INITIALIZE)
            {
                break;
            }
            DBG_PRINT("EvStatus command timeout, last value %04x.\n",EvStatus);
            DBG_ASSERT(0);
            DBG_LEAVE(DbgInfo);
            return(-ENXIO);
        }
    }

    if (Status)
    {
        DBG_PRINT("EvStat read error.\n");
        DBG_LEAVE(DbgInfo);
        return(Status);
    }

    if (P0) *P0 = S24tRead16(lp,NIC_RESP0);
    if (P1) *P1 = S24tRead16(lp,NIC_RESP1);
    if (P2) *P2 = S24tRead16(lp,NIC_RESP2);

    CmdStatus = S24tRead16(lp,NIC_STATUS);

    if ((CmdStatus & 0x3f) != (Command & 0x3f))
    {
        DBG_PRINT("S24tCommand Status %04x != Command %04x\n",(CmdStatus & 0x3f),(Command & 0x3f));
    }

    CmdStatus >>= 8;
    CmdStatus &= 0x7f;

    if (CmdStatus)
    {
        DBG_PRINT("CmdStatus %04x Command %04x P0 %04x P1 %04x P2 %04x\n",CmdStatus,Command,(P0 ? *P0 : 0),(P1 ? *P1 : 0),(P2 ? *P2 : 0));
        return(-EFAULT);
    }
    DBG_ASSERT(!CmdStatus);

    if (Result) *Result = CmdStatus;

    S24tWrite16(lp,NIC_EVACK,ACK_CMD);

    DBG_LEAVE(DbgInfo);
    return(0);
}

STATIC __u16
S24tRead16(
    struct S24T_private *lp,
    __u16               offset
    )
{
    __u16 val;
    if (!S24tCardInserted(lp))
    {
        return(S24T_EJECTED);
    }

    val = inw( lp->dev->base_addr+offset );

    return( val );
}

STATIC void
S24tReads(
    struct S24T_private *lp,
    __u8                *dest,
    __u16               offset,
    __u32               len
    )
{
    if (!S24tCardInserted(lp))
    {
        memset(dest,(char)S24T_EJECTED,len);
        return;
    }
    insw( lp->dev->base_addr+offset, dest, len/2 );
    if (len & 1)
    {
        *(dest+(len-1)) = (__u8)inw(lp->dev->base_addr+offset);
    }
}

STATIC void
S24tWrite16(
    struct S24T_private *lp,
    __u16               offset,
    __u16               val
    )
{
    if (!S24tCardInserted(lp))
    {
        DBG_PRINT("Write attempt while card not inserted.\n");
        return;
    }

    outw( val, lp->dev->base_addr+offset );
}

STATIC void
S24tWrites(
    struct S24T_private *lp,
    __u8                *src,
    __u16               offset,
    __u32               len
    )
{
    if (!S24tCardInserted(lp))
    {
        DBG_PRINT("Write string attempt while card not inserted.\n");
        return;
    }

    outsw( lp->dev->base_addr+offset, src, len/2 );
    if (len & 1)
    {
        outb(*(src+(len-1)), lp->dev->base_addr+offset);
    }
}

void
S24tHalt(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tHalt")

    DBG_ENTER(DbgInfo);

    DBG_PRINT("lp->Cor[OFFSET_COR] = %x\n",lp->Cor[OFFSET_COR]);
    lp->Cor[OFFSET_COR] |= COR_RESET;
    mdelay(1);
    DBG_PRINT("lp->Cor[OFFSET_COR] = %x\n",lp->Cor[OFFSET_COR]);

    DBG_LEAVE(DbgInfo);
}

STATIC void
S24ClearHalt(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24ClearHalt")

    DBG_ENTER(DbgInfo);

    DBG_PRINT("lp->Cor[OFFSET_COR] = %x\n",lp->Cor[OFFSET_COR]);
    lp->Cor[OFFSET_COR] &= (~COR_RESET);
    mdelay(1);
    DBG_PRINT("lp->Cor[OFFSET_COR] = %x\n",lp->Cor[OFFSET_COR]);

    DBG_LEAVE(DbgInfo);
}

STATIC void
S24SetIoMode(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24SetIoMode")
    DBG_ENTER(DbgInfo);

    DBG_PRINT("lp->Cor[OFFSET_COR] = %x\n",lp->Cor[OFFSET_COR]);
    lp->Cor[OFFSET_COR] |= COR_IO_MODE;
    mdelay(1);
    DBG_PRINT("lp->Cor[OFFSET_COR] = %x\n",lp->Cor[OFFSET_COR]);

    DBG_LEAVE(DbgInfo);
}

#if FIXME
STATIC int
S24tResetCOR(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tResetCOR")
    int Status;

    DBG_ENTER(DbgInfo);

    S24tHalt(lp);

    S24ClearHalt(lp);
    S24ClearHalt(lp);

    S24SetIoMode(lp);
    S24SetIoMode(lp);

    mdelay(10);

    Status = S24tCardInserted(lp);

    DBG_LEAVE(DbgInfo);
    return(Status ? 0 : (-ENODEV));
}
#endif

STATIC int
S24tReadMacAddress(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tReadMacAddress")
    int Status;

    DBG_ENTER(DbgInfo);

    Status =
    S24tCopyUp(
        lp,
        RID_CFG_MAC_ADDRESS,
        (__u8 *)&lp->InfoRec.Rec,
        sizeof(S24tRidType) + ETH_ALEN,
        CR_ACCESS
    );
    if (Status)
    {
        DBG_PRINT("S24tReadMacAddress failed\n");
    }
    DBG_NOTICE(DbgInfo,"Mac Addr %02x:%02x:%02x:%02x:%02x:%02x\n",
        lp->InfoRec.Data[0],
        lp->InfoRec.Data[1],
        lp->InfoRec.Data[2],
        lp->InfoRec.Data[3],
        lp->InfoRec.Data[4],
        lp->InfoRec.Data[5]
    );
    DBG_LEAVE(DbgInfo);
    return(0);
}

STATIC void
S24tParseInfoRec(
    struct S24T_private *lp
    )
{
    DBG_FUNC("S24tParseInfoRec")

    DBG_NOTICE(DbgInfo,"Info Rec %s\n",S24tRidDebug(lp->InfoRec.Rec.Rid));

    switch (lp->InfoRec.Rec.Rid)
    {
    case RID_INFO_LINK_STATUS:
#if FIXME
            /*
             * Note transitions.
             */
            if (lp->LinkStatus != *(__u16 *)lp->InfoRec.Data)
            {
                if (*(__u16 *)lp->InfoRec.Data == 1)
                {
                    lp->dev->tbusy = 0;
                }
                else
                {
                    lp->dev->tbusy = 1;
                }
            }
#endif
            lp->LinkStatus = *(__u16 *)lp->InfoRec.Data;
            DBG_PRINT("%s: Link status %u\n",lp->dev->name,(uint)lp->LinkStatus);
#if FIXME
            S24tGeneralInfo(lp);
#endif
            break;
        default:
            break;
    }
}

#if DBG
typedef struct {
    __u16   Rid;
    char    *Name;
} RidDebugType;

RidDebugType RidDebug[] = {
    {RID_INFO_HANDOVER_ADDR,        "RID_INFO_HANDOVER_ADDR"},
    {RID_INFO_COMM_TALLIES,         "RID_INFO_COMM_TALLIES"},
    {RID_INFO_SCAN_RESULTS,         "RID_INFO_SCAN_RESULTS"},
    {RID_INFO_LINK_STATUS,          "RID_INFO_LINK_STATUS"},
    {RID_INFO_ASSOCIATION_STATUS,   "RID_INFO_ASSOCIATION_STATUS"},
    {RID_CFG_PORT_TYPE,             "RID_CFG_PORT_TYPE"},
    {RID_CFG_MAC_ADDRESS,           "RID_CFG_MAC_ADDRESS"},
    {RID_CFG_SSID,                  "RID_CFG_SSID"},
    {RID_CFG_CHANNEL,               "RID_CFG_CHANNEL"},
    {RID_CFG_OWN_SSID,              "RID_CFG_OWN_SSID"},
    {RID_CFG_OWN_ATIM,              "RID_CFG_OWN_ATIM"},
    {RID_CFG_SYSTEM_SCALE,          "RID_CFG_SYSTEM_SCALE"},
    {RID_CFG_MAX_DATA_LENGTH,       "RID_CFG_MAX_DATA_LENGTH"},
    {RID_CFG_WDS_ADDRESS,           "RID_CFG_WDS_ADDRESS"},
    {RID_CFG_PM_ENABLE,             "RID_CFG_PM_ENABLE"},
    {RID_CFG_PM_EPS,                "RID_CFG_PM_EPS"},
    {RID_CFG_MULTICAST_RCV,         "RID_CFG_MULTICAST_RCV"},
    {RID_CFG_MAX_SLEEP_TIME,        "RID_CFG_MAX_SLEEP_TIME"},
    {RID_CFG_PM_HOLDOVER_TIME,      "RID_CFG_PM_HOLDOVER_TIME"},
    {RID_CFG_NAME,                  "RID_CFG_NAME"},
    {RID_CFG_MKK_FLAG,              "RID_CFG_MKK_FLAG"},
    {RID_CFG_PREFERRED_BSS_ID,      "RID_CFG_PREFERRED_BSS_ID"},
    {RID_CFG_MANDATORY_BSS_ID,      "RID_CFG_MANDATORY_BSS_ID"},
    {RID_CFG_AUTHENTICATION,        "RID_CFG_AUTHENTICATION"},
    {RID_CFG_ENCRYPT_INDEX,         "RID_CFG_ENCRYPT_INDEX"},
    {RID_CFG_KEY1,                  "RID_CFG_KEY1"},
    {RID_CFG_KEY2,                  "RID_CFG_KEY2"},
    {RID_CFG_KEY3,                  "RID_CFG_KEY3"},
    {RID_CFG_KEY4,                  "RID_CFG_KEY4"},
    {RID_CFG_ENCRYPT_FLAG,          "RID_CFG_ENCRYPT_FLAG"},
    {RID_CFG_AUTH_TYPE,             "RID_CFG_AUTH_TYPE"},
    {RID_CFG_KEY_LENGTH,            "RID_CFG_KEY_LENGTH"},
    {RID_GROUP_ADDR,                "RID_GROUP_ADDR"},
    {RID_CREATE_IBSS,               "RID_CREATE_IBSS"},
    {RID_FRAG_THRESHOLD,            "RID_FRAG_THRESHOLD"},
    {RID_RTS_THRESHOLD,             "RID_RTS_THRESHOLD"},
    {RID_TX_RATE,                   "RID_TX_RATE"},
    {RID_PROMISCUOUS_MODE,          "RID_PROMISCUOUS_MODE"},
    {RID_DIVERSITY,                 "RID_DIVERSITY"},
    {RID_OPTIONS,                   "RID_OPTIONS"},
    {RID_IF_FREQUENCY,              "RID_IF_FREQUENCY"},
    {RID_BASIC_RATES,               "RID_BASIC_RATES"},
    {RID_TX_INHIBIT,                "RID_TX_INHIBIT"},
    {RID_CLEAR_STATS,               "RID_CLEAR_STATS"},
    {RID_TICK_TIME,                 "RID_TICK_TIME"},
    {RID_MAX_DOWNLOAD_TIME,         "RID_MAX_DOWNLOAD_TIME"},
    {RID_DOWNLOAD_BUFFER,           "RID_DOWNLOAD_BUFFER"},
    {RID_PRI_IDENTITY,              "RID_PRI_IDENTITY"},
    {RID_PRI_SUP_RANGE,             "RID_PRI_SUP_RANGE"},
    {RID_PRIMARY_VERSION,           "RID_PRIMARY_VERSION"},
    {RID_PRIMARY_DATE,              "RID_PRIMARY_DATE"},
    {RID_NIC_SERIAL,                "RID_NIC_SERIAL"},
    {RID_NIC_FW_ID,                 "RID_NIC_FW_ID"},
    {RID_CHANNEL_LIST,              "RID_CHANNEL_LIST"},
    {RID_REGULATORY,                "RID_REGULATORY"},
    {RID_TEMP_TYPE,                 "RID_TEMP_TYPE"},
    {RID_CIS,                       "RID_CIS"},
    {RID_COUNTRY_CODE,              "RID_COUNTRY_CODE"},
    {RID_COUNTRY_NAME,              "RID_COUNTRY_NAME"},
    {RID_MKK_CALL_SIGN,             "RID_MKK_CALL_SIGN"},
    {RID_MKK_RSSI,                  "RID_MKK_RSSI"},
    {RID_ORDINALS,                  "RID_ORDINALS"},
    {RID_STA_IDENTITY,              "RID_STA_IDENTITY"},
    {RID_STA_SUP_RANGE,             "RID_STA_SUP_RANGE"},
    {RID_MFI_ACT_RANGE,             "RID_MFI_ACT_RANGE"},
    {RID_CFI_ACT_RANGE,             "RID_CFI_ACT_RANGE"},
    {RID_SECONDARY_VERSION,         "RID_SECONDARY_VERSION"},
    {RID_SECONDARY_DATE,            "RID_SECONDARY_DATE"},
    {RID_PORT_STATUS,               "RID_PORT_STATUS"},
    {RID_CURRENT_SSID,              "RID_CURRENT_SSID"},
    {RID_CURRENT_BSSID,             "RID_CURRENT_BSSID"},
    {RID_COMMS_QUALITY,             "RID_COMMS_QUALITY"},
    {RID_CURR_TX_RATE,              "RID_CURR_TX_RATE"},
    {RID_SHORT_RETRY,               "RID_SHORT_RETRY"},
    {RID_LONG_RETRY,                "RID_LONG_RETRY"},
    {RID_PRIVACY,                   "RID_PRIVACY"},
    {RID_PHY_TYPE,                  "RID_PHY_TYPE"},
    {RID_CURRENT_CHANNEL,           "RID_CURRENT_CHANNEL"},
    {RID_POWER_MODE,                "RID_POWER_MODE"},
    {RID_CCA_MODE,                  "RID_CCA_MODE"},
    {0,                             "Unknown"}
};

STATIC char *
S24tRidDebug(
    __u16 Rid
    )
{
    RidDebugType *p = RidDebug;
    while (p->Rid)
    {
        if (p->Rid == Rid)
        {
            return(p->Name);
        }
        p++;
    }
    return(p->Name);
}
#endif

